#ifndef MACH_BCM2708_IPC_H
#define MACH_BCM2708_IPC_H

/******************************************************************************
 FIFO helpers
 *****************************************************************************/

typedef struct
{
	/* ptr to the read/write variables */
   volatile unsigned long *write;
   volatile unsigned long *read;

	/* pointer to the base memory for the fifo */
   volatile void *base;

	/* size of the fifo in terms of entries */
   uint32_t size;

	/* size of each entry */
   uint32_t entry_size;

} IPC_FIFO_T;

static inline void ipc_fifo_reset(IPC_FIFO_T *fifo)
{
   *fifo->write = 0;
   *fifo->read = 0;
}

static inline
void ipc_fifo_setup(IPC_FIFO_T *fifo, volatile unsigned long *write,
			volatile unsigned long *read, volatile void *base_ptr,
			const uint32_t size, const uint32_t entry_size)
{
   fifo->write = write;
   fifo->read = read;
   fifo->size = size;
   fifo->entry_size = entry_size;
   fifo->base = base_ptr;

   ipc_fifo_reset( fifo );
}

static inline
void ipc_fifo_setup_no_reset(IPC_FIFO_T *fifo, volatile unsigned long *write,
				volatile unsigned long *read, volatile void *base_ptr,
				const uint32_t size, const uint32_t entry_size)
{
   fifo->write = write;
   fifo->read = read;
   fifo->size = size;
   fifo->entry_size = entry_size;
   fifo->base = base_ptr;
}

static inline int32_t ipc_fifo_full( const IPC_FIFO_T *fifo )
{
	int32_t read  = *fifo->read;
	int32_t write = *fifo->write;
        return write - read == fifo->size ? 1 : 0;
}

static inline int32_t ipc_fifo_empty( const IPC_FIFO_T *fifo )
{
   uint32_t read = *fifo->read;
   uint32_t write = *fifo->write;
	return write == read ? 1 : 0;
}

static inline int32_t ipc_fifo_num_entries(const IPC_FIFO_T *fifo)
{
	int32_t read  = *fifo->read;
	int32_t write = *fifo->write;
   return (write - read);
}

static inline int32_t ipc_fifo_write( const IPC_FIFO_T *fifo, const void *entry )
{
	if (!ipc_fifo_full(fifo)) {
      uint32_t write = *fifo->write;
		void *fifo_entry = (void *)((int)fifo->base + fifo->entry_size * (write % fifo->size));

      memcpy( fifo_entry, entry, fifo->entry_size );
		*fifo->write = write + 1;
      return(1);
   }
   return(0);
}

static inline int32_t ipc_fifo_read( const IPC_FIFO_T *fifo, void *entry )
{
	if (!ipc_fifo_empty(fifo)) {
      uint32_t read = *fifo->read;
		void *fifo_entry = (void *)((int)fifo->base + fifo->entry_size * (read % fifo->size));

      memcpy( entry, fifo_entry, fifo->entry_size );

		/* inc the read ptr */
		*fifo->read = read + 1;
      return(1);
   }
   return(0);
}

static inline const void *ipc_fifo_peek(const IPC_FIFO_T *fifo, int index)
{
	return (const void *)((int)fifo->base + fifo->entry_size * (index % fifo->size));
}

#endif /* ifndef MACH_BCM2708_IPC_H */
