//#include <linux/vmalloc.h>
#include <linux/mm.h> // get_user_pages
#include <linux/pagemap.h> // page_cache_release
#include <linux/dma-mapping.h>
#include <linux/semaphore.h>
//#include <linux/module.h>
#include <linux/kernel.h>

#include "bcm2708_zerocopy.h"
#include "bcm2708_mdec.h"


#define assert(c) BUG_ON(!(c))
#define page_to_bus(p) __phys_to_bus(page_to_phys(p))


// ===================================================================

static void AddFragment( MEDIA_DEC_FRAGMENT * * Fragment, uint32_t VideoCorePhysical, uint32_t Bytes )
{
  MEDIA_DEC_FRAGMENT * Current = *Fragment;
  if (Current->Physical + Current->Bytes == VideoCorePhysical)
    Current->Bytes += Bytes;
  else
  {
    *Fragment = ++Current;
    Current->Physical = VideoCorePhysical;
    Current->Bytes    = Bytes;
  }
}

static unsigned int MapPages( void * Buffer, unsigned int Bytes, MEDIA_DEC_FRAGMENT * Fragment, unsigned int Pages, struct page * * Page, int Write )
{
  unsigned int            StartOffset = (unsigned int) Buffer & (PAGE_SIZE - 1);
  enum dma_data_direction Direction   = Write != 0 ? DMA_FROM_DEVICE : DMA_TO_DEVICE;
  MEDIA_DEC_FRAGMENT    * Current     = Fragment;
//printk( KERN_ERR "%s: b %p +%4x f %p n %2x page %p w %d\n",
//  __FUNCTION__, Buffer, Bytes, Fragment, Pages, Page, Write );
  if (Pages == 1)
  {
    Current->Physical = dma_map_page( NULL, Page[0], StartOffset, Bytes, Direction );
    Current->Bytes    = Bytes;
  }
  else
  {
    unsigned int i;
    unsigned int LastOffset = (StartOffset + Bytes - 1) & ~PAGE_MASK;
    uint32_t     Physical;
    uint32_t     Bytes;

    // First fragment
    Current->Bytes    = PAGE_SIZE - StartOffset;
    Current->Physical = dma_map_page( NULL, *Page, StartOffset, Current->Bytes, Direction );
//printk( KERN_ERR "%s: Map page (%4u) %p +%4x phys %08x", __FUNCTION__, 0, *Page, Current->Bytes, Current->Physical );
    for( i = 1, ++Page ;
         i < Pages - 1 ;
       ++i,     ++Page )
    {
      Physical = dma_map_page( NULL, *Page, 0, Bytes = PAGE_SIZE, Direction );
      AddFragment( &Current, Physical, Bytes );
    }
    Physical = dma_map_page( NULL, *Page, 0, Bytes = LastOffset + 1, Direction );
    AddFragment( &Current, Physical, Bytes );
  }
//printk( KERN_ERR "%s: Physical fragments %2x\n", __FUNCTION__, Current - Fragment + 1 );
  return Current - Fragment + 1;
}


// ===================================================================

void UnmapPageClean( struct page * Page, dma_addr_t Physical, unsigned int Bytes, enum dma_data_direction Direction )
{
  dma_unmap_page( NULL, Physical, Bytes, Direction );
  page_cache_release( Page );
}

void UnmapPageDirty( struct page * Page, dma_addr_t Physical, unsigned int Bytes, enum dma_data_direction Direction )
{
//printk( KERN_ERR "%s: release page %p p %08x +%4x\n", __FUNCTION__, Page, Physical, Bytes );
  dma_unmap_page( NULL, Physical, Bytes, Direction );
  if (!PageReserved( Page ))
    SetPageDirty( Page ); // We've read data in, so they're now dirty
  page_cache_release( Page );
}

static void UnmapPages( const MEDIA_DEC_FRAGMENT * Fragment, unsigned int Fragments, struct page * * Page, unsigned int Pages, int Write )
{
  enum dma_data_direction Direction = Write != 0 ? DMA_FROM_DEVICE : DMA_TO_DEVICE;
  void (* UnmapPage)( struct page * Page, dma_addr_t Physical, unsigned int Bytes, enum dma_data_direction Direction );
  unsigned int StartOffset = Fragment[0].Physical                                               & ~PAGE_MASK;
  unsigned int LastOffset  = (Fragment[Fragments-1].Physical + Fragment[Fragments-1].Bytes - 1) & ~PAGE_MASK;

  UnmapPage = Write == 0 ? UnmapPageClean : UnmapPageDirty;
//printk( KERN_ERR "%s: so %4x lo %4x (unmap %p)\n", __FUNCTION__, StartOffset, LastOffset, UnmapPage );
  if (Pages == 1)
  {
    (*UnmapPage)( Page[0], page_to_bus( Page[0] ) + StartOffset, LastOffset - StartOffset + 1, Direction );
  }
  else
  {
    unsigned int i;
    (*UnmapPage)( Page[0], page_to_bus( Page[0] ) + StartOffset, PAGE_SIZE - StartOffset, Direction );
    for( i = 1 ; i < Pages - 1 ; ++i )
      (*UnmapPage)( Page[i], page_to_bus( Page[i] ), PAGE_SIZE, Direction );
    (*UnmapPage)( Page[i], page_to_bus( Page[i] ), LastOffset + 1, Direction );
  }
}

// ===================================================================


int ZeroCopy_CreateFragmentList(
  struct task_struct * ClientTask,
  void               * User,
  unsigned int         UserBytes,
  MEDIA_DEC_FRAGMENT * Fragment,
  struct page      * * Page,
  unsigned int       * Pages,
  unsigned int         MaxPages,
  int                  Write )
{
  unsigned int Fragments;
  unsigned int Virtual = (unsigned int) User;
  unsigned int n       = *Pages = ((Virtual & ~PAGE_MASK) + UserBytes - 1) / PAGE_SIZE + 1;
  int          Status;
//printk( KERN_ERR "%s: ct %p u %p +%4x f %p p %p n %4x w %d\n",
//  __FUNCTION__, ClientTask, User, UserBytes, Fragment, Page, MaxPages, Write );
  assert( Virtual < PAGE_OFFSET ); // We currently only handle user-side addresses
  assert( UserBytes != 0 );
  assert( n <= MaxPages );

  // Lock the user's pages
  down_read( &ClientTask->mm->mmap_sem );
  Status = get_user_pages(
             ClientTask,
             ClientTask->mm,
             Virtual,
             n,
             Write,
             0 /* Force */,
             Page,
             NULL /* VMAs */ );
  up_read( &ClientTask->mm->mmap_sem );
  if (Status != n)
     return -EFAULT;

  // Map them for DMA (handles cache coherency)
  Fragments = MapPages( User, UserBytes, Fragment, n, Page, Write );
  return (int) Fragments;
}

void ZeroCopy_DeleteFragmentList(
  MEDIA_DEC_FRAGMENT * Fragment,
  unsigned int         Fragments,
  struct page      * * Page,
  unsigned int         Pages,
  int                  Write )
{
//printk( KERN_ERR "%s: pages %u\n", __FUNCTION__, Pages );
  UnmapPages( Fragment, Fragments, Page, Pages, Write );
}


// ===================================================================
