/*
 *  wdog.c - gencmd-based watchdog driver
 *
 *  Copyright (C) 2011 Roku, Inc.
 *
 * This device handles watchdog timeouts longer than provided by the hardware.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/sysrq.h>
#include <linux/sched.h>

#include "../misc/vmcs_services/gencmd/gencmd.h"

extern int die_counter;
extern int printk_time;
extern int is_roku_application(struct task_struct *t);

#define HW_TIMEOUT_MS       16000 /* note: maximum is 16777 */
#define CHECK_INTERVAL_MS   1000  /* should be much less than HW_TIMEOUT_MS */
#define WARN_PCT            80    /* warning occurs at this percent of timeout */

static void wdog_warn(unsigned long arg); /* forward declaration */

static struct task_struct * s_wdog_thread = NULL; /* thread to update timers */
static uint64_t             s_wdog_time = 0;      /* watchdog expiration time */
DEFINE_MUTEX(               s_mutex );
DEFINE_TIMER(               s_wdog_warn_timer, wdog_warn, 0, 0);


static uint64_t now_ms(void)
{
    ktime_t ktime = ktime_get();
    uint64_t ns = ktime_to_ns(ktime);
    do_div(ns, 1000000); /* convert nanosec to millisec */
    return ns;
}

static void wdog_set_hw_timeout(long timeout)
{
    /* Hardware will reboot in timeout millisec. */
    char response[16];
    vc_gencmd(response, sizeof(response), "wdog %ld", timeout);
}

static void wdog_disable(void)
{
    mutex_lock(&s_mutex);
    del_timer_sync(&s_wdog_warn_timer);
    s_wdog_time = 0;
    wdog_set_hw_timeout(0);
    mutex_unlock(&s_mutex);
}

static void wdog_warn(unsigned long arg)
{
    struct task_struct *p;
    int save_printk_time = printk_time;

    printk("Watchdog Warning at %ld!\n", (long) now_ms());
    printk_time = 0;
    handle_sysrq('t',0); /* dump scheduler state */
    handle_sysrq('m',0); /* dump memory stats */
    dump_stack();
    printk_time = save_printk_time;

#if 1 /* ifdef CONFIG_SPECIAL_APPLICATION */
    for_each_process(p) {
        if (is_roku_application(p)) {
            send_sig(SIGKILL, p, 1);
            break;
        }
    }
#endif
}

static int wdog_thread(void* arg)
{
    while (!kthread_should_stop()) {
        msleep_interruptible(CHECK_INTERVAL_MS);
        mutex_lock(&s_mutex);
        if (s_wdog_time > 0) {
            /* Watchdog is enabled. */
            if (now_ms() < s_wdog_time) {
                /* Tickle the hardware. */
                if (die_counter == 0)
                    wdog_set_hw_timeout(HW_TIMEOUT_MS);
            } else {
                /* Software timeout has expired: force reboot. */
                console_loglevel = 7; /* force following printk to the console */
                printk("watchdog expired: rebooting now at %ld\n", (long) now_ms());
                wdog_set_hw_timeout(1);
                for (;;) ;
            }
        }
        mutex_unlock(&s_mutex);
    }
    return 0;
}

static void wdog_set_timeout(long incr_ms)
{
    uint64_t now = now_ms();
    mutex_lock(&s_mutex);
    s_wdog_time = now + incr_ms;
    mod_timer(&s_wdog_warn_timer, jiffies + ((HZ * WARN_PCT * incr_ms) / (1000*100)));
    mutex_unlock(&s_mutex);
}

static const struct watchdog_info wdog_info = {
    .identity     = "bcm_vcore_wdt",
    .options     = WDIOF_SETTIMEOUT
};

static long wdog_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    void __user *argp = (void __user *)arg;
    int __user *p = argp;
    int seconds;

    switch (cmd) {
    case WDIOC_GETSUPPORT:
        return copy_to_user(argp, &wdog_info, sizeof(wdog_info)) ? -EFAULT : 0;

    case WDIOC_GETSTATUS:
    case WDIOC_GETBOOTSTATUS:
        return put_user(0, p);

    case WDIOC_SETTIMEOUT:
        if (get_user(seconds, p))
            return -EFAULT;
        if (seconds == 0)
            wdog_disable();
        else
            wdog_set_timeout(seconds * 1000);
        return 0;

    default:
        return -ENOTTY;
    }
}

static int wdog_open(struct inode *inode, struct file *filp)
{
    return 0;
}

static int wdog_release(struct inode *inode, struct file *filp)
{
    return 0;
}

static const struct file_operations wdog_fops = {
    .owner        = THIS_MODULE,
    .llseek        = no_llseek,
    .unlocked_ioctl    = wdog_ioctl,
    .open        = wdog_open,
    .release    = wdog_release,
};

static struct miscdevice wdog_dev = {
    .minor        = WATCHDOG_MINOR,
    .name        = "watchdog",
    .fops        = &wdog_fops,
};

static int __init wdog_init(void)
{
    int ret = misc_register(&wdog_dev);
    if (ret < 0) {
        printk(KERN_ERR "bcm_vcore_wdt failed to register device\n");
    } else {
        wdog_disable();
        s_wdog_thread = kthread_run(wdog_thread, 0, "bcm_watchdog");
    }
    return ret;
}

static void __exit wdog_exit(void)
{
    wdog_disable();
    kthread_stop(s_wdog_thread);
    misc_deregister(&wdog_dev);
}

module_init(wdog_init);
module_exit(wdog_exit);

MODULE_AUTHOR("Mark Nudelman");
MODULE_DESCRIPTION("Watchdog driver for Broadcom Videocore");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
