#include <linux/module.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <mach/platform.h>
#include <linux/stat.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <mach/ipc.h>

#include <linux/broadcom/vc_loopaudio.h>

static struct {
    volatile VC_LOOPAUDIO_MAILBOX_T *mailbox;
} loopaudio_state;

/* -------------------------------------------------------------------- */
/* Implementation of loopaudio driver */

static int __devexit loopaudio_remove(struct platform_device *pdev)
{
    return 0;
}

static int __devinit loopaudio_probe(struct platform_device *pdev)
{
    struct resource *res;

    loopaudio_state.mailbox = NULL;

    res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (res == NULL) {
        dev_err(&pdev->dev, "failed to get memory region resource\n");
        return -ENODEV;
    }

    /* Get pointer to videocore loopaudio buffer. */
    loopaudio_state.mailbox = (void __iomem *) (res->start);

    return 0;
}

static struct platform_driver loopaudio_driver = {
    .probe     = loopaudio_probe,
    .remove    = __devexit_p (loopaudio_remove),
    .driver    = {
      .name    =  "bcm2835_LAUD",
      .owner   =   THIS_MODULE,
                 },
};

/* -------------------------------------------------------------------- */
/* Loopaudio device (implementation of /dev/vc-loopaudio) */

#define LOOPAUDIO_DRIVER_NAME  "vc-loopaudio"

static uint64_t side_channel_total_read = 0;

struct loopaudio {
#ifdef LOOPAUDIO_USER_IO
    uint64_t total_read;
#endif
};

static int loopaudio_getflags(uint32_t* pflags)
{
    if (!loopaudio_state.mailbox) return -ENOTCONN;
    *pflags = loopaudio_state.mailbox->lflags;
    return 0;
}

static int loopaudio_setflags(uint32_t flags)
{
    if (!loopaudio_state.mailbox) return -ENOTCONN;
    loopaudio_state.mailbox->lflags = flags;
    return 0;
}

static int loopaudio_open(struct inode *inode, struct file *file)
{
    struct loopaudio* loopaudio;
    if (!loopaudio_state.mailbox) return -ENOTCONN;
    loopaudio = kmalloc(sizeof(struct loopaudio), GFP_KERNEL);
    if (!loopaudio)
        return -ENOMEM;
#ifdef LOOPAUDIO_USER_IO
    loopaudio->total_read = loopaudio_state.mailbox->written;
#endif
    file->private_data = loopaudio;
    return 0;
}

static int loopaudio_release(struct inode *inode, struct file *file)
{
    struct loopaudio* loopaudio = (struct loopaudio*) file->private_data;
    kfree(loopaudio);
    return 0;
}

void vc_side_channel_read(uint32_t *cnt, unsigned char** buf, uint32_t max_chunk)
{
    uint32_t max;
    if (loopaudio_state.mailbox->written - side_channel_total_read > VC_LOOPAUDIO_BUFFER_SIZE/2)
        side_channel_total_read = loopaudio_state.mailbox->written;

    *buf = (unsigned char*) &loopaudio_state.mailbox->buffer[side_channel_total_read % VC_LOOPAUDIO_BUFFER_SIZE];
    max = VC_LOOPAUDIO_BUFFER_SIZE - (side_channel_total_read % VC_LOOPAUDIO_BUFFER_SIZE);
    *cnt = loopaudio_state.mailbox->written - side_channel_total_read;
    if (*cnt > max) *cnt = max;
    if (*cnt > max_chunk) *cnt = max_chunk;
    side_channel_total_read += *cnt;
}
EXPORT_SYMBOL(vc_side_channel_read);

static int loopaudio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    // struct loopaudio* loopaudio = (struct loopaudio*) file->private_data;
    int ret;

    switch (cmd) {
    case VC_IOCTL_LOOPAUDIO_SETFLAGS: {
        uint32_t flags;
        if (copy_from_user(&flags, (void __user*) arg, sizeof(flags)))
            return -EFAULT;
        return loopaudio_setflags(flags);
    }
    case VC_IOCTL_LOOPAUDIO_GETFLAGS: {
        uint32_t flags;
        ret = loopaudio_getflags(&flags);
        if (ret) return ret;
        if (copy_to_user((void __user*) arg, &flags, sizeof(flags)))
            return -EFAULT;
        return 0;
    }
    default:
        printk(KERN_ERR"[%s]:unknown command cmd=%d\n", __func__, cmd);
        return -EIO;
    }
}

struct file_operations loopaudio_fops = {
	.owner =	THIS_MODULE,
	.ioctl =	loopaudio_ioctl,
	.open =		loopaudio_open,
	.release =	loopaudio_release,
};

struct miscdevice loopaudio_device = {
    .minor =    MISC_DYNAMIC_MINOR,
    .name =     LOOPAUDIO_DRIVER_NAME,
    .fops =     &loopaudio_fops
};

/* -------------------------------------------------------------------- */

static int __init loopaudio_init(void)
{
	int ret;
    printk(KERN_ERR "Roku loopaudio driver installed");

    /* Register driver */
    ret = platform_driver_register(&loopaudio_driver);
    if (ret < 0)
        return ret;

    /* Register device */
    ret = misc_register(&loopaudio_device);
    if (ret < 0) {
        printk(KERN_ERR "%s:failed to register device\n", LOOPAUDIO_DRIVER_NAME);
        return ret;
    }
    return 0;
}

static void __exit loopaudio_exit(void)
{
    misc_deregister(&loopaudio_device);
    platform_driver_unregister(&loopaudio_driver);
}

module_init(loopaudio_init);
module_exit(loopaudio_exit);

MODULE_AUTHOR("Roku");
MODULE_DESCRIPTION("Loopaudio Kernel Driver");
