/*
 *  mhl_core.c - core mhl_intf kernel driver
 *
 *  Copyright (C) 2012 Broadcom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This device provides a shared mechanism for writing to the mailboxes,
 * semaphores, doorbells etc. that are shared between the ARM and the VideoCore
 * processor
 */

#include <linux/module.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <linux/syscalls.h>
#include <linux/proc_fs.h>
#include <linux/semaphore.h>
#include <asm/sizes.h>
#include <asm/io.h>
#include <mach/platform.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <mach/ipc.h>
#include <linux/ctype.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <mach/fifo.h>

#include "mhl_intf.h"
#include "mhl_intf_regs.h"
#include "mhl_driver.h"

/* macros */
#define MAKEFOURCC(ch0, ch1, ch2, ch3) ((unsigned int)(unsigned char)(ch0) \
      | ((unsigned int)(unsigned char)(ch1) << 8) \
      | ((unsigned int)(unsigned char)(ch2) << 16) \
      | ((unsigned int)(unsigned char)(ch3) << 24 ))

#define MHL_INTF_FOURCC MAKEFOURCC( 'M', 'H', 'L', 'I')     // must match the platform_driver name below!
//access macro
#define MHL_INTF_REGISTER_RW(base, offset) (*(volatile unsigned long *)((base) + (offset)))
#define MHL_INTF_REGISTER_RW_BYTE(base, offset) (*(volatile unsigned char *)((base) + (offset)))
// obviously these macros that refer to state base address can only be used after probe has set this up...
#define MHL_IPC_REG_RW(offset) (*(volatile unsigned long *)((uint32_t)mhl_intf_state.base_address + (offset)))
#define MHL_IPC_REG_RW_BYTE(offset) (*(volatile unsigned char *)((uint32_t)mhl_intf_state.base_address + (offset)))


/* structure decalaration */
static struct 
{
   volatile unsigned char *base_address;
   struct resource *mem_res;
   unsigned int irq;
   struct proc_dir_entry *dump_info;
   struct semaphore work_status; /* 0 = busy processing, 1 = free */
   int initialised;
   char cmd_buf[128];
   IPC_FIFO_T rxfifo;
   struct semaphore rxfifo_num_entries;
   struct semaphore devcap_mutex;        // for syncing read/write to the state devcap info data block.
   MHL_DEVCAP_INFO_T devcap_info;
   int interrupt_count;
   int rx_work_count;
} mhl_intf_state;


typedef struct {
   const char *name;
   uint32_t value;
} lookup_table_t;

static lookup_table_t reg_name_table[] =
{
   {"MHL_IPC_TX_CODE",                    MHL_IPC_TX_CODE_OFFSET           },
   {"MHL_IPC_TX_PARAM",                   MHL_IPC_TX_PARAM_OFFSET          },
   {"MHL_IPC_TX_COUNT",                   MHL_IPC_TX_COUNT_OFFSET          },
   {"MHL_IPC_RXFIFO_WRITEPTR_OFFSET",     MHL_IPC_RXFIFO_WRITEPTR_OFFSET   },
   {"MHL_IPC_RXFIFO_READPTR_OFFSET",      MHL_IPC_RXFIFO_READPTR_OFFSET    },
   {"MHL_IPC_RXFIFO_FIFO_SIZE_OFFSET",    MHL_IPC_RXFIFO_FIFO_SIZE_OFFSET  },
   {"MHL_IPC_RXFIFO_ENTRY_SIZE_OFFSET",   MHL_IPC_RXFIFO_ENTRY_SIZE_OFFSET },
   {"MHL_IPC_RXFIFO_FIFO_DATA_OFFSET",    MHL_IPC_RXFIFO_FIFO_DATA_OFFSET  },
   {NULL, 0}
};

/*
static lookup_table_t ioctl_name_table[] =
{
   {"IOCTL_MHL_GET_INFO",     IOCTL_MHL_GET_INFO_ID   },
   {"IOCTL_MHL_GET_EVENT",    IOCTL_MHL_GET_EVENT_ID  },
   {"IOCTL_MHL_PUT_EVENT",    IOCTL_MHL_PUT_EVENT_ID  },
   {NULL, 0}
};
*/

/* local helper */

static void mhl_intf_vc_arm_lock(void)
{
    MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_VC_ARM_FLAG1_OFFSET ) = 1;
    MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_VC_ARM_TURN_OFFSET ) = 2;

    while(( MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_VC_ARM_FLAG0_OFFSET ) == 1) &&
          ( MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_VC_ARM_TURN_OFFSET ) == 2))
    {
       //_nop();
        schedule_timeout(1);
    }
}

static void mhl_intf_vc_arm_unlock( void )
{
     MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_VC_ARM_FLAG1_OFFSET ) = 0;
}


// blocks waiting to get an event from the fifo.
static int mhl_rx_event_fifo_get(MHL_IPC_EVENT_T *event)
{
   int status;

   while ( 1 )
   {
      /* is there at least one entry in the FIFO?
       * note is safe to check the number of entries of a FIFO 
       * without a lock.
       */
      if ( 0 != ipc_fifo_num_entries(&mhl_intf_state.rxfifo) )
      {
         mhl_intf_vc_arm_lock();
         status = ipc_fifo_read(&mhl_intf_state.rxfifo, event);
         mhl_intf_vc_arm_unlock();
         status = !status; // 0 means success.
         if ( 0 == status )
         {
            // we read one entry, break out of while()
            break;
         }
      }

      // Wait for Interrupt routine to wake us
      status = down_interruptible(&mhl_intf_state.rxfifo_num_entries);
      mhl_intf_state.rx_work_count++;
      if ( 0 != status )
      {
         // down was interrupted (by CTRL-C or whatever)
         break;
      }
   }

   return status;
}


static void copy_devcap_from_ipc(void)
{
   // copy the devcap info to local state
   unsigned int i=0;
   down(&mhl_intf_state.devcap_mutex);    // for the state devcap structure.
   mhl_intf_vc_arm_lock();                // for the ipc regs
   mhl_intf_state.devcap_info.struct_size_bytes = sizeof(MHL_DEVCAP_INFO_T);
   mhl_intf_state.devcap_info.valid          = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_VALID_OFFSET);
   mhl_intf_state.devcap_info.dev_state      = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4;
   mhl_intf_state.devcap_info.mhl_version    = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4;
   mhl_intf_state.devcap_info.dev_cat        = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.adopter_id_h   = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.adopter_id_l   = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.vid_link_mode  = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.aud_link_mode  = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.video_type     = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.log_dev_map    = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.bandwidth      = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.feature_flag   = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.device_id_h    = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.device_id_l    = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_state.devcap_info.scratchpad_size= MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4;
   mhl_intf_state.devcap_info.int_stat_size  = MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET + i );  i+=4; 
   mhl_intf_vc_arm_unlock();
   up(&mhl_intf_state.devcap_mutex);
}


static int parse_event(MHL_EVENT_T *out, MHL_IPC_EVENT_T *in)
{
   int result = -1;

   if (in && out)
   {
      mhl_intf_print("MHL rx work func received event: %d,%d.\n", in->event_type, in->payload);

      out->event_type = 0;    // initialise with no event

      // now parse the event_code...
      switch (in->event_type)
      {
         case MHL_IPC_RX_EVENT_RCP :
            out->event_type = EVENT_TYPE_KEY;
            out->payload    = in->payload;
            out->error      = EVENT_ERROR_NONE;
            break;
         case MHL_IPC_RX_EVENT_RCPK :
            out->event_type = EVENT_TYPE_KEY_REPLY;
            out->payload    = in->payload;
            out->error      = EVENT_ERROR_NONE;
            break;
         case MHL_IPC_RX_EVENT_RCPE :
            out->event_type = EVENT_TYPE_KEY_REPLY;
            out->payload    = in->payload;
            out->error      = EVENT_ERROR_INEFFECTIVE_KEY;
            break;
         case MHL_IPC_RX_EVENT_DEVCAP :
            copy_devcap_from_ipc();
            out->event_type = EVENT_TYPE_HPD_ACTIVE;
            out->payload    = 0;
            out->error      = 0;
            break;
         case MHL_IPC_RX_EVENT_CONNECT :
            out->event_type = EVENT_TYPE_HPD_ACTIVE;
            out->payload    = 0;
            out->error      = 0;
            break;
         case MHL_IPC_RX_EVENT_DISCONNECT :
            out->event_type = EVENT_TYPE_HPD_INACTIVE;
            out->payload    = 0;
            out->error      = 0;
            mhl_intf_state.devcap_info.valid = 0;  // reset the devcap info to invalid.
            break;
         case MHL_IPC_RX_EVENT_CONTENT_ON :
            out->event_type = EVENT_TYPE_CONTENT_ON;
            out->payload    = 0;
            out->error      = 0;
            break;
         case MHL_IPC_RX_EVENT_CONTENT_OFF :
            out->event_type = EVENT_TYPE_CONTENT_OFF;
            out->payload    = 0;
            out->error      = 0;
            break;
         default:    
            mhl_intf_print("Unknown event!\n");
            // ignore it.
      }
      
      if (out->event_type)
      {
         // found a known event
         result = 0; // success!
      }
   }

   return result;
}


// Search the input string for alpha-numeric chars.
// Upon return, token_start = pointer to any alpha-num string found.
//              next_char = pointer to the char after the end of the token.
// Returns non-zero if found a token.
static char * str_find_token(char * input, char **token_start, char **next_char)
{
   char * start = NULL;
   char * next = NULL;
   if (input && input[0])
   {
      start = input;
      while (*start && isspace(*start))
         start++;
      // keyp_p now points to start of token. 
      // Next, progress through the token converting to upper case...
      next = start;
      while (isalnum(*next) || (*next == '_'))
      {
         *next = toupper(*next);
         next++;
      }
      // Now we're at the end of the token.

      // did we find any alpha-num chars?
      if (start==next) 
      {
         // nope.
         start=next=0;
      }
      else if (*next)
      {
         // there's still more string to be parsed. Terminate the token string.
         *next = 0;
         next++;
      }
      else
      {
         // end of string, no further parameters.
         next = NULL;
      }
   }
   if (token_start) *token_start = start;
   if (next_char) *next_char = next;
   return start;
}

/* global function */

// return the size of the response str.
int bcm2835_mhl_intf_parse_strcmd(char *response, int maxlen, const char *format, ...)
{
   va_list args;
   int ret = -ENXIO, len;
   char buf[128];
   char *key_p=NULL, *value1_p=NULL, *value2_p=NULL;

   mhl_intf_print("bcm2835_mhl_intf_parse_strcmd -enter\n");
   if (!mhl_intf_state.initialised)
   {
      printk(KERN_ERR"mhl_intf driver not yet initialised\n");
      //ret = -ENXIO;
      goto done;
   }
   // mark the tv_intf state as busy...
   //down(&tv_intf_state.work_status);
   va_start(args, format);
   len = vsnprintf(buf, 127, format, args );
   va_end(args);
   buf[127] = '\0';

   strcpy(mhl_intf_state.cmd_buf, buf );
   // Find beginning of key (strip leading whitespace, convert to uppercase)
   str_find_token(buf, &key_p, &value1_p);
   if (key_p)
   {
      if (0==strcmp(key_p, "IOCTL_MHL_GET_EVENT"))
      {
         MHL_EVENT_T event;
         ret = bcm2835_mhl_ioctl_event_get(&event);
         len = snprintf(response, maxlen,
                  "IOCTL_MHL_GET_EVENT:  ret=%d event_type=%u payload=%u err=%u.\n", 
                  ret, event.event_type, event.payload, event.error);
         response[maxlen-1] = 0;   // enforce termination
         ret = len;
      }
      else if (0==strcmp(key_p, "IOCTL_MHL_GET_INFO"))
      {
         MHL_DEVCAP_INFO_T info;
         ret = bcm2835_mhl_ioctl_devcap_get(&info);
         len = snprintf(response, maxlen,
                        "IOCTL_MHL_GET_INFO:  ret=%d sz=%u valid=%u  0x%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X,%02X.\n", 
                        ret, info.struct_size_bytes, info.valid, 
                        info.dev_state,
                        info.mhl_version,
                        info.dev_cat,
                        info.adopter_id_h,
                        info.adopter_id_l,
                        info.vid_link_mode,
                        info.aud_link_mode,
                        info.video_type,
                        info.log_dev_map,
                        info.bandwidth,
                        info.feature_flag,
                        info.device_id_h,
                        info.device_id_l,
                        info.scratchpad_size,
                        info.int_stat_size     );
         response[maxlen-1] = 0;   // enforce termination
         ret = len;
      }
      else if (0==strcmp(key_p, "IOCTL_MHL_PUT_EVENT"))
      {
         str_find_token(value1_p, &value1_p, &value2_p);
         str_find_token(value2_p, &value2_p, NULL);
         if (value1_p && value2_p)
         {
            MHL_EVENT_T event;
            event.event_type = simple_strtol(value1_p, (char**)NULL, 0);
            event.payload = simple_strtol(value2_p, (char**)NULL, 0);
            ret = bcm2835_mhl_ioctl_event_put(&event);
            len = snprintf(response, maxlen,
                     "IOCTL_MHL_PUT_EVENT:  ret=%d event=0x%02X param=0x%02X.\n", 
                     ret, event.event_type, event.payload );
            response[maxlen-1] = 0;   // enforce termination
            ret = len;
         }
         else
         {
            mhl_intf_print("No parameter found.\n");
         }
      }
      else if (0==strcmp(key_p, "IOCTL_MHL_CONTROL_BEHAVIOUR"))
      {
         str_find_token(value1_p, &value1_p, &value2_p);
         str_find_token(value2_p, &value2_p, NULL);
         if (value1_p && value2_p)
         {
            MHL_BEHAVIOUR_T control;
            control.key = simple_strtol(value1_p, (char**)NULL, 0);
            control.value = simple_strtol(value2_p, (char**)NULL, 0);

            ret = bcm2835_mhl_ioctl_control_behaviour(&control);
            len = snprintf(response, maxlen,
                     "IOCTL_MHL_CONTROL_BEHAVIOUR:  ret=%d key=0x%02X val=0x%02X.\n", 
                     ret, control.key, control.value );
            response[maxlen-1] = 0;   // enforce termination
            ret = len;
         }
         else
         {
            mhl_intf_print("No parameter found.\n");
         }
      }
      else
      {
         mhl_intf_print("Unknown ioctl name.\n");
      }
   }
   else
   {
      mhl_intf_print("No key name found.\n");
   }

   // mark mhl_intf state as free...
   //up(&tv_intf_state.work_status);
   //ret = -ENXIO;
done:
   mhl_intf_print("-exiting\n");
   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_intf_parse_strcmd);


/* static functions */
static int dump_mhl_intf_ipc_block( char *buffer, char **start, off_t offset, 
                                    int buffer_space_bytes, int *eof, void *context )
{
   char *buffer_original;
   int len;
   int ret = 0;
   uint32_t i;
   *eof = 1;

   buffer_original = buffer;

   mhl_intf_print( "params: 0x%08X 0x%08X 0x%08X %d 0x%08X 0x%08X.\n", 
                   (uint32_t)buffer, (uint32_t)start, (uint32_t)offset, buffer_space_bytes, (uint32_t)eof, (uint32_t)context );

   len = snprintf(buffer, buffer_space_bytes, "MHL_Intf IPC block registers:\n");
   //len++;
   ret += len;
   buffer_space_bytes -= len;
   buffer+= len;

   mhl_intf_vc_arm_lock();

   // dump the ipc regs...
   i=0;
   while (reg_name_table[i].name) 
   {
      len = snprintf( buffer, buffer_space_bytes, "%s(0x%08X)=0x%08X\n", 
                      reg_name_table[i].name,
                      (uint32_t)&MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, reg_name_table[i].value),
                      MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, reg_name_table[i].value) );
      ret += len;
      buffer_space_bytes -= len;
      buffer+= len;
      i++;
   }

   // dump fifo data
   len = snprintf( buffer, buffer_space_bytes, "rxfifo r=%lu w=%lu n=%d \n",
                   (*mhl_intf_state.rxfifo.read) % mhl_intf_state.rxfifo.size,
                   (*mhl_intf_state.rxfifo.write) % mhl_intf_state.rxfifo.size,
                   ipc_fifo_num_entries(&mhl_intf_state.rxfifo) );
   ret += len;
   buffer_space_bytes -= len;
   buffer+= len;
   for (i=0; i<mhl_intf_state.rxfifo.size; i++) 
   {
      MHL_IPC_EVENT_T *entry;
      entry = (MHL_IPC_EVENT_T *)ipc_fifo_peek(&mhl_intf_state.rxfifo, i);
      len = snprintf( buffer, buffer_space_bytes, "rxfifo[%02d]=0x%02X,0x%02X\n", i,
                      entry->event_type, entry->payload );
      ret += len;
      buffer_space_bytes -= len;
      buffer+= len;
   }

   // dump DEVCAP regs
   for (i=0; i<MHL_IPC_RX_DEVCAP_LENGTH; i++)
   {
      len = snprintf(buffer, buffer_space_bytes,
                    "devcap[%02d]=0x%02X\n",
                    i, MHL_INTF_REGISTER_RW_BYTE(mhl_intf_state.base_address, MHL_IPC_RX_DEVCAP_OFFSET+(4*i)) );
      ret += len;
      buffer_space_bytes -= len;
      buffer += len;
   }

   mhl_intf_vc_arm_unlock();

   len = snprintf(buffer, buffer_space_bytes, "devcap valid=%d\n", mhl_intf_state.devcap_info.valid);
   ret += len;
   buffer_space_bytes -= len;
   buffer+= len;
   mhl_intf_print( "buffer=0x%08X space=%d.\n", (uint32_t)buffer, buffer_space_bytes);

   len = snprintf(buffer, buffer_space_bytes, "mhl ipc isr=%d wx=%d\n",
                  mhl_intf_state.interrupt_count,
                  mhl_intf_state.rx_work_count );
   ret += len;
   buffer_space_bytes -= len;
   buffer+= len;

   len = snprintf(buffer, buffer_space_bytes, "cmd_buf=%s\n", mhl_intf_state.cmd_buf);
   len++;
   ret += len;
   buffer_space_bytes -= len;
   buffer+= len;

   buffer[buffer_space_bytes-1] = 0;   // enforce termination.

   mhl_intf_print( "dump=<%s>\n", buffer_original);

   return ret;
}


int proc_bcm2835_mhl_intf(struct file *file, const char __user *buffer, unsigned long buffer_len, void *data)
{
   char req_str[200], resp_str[100];
   int ret;

   if( buffer_len > (sizeof(req_str) - 1) ) {
      printk(KERN_ERR"%s:mhl_intf max length=%d\n", __func__, sizeof(req_str));
      return -ENOMEM;
   }
   /* write data to buffer */
   if( copy_from_user( req_str, buffer, buffer_len) )
      return -EFAULT;
   req_str[buffer_len] = '\0';

   printk("mhl_intf: %s\n", req_str);
   ret = bcm2835_mhl_intf_parse_strcmd(resp_str, sizeof(resp_str), req_str);

   if (0 > ret)
      printk(KERN_ERR"response: %s\n", resp_str);

   return buffer_len;
}


int bcm2835_mhl_ioctl_event_get(MHL_EVENT_T *ioctl_event)
{
   int ret = -1;
   MHL_IPC_EVENT_T ipc_event;
   if (ioctl_event)
   {
      ret = mhl_rx_event_fifo_get(&ipc_event);    // blocking call will wait for an event to arrive.

      if (0==ret)
      {
         ret = parse_event(ioctl_event, &ipc_event);
         mhl_intf_print("bcm2835_mhl_ioctl_event_get:  ret=%d event_type=%u payload=%u err=%u.\n", 
                         ret, ioctl_event->event_type, ioctl_event->payload, ioctl_event->error);
      }
      else
      {
         mhl_intf_print("bcm2835_mhl_ioctl_event_get: wait for event got cancelled. err=%d.\n", ret);
      }
   }
   else
   {
      mhl_intf_print("bcm2835_mhl_ioctl_event_get: bad argument: 0x%08lX.\n", (unsigned long int)ioctl_event);
   }

   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_ioctl_event_get);


int bcm2835_mhl_ioctl_get_num_events(uint32_t *num_events)
{
   int ret = -1;

   if (num_events)
   {
      mhl_intf_vc_arm_lock();
      ret = ipc_fifo_num_entries(&mhl_intf_state.rxfifo);
      mhl_intf_vc_arm_unlock();

      mhl_intf_print("bcm2835_mhl_ioctl_get_num_events:  num_events=%d.\n", ret);
      if (ret>=0)
      {
         // ok
         *num_events = (uint32_t)ret;
         ret = 0;
      }
   }
   else
   {
      mhl_intf_print("bcm2835_mhl_ioctl_get_num_events: bad argument: 0x%08lX.\n", (unsigned long int)num_events);
   }

   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_ioctl_get_num_events);


static void mhl_ipc_tx_write(uint8_t code, uint8_t param)
{
//         unsigned long int time = sched_clock();
   while (MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_TX_COUNT_OFFSET)>0) {}   // wait for vc to receive the last message before we clobber it with a new message.
//         time = sched_clock() - time;
   mhl_intf_vc_arm_lock();
   MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_TX_CODE_OFFSET)  = code;
   MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_TX_PARAM_OFFSET) = param;
   MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_TX_COUNT_OFFSET) += 1;
   mhl_intf_vc_arm_unlock();
   ipc_notify_vc_event(mhl_intf_state.irq);
}


int bcm2835_mhl_ioctl_event_put(MHL_EVENT_T *event)
{
   int ret = 0;

   if (event)
   {
      uint8_t event_code = 0, param = 0;

      switch (event->event_type)
      {
         case EVENT_TYPE_KEY :
            event_code = MHL_IPC_TX_EVENT_RCP;
            param = event->payload;
            break;
         case EVENT_TYPE_KEY_REPLY :
            if (event->error)
            {
               event_code = MHL_IPC_TX_EVENT_RCPE;
               param = 1;  // ineffective key.
            }
            else
            {
               event_code = MHL_IPC_TX_EVENT_RCPK;
               param = event->payload;
            }
            break;
         case EVENT_TYPE_CONTENT_ON :
            event_code = MHL_IPC_TX_EVENT_CONTENT_ON;
            param = event->payload;
            break;
         case EVENT_TYPE_CONTENT_OFF :
            event_code = MHL_IPC_TX_EVENT_CONTENT_OFF;
            param = event->payload;
            break;
         default :
            mhl_intf_print("bcm2835_mhl_ioctl_event_put:  unknown event_type=%u payload=%u err=%u.\n", 
                           event->event_type, event->payload, event->error);
      }

      if (event_code)   // if got a valid event then write it to IPC...
      {
         mhl_intf_print("bcm2835_mhl_ioctl_event_put:  event_type=%u payload=%u err=%u.\n", 
                        event->event_type, event->payload, event->error);
         mhl_ipc_tx_write(event_code, param);
      }
   }
   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_ioctl_event_put);


int bcm2835_mhl_ioctl_control_behaviour(MHL_BEHAVIOUR_T *ctl)
{
   int ret = -1;

   if (ctl)
   {
      switch (ctl->key)
      {
         case MHL_BEHAVIOUR_RCP :
            mhl_intf_print("bcm2835_mhl_ioctl_control_behaviour: sending MHL_IPC_TX_RCP_ENABLE.\n");
            mhl_ipc_tx_write(MHL_IPC_TX_EVENT_RCP_ENABLE, ctl->value);
            ret = 0;
            break;
         default :
            mhl_intf_print("bcm2835_mhl_ioctl_control_behaviour: unkown control key=0x%02X val=0x%02X.\n",
                           ctl->value, ctl->key);
            break;
      }
   }

   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_ioctl_control_behaviour);


int bcm2835_mhl_ioctl_devcap_get(MHL_DEVCAP_INFO_T *info)
{
   int ret = -1;

   if (info)
   {
      if (mhl_intf_state.initialised)
      {
         if (!mhl_intf_state.devcap_info.valid)
         {
            // haven't got a valid devcap so get latest one.
            copy_devcap_from_ipc();
         }

         if (mhl_intf_state.devcap_info.valid)
         {
            down(&mhl_intf_state.devcap_mutex);    // for the state devcap structure.
            memcpy(info, &mhl_intf_state.devcap_info, mhl_intf_state.devcap_info.struct_size_bytes);
            up(&mhl_intf_state.devcap_mutex);      // for the state devcap structure.
            ret = 0;
         }
         else
         {
            mhl_intf_print("bcm2835_mhl_ioctl_devcap_get: devcap info not valid.\n" );
            ret = -2;
            // don't have any info so see if we can get some...
            mhl_ipc_tx_write(MHL_IPC_TX_EVENT_REQ_INFO, 0);
         }
      }
      else
      {
         mhl_intf_print("bcm2835_mhl_ioctl_devcap_get: mhl_intf_state not initialised!\n" );
      }
   }
   else
   {
      mhl_intf_print("bcm2835_mhl_ioctl_devcap_get: bad argument: 0x%08lX.\n", (long unsigned int)info);
   }

   return ret;
}
EXPORT_SYMBOL(bcm2835_mhl_ioctl_devcap_get);


/*
 * Interrupt Service Routine. Called when the VC rings
 * the doorbell interrupt and a MHL interface action
 * has been indicated.
 */
static irqreturn_t bcm2835_mhl_intf_isr(int irq, void *dev_id)
{
   mhl_intf_print("bcm2835_mhl_intf_isr(): %d entries.\n", ipc_fifo_num_entries(&mhl_intf_state.rxfifo) );

   mhl_intf_state.interrupt_count++;

   up(&mhl_intf_state.rxfifo_num_entries);

   return IRQ_HANDLED;
}


static int __devexit bcm2835_mhl_intf_remove(struct platform_device *pdev)
{
   if( mhl_intf_state.initialised )
   {
      mhl_intf_driver_exit();
      remove_proc_entry("bcm2835_mhlif", NULL);
      free_irq(mhl_intf_state.irq, NULL);
      release_region(mhl_intf_state.mem_res->start, resource_size(mhl_intf_state.mem_res));
      mhl_intf_state.initialised = 0;
   }

   return 0;
}


static int __devinit bcm2835_mhl_intf_probe(struct platform_device *pdev)
{
   struct resource *res;
   int ret = -ENOENT;
   int num_entries;

   mhl_intf_print("probe=%p\n", pdev);

   /* initialize the mhl_intf struct */
   memset(&mhl_intf_state, 0, sizeof(mhl_intf_state));

   /* get the mhl_intf memory region */
   res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
   if (res == NULL) {
      dev_err(&pdev->dev, "failed to get memory region resource\n");
      ret = -ENODEV;
      goto err_platform_res;
   }

   mhl_intf_print("mhl_intf memory region start=%p end=%p\n", (void *)res->start,
         (void *)res->end);

   mhl_intf_state.base_address = (void __iomem *)(res->start);

   /* Request memory region */
   mhl_intf_state.mem_res = request_mem_region_exclusive(res->start, resource_size(res), "bcm2835mhl_intf_regs");
   if (mhl_intf_state.mem_res == NULL) {
      dev_err(&pdev->dev, "Unable to request mhl_intf memory region\n");
      ret = -ENOMEM;
      goto err_mem_region;
   }

   /* find our ipc id */
   mhl_intf_state.irq = platform_get_irq(pdev, 0);

   if( mhl_intf_state.irq < 0 ) {
      dev_err(&pdev->dev, "failed to get irq for mhl_intf fourcc=0x%x\n", MHL_INTF_FOURCC);
      ret = -ENODEV;
      goto err_irq;
   }

   mhl_intf_print("mhl_intf irqno=%d\n", mhl_intf_state.irq);

   mhl_intf_print("mhl_intf base_address   =0x%08X\n", (uint32_t)mhl_intf_state.base_address);
   mhl_intf_print("mhl_intf &rxfifobasebus =0x%08X\n", (uint32_t)&MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_DATA_OFFSET));
   mhl_intf_print("mhl_intf rxfifobasebus  =0x%08lX\n", MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_DATA_OFFSET));
   mhl_intf_print("mhl_intf rxfifobasevirt =0x%08X\n", (uint32_t)ipc_bus_to_virt(MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_DATA_OFFSET)));
   mhl_intf_print("mhl_intf rxfifowrite    =0x%08lX\n", MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_WRITEPTR_OFFSET));
   mhl_intf_print("mhl_intf rxfiforead     =0x%08lX\n", MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_READPTR_OFFSET));
   mhl_intf_print("mhl_intf rxfifosize     =0x%08lX\n", MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_SIZE_OFFSET));
   mhl_intf_print("mhl_intf rxfifoentrysz  =0x%08lX\n", MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_ENTRY_SIZE_OFFSET));

   // set up the local rxfifo struct to point to the IPC regs...
   ipc_fifo_setup_no_reset(&mhl_intf_state.rxfifo,
                      &MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_WRITEPTR_OFFSET),
                      &MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_READPTR_OFFSET),
                      //ipc_bus_to_virt(&MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_DATA_OFFSET)),
                      &MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_DATA_OFFSET),
                      MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_FIFO_SIZE_OFFSET),
                      MHL_INTF_REGISTER_RW(mhl_intf_state.base_address, MHL_IPC_RXFIFO_ENTRY_SIZE_OFFSET));

   // Videocore might have seen events happen already before this driver has 
   // started, so need to account for this when initialising the counting semaphore...
   num_entries = ipc_fifo_num_entries(&mhl_intf_state.rxfifo);

   // initialize the semaphores
   sema_init( &mhl_intf_state.rxfifo_num_entries, num_entries);
   sema_init( &mhl_intf_state.devcap_mutex, 1);

   /* register irq for mhl_intf service */
   ret = request_irq( mhl_intf_state.irq, bcm2835_mhl_intf_isr,
         IRQF_DISABLED, "bcm2835 mhl_intf interrupt", NULL);
   if (ret < 0) {
      dev_err(&pdev->dev, "failed to install mhl_intf irq handler(%d)\n",
               mhl_intf_state.irq);
      ret = -ENOENT;
      goto err_irq_handler;
   }

   /* create a proc entry */
   mhl_intf_state.dump_info = create_proc_entry( "bcm2835_mhlif",
                                                 (S_IWUSR | S_IWGRP | S_IRUSR | S_IRGRP ),
                                                 NULL );

   if( !mhl_intf_state.dump_info ) {
      dev_err(&pdev->dev, "failed to create bcm2835_mhl_intf proc entry\n");
      ret = -ENOENT;
      goto err_proc_entry;
   }
   else {
      mhl_intf_state.dump_info->write_proc = proc_bcm2835_mhl_intf;
      mhl_intf_state.dump_info->read_proc  = dump_mhl_intf_ipc_block;
   }

   /* register the driver */
   ret = mhl_intf_driver_init();
   if( ret < 0 ) {
      dev_err(&pdev->dev, "failed to register the driver\n");
      ret = -ENOENT;
      goto err_driver;
   }

   mhl_intf_state.initialised = 1;
   return 0;

err_driver:
   /* remove proc entry */
   remove_proc_entry("bcm2835_mhlif", NULL);
err_proc_entry:
   /* free irq */
   free_irq(mhl_intf_state.irq, NULL);
err_irq_handler:
err_irq:
   /* release region */
   release_region(mhl_intf_state.mem_res->start, resource_size(mhl_intf_state.mem_res));
err_mem_region:
err_platform_res:
   return ret;
}


static struct platform_driver bcm2835_mhl_intf_driver = {
   .probe   = bcm2835_mhl_intf_probe,
   .remove  = __devexit_p(bcm2835_mhl_intf_remove),
   .driver  = 
   {
      .name = "bcm2835_MHLI",    // must be of the form "bcm2835_"+fourcc from above.
      .owner = THIS_MODULE,
   },
};

static char __initdata banner[] = "VC4 MHL_Intf Driver,(C) 2012 Broadcom Corp\n";

static int __init bcm2835_mhl_intf_init( void )
{
   //volatile int wait_here = 1;
   //while (wait_here) {};
   printk(KERN_ERR"%s", banner);
   return platform_driver_register(&bcm2835_mhl_intf_driver);
}

static void __exit bcm2835_mhl_intf_exit( void )
{
   printk(KERN_ERR "bcm2835_mhl_intf_exit.\n");
   platform_driver_unregister(&bcm2835_mhl_intf_driver);
}

module_init(bcm2835_mhl_intf_init);
module_exit(bcm2835_mhl_intf_exit);

MODULE_AUTHOR("Broadcom");
MODULE_DESCRIPTION("MHL_Intf Kernel Driver");

/* ************************************ The End ***************************************** */
