/*
 *  mhl_intf_driver.c - mhl_intf driver interface
 *
 *  Copyright (C) 2012 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This device provides a shared mechanism for writing to the mailboxes,
 * semaphores, doorbells etc. that are shared between the ARM and the VideoCore
 * processor
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>

#include "mhl_intf.h"
#include "mhl_driver.h"

#define MHL_INTF_DRIVER_NAME  "mhl_intf"

struct mhl_intf_buffers {
   char cmd[MHL_INTF_CMD_SIZE];
   char resp[MHL_INTF_RESP_SIZE];
   int  resp_len;
   struct semaphore buf_sem;
};


/*driver functions  */
ssize_t mhl_intf_read(struct file *filp, char __user *buf, size_t count,
      loff_t *f_pos)
{
   struct mhl_intf_buffers *g_buf;
   int ret = 0;
   int len;

   mhl_intf_print("-IN buf=%p count=%d\n", buf, count);
   /* get the buffer */
   g_buf = (struct mhl_intf_buffers *) filp->private_data;

   ret = down_interruptible(&g_buf->buf_sem);
   if( ret < 0 ) {
      printk(KERN_ERR"[%s] error in getting buffer semaphore\n", __func__ );
      goto err_lock;
   }

   /* check the length */
   len = g_buf->resp_len;

   if( len > count )
      len = count;

   ret = copy_to_user(buf, g_buf->resp, len);
   if( ret < 0 ) {
      printk(KERN_ERR"[%s] error in copying date to user buffer\n", __func__);
      goto done;
   }

   mhl_intf_print("mhl_intf resp=%s\n", g_buf->resp);
   ret = len;

done:
   up(&g_buf->buf_sem);

err_lock:
   mhl_intf_print("-OUT ret=%d\n", ret);
   return ret;
}


ssize_t mhl_intf_write(struct file *filp, const char __user *buf, size_t count,
      loff_t *f_pos)
{
   struct mhl_intf_buffers *g_buf;
   int ret = 0;

   mhl_intf_print("-IN buf=%p count=%d\n", buf, count);

   /* check the length */
   if( count > (MHL_INTF_CMD_SIZE - 1) ) {
      printk(KERN_ERR"[%s] max cmd size=%d\n", __func__, MHL_INTF_CMD_SIZE);
      ret = -ENOMEM;
      goto err_len;
   }

   /* get the buffer */
   g_buf = (struct mhl_intf_buffers *) filp->private_data;

   ret = down_interruptible(&g_buf->buf_sem);
   if( ret < 0 ) {
      printk(KERN_ERR"[%s] error in getting buffer semaphore\n", __func__ );
      goto err_lock;
   }

   /* copy from user */
   if( copy_from_user(g_buf->cmd, buf, count) ) {
      ret = -EFAULT;
      printk(KERN_ERR"[%s] copy from user failed user_buf=%p\n",
            __func__, buf);
      goto done;
   }
   g_buf->cmd[count] = '\0';
   mhl_intf_print("cmd=%s\n", g_buf->cmd);

   /* call bcm2835_mhl_intf */
   g_buf->resp_len = bcm2835_mhl_intf_parse_strcmd( g_buf->resp, MHL_INTF_RESP_SIZE, g_buf->cmd);
   ret = count;

done:
   up(&g_buf->buf_sem);
err_lock:
err_len:
   mhl_intf_print("-OUT ret=%d\n", ret);
   return ret;
}


int mhl_intf_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
   int ret = 0;
   unsigned long uncopied;
   u8 *ioctl_cmd_buf;

   mhl_intf_print("mhl_intf_ioctl: cmd = %x, size=%d\n", cmd, _IOC_SIZE(cmd));

   BUG_ON(MHL_INTF_MAX_IOCTL_CMD_SIZE < _IOC_SIZE(cmd));

   ioctl_cmd_buf = (u8 *)kmalloc(MHL_INTF_MAX_IOCTL_CMD_SIZE, GFP_KERNEL);
   if (!ioctl_cmd_buf)
      return -ENOMEM;
   mhl_intf_print("mhl_intf_ioctl: Got cmd buffer.\n");

   if (0 != _IOC_SIZE(cmd))
   {
      uncopied = copy_from_user(ioctl_cmd_buf, (void *)arg, _IOC_SIZE(cmd));
      if (uncopied != 0) {
         kfree(ioctl_cmd_buf);
         return -EFAULT;
      }
      mhl_intf_print("mhl_intf_ioctl: Copied args from user space. cmd=0x%02X argp=0x%08X argval=0x%08X.\n", cmd, (uint32_t)ioctl_cmd_buf, *(uint32_t *)ioctl_cmd_buf);

      switch (cmd) 
      {
         case IOCTL_MHL_GET_INFO :
            ret = bcm2835_mhl_ioctl_devcap_get((MHL_DEVCAP_INFO_T *)ioctl_cmd_buf);
            uncopied = copy_to_user((void *)arg, ioctl_cmd_buf, _IOC_SIZE(cmd));
            if (uncopied != 0)
               ret = -EFAULT;
            mhl_intf_print("mhl_intf_ioctl: Copied results to user space.\n");
            break;
         case IOCTL_MHL_GET_EVENT :
            ret = bcm2835_mhl_ioctl_event_get((MHL_EVENT_T *)ioctl_cmd_buf);
            uncopied = copy_to_user((void *)arg, ioctl_cmd_buf, _IOC_SIZE(cmd));
            if (uncopied != 0)
               ret = -EFAULT;
            mhl_intf_print("mhl_intf_ioctl: Copied results to user space.\n");
            break;
         case IOCTL_MHL_PUT_EVENT :
            ret = bcm2835_mhl_ioctl_event_put((MHL_EVENT_T *)ioctl_cmd_buf);
            break;
         case IOCTL_MHL_CONTROL_BEHAVIOUR :
            ret = bcm2835_mhl_ioctl_control_behaviour((MHL_BEHAVIOUR_T *)ioctl_cmd_buf);
            break;
         case IOCTL_MHL_GET_NUM_EVENTS :
            ret = bcm2835_mhl_ioctl_get_num_events((uint32_t *)ioctl_cmd_buf);
            uncopied = copy_to_user((void *)arg, ioctl_cmd_buf, _IOC_SIZE(cmd));
            if (uncopied != 0)
               ret = -EFAULT;
            mhl_intf_print("mhl_intf_ioctl: Copied results to user space.\n");
            break;
         default:
            printk(KERN_ERR "Wrong IOCTL cmd (%x). Expected %x, %x, %x, or %x.\n",
                              cmd,
                              IOCTL_MHL_GET_INFO,
                              IOCTL_MHL_GET_EVENT,
                              IOCTL_MHL_PUT_EVENT,
                              IOCTL_MHL_CONTROL_BEHAVIOUR);
            ret = -EFAULT;
            break;
      }
   }

   kfree(ioctl_cmd_buf);
   return ret;
}


int mhl_intf_open (struct inode *inode, struct file *filp)
{
   struct mhl_intf_buffers *g_buf;
   int ret = 0;

   mhl_intf_print("-IN\n");
   g_buf = kmalloc(sizeof(struct mhl_intf_buffers), GFP_KERNEL);
   if( !g_buf ) {
      printk(KERN_ERR"[%s]:failed to allocate mhl_intf buffers\n", __func__);
      ret = -ENOMEM;
   }
   else {
      filp->private_data = g_buf;
      g_buf->resp_len = 0;
      sema_init(&g_buf->buf_sem, 1);
   }
   mhl_intf_print("-OUT ret=%d\n", ret);
   return ret;
}


int mhl_intf_release(struct inode *inode, struct file *filp)
{
   int ret;
   // int scratch;
   struct mhl_intf_buffers *g_buf;

   mhl_intf_print("-IN");
   //bcm2835_tv_ioctl_fin(&scratch);
   g_buf = (struct mhl_intf_buffers *) filp->private_data;
   ret = down_interruptible(&g_buf->buf_sem );
   if( ret < 0 ) {
      printk(KERN_ERR"[%s]:failed to get buffer semaphore\n", __func__);
      goto done;
   }
   kfree(g_buf);
   ret = 0;
done:
   mhl_intf_print("-OUT ret=%d\n", ret);
   return ret;
}


struct file_operations mhl_intf_fops = {
   .owner   =  THIS_MODULE,
   .read    =  mhl_intf_read,
   .write   =  mhl_intf_write,
   .ioctl   =  mhl_intf_ioctl,
   .open    =  mhl_intf_open,
   .release =  mhl_intf_release,
};

struct miscdevice mhl_intf_dev = {
   .minor   =  MISC_DYNAMIC_MINOR,
   .name    =  MHL_INTF_DRIVER_NAME,
   .fops    =  &mhl_intf_fops
};


int __devinit mhl_intf_driver_init(void)
{
   int ret;
   printk(KERN_ERR "mhl_intf_driver_init.\n");

   ret = misc_register(&mhl_intf_dev);
   if (ret < 0) {
      printk(KERN_ERR "%s:failed to register device\n",
             MHL_INTF_DRIVER_NAME);
   }

   return ret;
}

void __devexit mhl_intf_driver_exit(void)
{
   printk(KERN_ERR "mhl_intf_driver_exit.\n");
   misc_deregister(&mhl_intf_dev);
}
