/*
 *  mhl_intf.h - mhl_intf driver interface
 *
 *  Copyright (C) 2012 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This device provides a shared mechanism for writing to the mailboxes,
 * semaphores, doorbells etc. that are shared between the ARM and the VideoCore
 * processor
 */

#ifndef _MHL_INTF_H
#define _MHL_INTF_H

//#define MHL_INTF_DEBUG

#ifdef MHL_INTF_DEBUG

#ifdef __KERNEL__
#define refmt(fmt) "[%s]: " fmt, __func__
#define mhl_intf_print(fmt, ...) \
	printk(KERN_ERR refmt(fmt), ##__VA_ARGS__)
#else    // __KERNEL__
#define refmt(fmt) "[%s]: " fmt, __func__
#define mhl_intf_print(fmt, ...) \
    printf(refmt(fmt), ##__VA_ARGS__)
#endif   // __KERNEL__

#else    // MHL_INTF_DEBUG
#define mhl_intf_print(fmt, ...)
#endif   // MHL_INTF_DEBUG

#define MHL_INTF_CMD_SIZE	512
#define MHL_INTF_RESP_SIZE	512
#define MHL_INTF_MAX_IOCTL_CMD_SIZE (18*4)


//------------ from the MHL interface spec from Roku....

typedef unsigned char byte;

typedef enum
{
   IOCTL_MHL_GET_INFO_ID            = 0x1,
   IOCTL_MHL_GET_EVENT_ID           = 0x2,
   IOCTL_MHL_PUT_EVENT_ID           = 0x3,
   IOCTL_MHL_CONTROL_BEHAVIOUR_ID   = 0x4,
   IOCTL_MHL_GET_NUM_EVENTS_ID      = 0x5,
} IOCTL_MHL_T;


//The IOCTL_MHL_GET_INFO command will fill an instance of this structure:

// From MHL 1.1 Spec. Table 7-21: Device Capability Registers
typedef struct
{
   // The size of this structure:
   unsigned short struct_size_bytes;

   // A flag to indicate if the contents are current and valid.
   byte valid;

   // The MHL dev cap values:
   byte dev_state;
   byte mhl_version;
   byte dev_cat;
   byte adopter_id_h;
   byte adopter_id_l;
   byte vid_link_mode;
   byte aud_link_mode;
   byte video_type;
   byte log_dev_map;
   byte bandwidth;
   byte feature_flag;
   byte device_id_h;
   byte device_id_l;
   byte scratchpad_size;
   byte int_stat_size;
} MHL_DEVCAP_INFO_T;


//To handle incoming Remote Control Protocol (RCP) messages (and their success or failure
//replies) and Remote Action Protocol (RAP) messages, the IOCTL_MHL_GET_EVENT command will fill
//an instance of this structure:

typedef enum
{
   // The KEY events are used for RCP messages
   EVENT_TYPE_KEY = 0x10,
   EVENT_TYPE_KEY_REPLY = 0x11,

   // The CONTENT events are used for RAP messages
   EVENT_TYPE_CONTENT_ON = 0x20,
   EVENT_TYPE_CONTENT_OFF = 0x21,

   // The HPD events are used for HPD MSC messages
   EVENT_TYPE_HPD_ACTIVE = 0x30,
   EVENT_TYPE_HPD_INACTIVE = 0x31,

} EVENT_TYPE_T;


typedef enum
{
   EVENT_ERROR_NONE = 0x00, // No error
   EVENT_ERROR_INEFFECTIVE_KEY = 0x01,
} EVENT_ERROR_T;


typedef struct
{
   byte event_type;
   // This is the message payload value (if any) e.g. the key value for
   // an RCP key message. The other messages dont contain payloads
   byte payload;
   // This is an error value for replies to key events.
   byte error;
} MHL_EVENT_T;


typedef enum
{
   MHL_BEHAVIOUR_RCP = 0x01,
} MHL_BEHAVIOUR_KEY_T;


typedef struct
{
   byte key;      // MHL_BEHAVIOUR_KEY_T
   byte value;
} MHL_BEHAVIOUR_T;


//The IOCTL events should be defined like this:
#define IOCTL_MHL_GET_INFO          _IOR('S', IOCTL_MHL_GET_INFO_ID,  MHL_DEVCAP_INFO_T )
#define IOCTL_MHL_GET_EVENT         _IOR('S', IOCTL_MHL_GET_EVENT_ID, MHL_EVENT_T)
#define IOCTL_MHL_PUT_EVENT         _IOW('S', IOCTL_MHL_PUT_EVENT_ID, MHL_EVENT_T)
#define IOCTL_MHL_CONTROL_BEHAVIOUR _IOW('S', IOCTL_MHL_CONTROL_BEHAVIOUR_ID, MHL_BEHAVIOUR_T)
#define IOCTL_MHL_GET_NUM_EVENTS    _IOR('S', IOCTL_MHL_GET_NUM_EVENTS_ID, uint32_t)


//---------------------------------------------------------------
// These functions are shared out to the driver to implement ioctl calls...
int bcm2835_mhl_ioctl_event_get(MHL_EVENT_T *ctl);
int bcm2835_mhl_ioctl_event_put(MHL_EVENT_T *ctl);
int bcm2835_mhl_ioctl_devcap_get(MHL_DEVCAP_INFO_T *info);
int bcm2835_mhl_ioctl_control_behaviour(MHL_BEHAVIOUR_T *ctl);
int bcm2835_mhl_ioctl_get_num_events(uint32_t *num_events);

// ------------------------------------------------------------------
// Shared out to the driver to implement /proc write argument parsing...
int bcm2835_mhl_intf_parse_strcmd(char *response, int maxlen, const char *format, ...);

#endif   // _MHL_INTF_H

