/*****************************************************************************
* Copyright 2012 Broadcom Corporation.  All rights reserved.
*
* Unless you and Broadcom execute a separate written software license
* agreement governing use of this software, this software is licensed to you
* under the terms of the GNU General Public License version 2, available at
* http://www.broadcom.com/licenses/GPLv2.php (the "GPL").
*
* Notwithstanding the above, under no circumstances may you combine this
* software in any way with any other Broadcom software provided under a
* license other than the GPL, without Broadcom's express prior written
* consent.
*****************************************************************************/

/***************************************************************************
*
*  @file    mhl_ipc_regs.h
*
****************************************************************************/

/***************************************************************************
 *
 * The MHL Interface provides the mechanism for the ARM processor to monitor
 * the status of the MHL link, and transmit and receive button presses.
 *
 * Control and status is passed by virtual registers in shared memory.
 *
 * Messages are sent by first claiming the MHL IPC mutex, then writing the 
 * desired values to the registers, releasing the mutex, then the ARM/VC 
 * doorbell interrupt is issued.
 *
 * Upon receiving the doorbell interrupt, a message can be received by
 * claiming the MHL IPC mutex, reading the registers, then releasing the mutex.
 *
 * The MHL IPC mutex is implemented by Peterson's Algorithm on the 3 shared 
 * registers FLAG0, FLAG1, and TURN.
 *
 * Most messages are composed of just two elements: a 1-byte event code and 
 * a 1-byte parameter. A Device Capability message is a little different as 
 * it's composed of a 1-byte event code and a 16-byte block of DEVCAP data.
 *
 ***************************************************************************/

#ifndef MHL_IPC_REGS_H
#define MHL_IPC_REGS_H

#if defined( __cplusplus )
extern "C"
{
#endif

typedef enum {
   MHL_IPC_TX_EVENT_RCP           = (1 << 0),   // 0x01
   MHL_IPC_TX_EVENT_RCPK          = (1 << 1),   // 0x02
   MHL_IPC_TX_EVENT_RCPE          = (1 << 2),   // 0x04
   MHL_IPC_TX_EVENT_CONTENT_ON    = (1 << 3),   // 0x08
   MHL_IPC_TX_EVENT_CONTENT_OFF   = (1 << 4),   // 0x10
   MHL_IPC_TX_EVENT_REQ_INFO      = (1 << 5),   // 0x20
   MHL_IPC_TX_EVENT_RCP_ENABLE    = (1 << 6),   // 0x40
} MHL_IPC_TX_EVENT_T;

typedef enum {
   MHL_IPC_RX_EVENT_RCP           = (1 << 0),   // 0x01
   MHL_IPC_RX_EVENT_RCPK          = (1 << 1),   // 0x02
   MHL_IPC_RX_EVENT_RCPE          = (1 << 2),   // 0x04
   MHL_IPC_RX_EVENT_CONNECT       = (1 << 3),   // 0x08
   MHL_IPC_RX_EVENT_DISCONNECT    = (1 << 4),   // 0x10
   MHL_IPC_RX_EVENT_DEVCAP        = (1 << 5),   // 0x20
   MHL_IPC_RX_EVENT_CONTENT_ON    = (1 << 6),   // 0x40
   MHL_IPC_RX_EVENT_CONTENT_OFF   = (1 << 7),   // 0x80
} MHL_IPC_RX_EVENT_T;


typedef struct
{
   uint8_t event_type;

   // This is the message payload value (if any) e.g. the key value for
   // an RCP key message. The other messages dont contain payloads
   uint8_t payload;

   uint16_t padding;
} MHL_IPC_EVENT_T;
int debug[(sizeof(MHL_IPC_EVENT_T)==4) ? 1:-1];


// Registers for Peterson's algorithm for locking other MHL IPC registers.
#define MHL_IPC_VC_ARM_FLAG0_OFFSET       0
#define MHL_IPC_VC_ARM_FLAG1_OFFSET       4
#define MHL_IPC_VC_ARM_TURN_OFFSET        8

// For receiving events from the MHL port to the ARM. These are r+w by both sides...
#define MHL_IPC_RXFIFO_BASE_OFFSET        12
#define MHL_IPC_RXFIFO_WRITEPTR_OFFSET       (MHL_IPC_RXFIFO_BASE_OFFSET +  0)
#define MHL_IPC_RXFIFO_READPTR_OFFSET        (MHL_IPC_RXFIFO_BASE_OFFSET +  4)
#define MHL_IPC_RXFIFO_FIFO_SIZE_OFFSET      (MHL_IPC_RXFIFO_BASE_OFFSET +  8)
#define MHL_IPC_RXFIFO_ENTRY_SIZE_OFFSET     (MHL_IPC_RXFIFO_BASE_OFFSET + 12)
#define MHL_IPC_RXFIFO_FIFO_DATA_OFFSET      (MHL_IPC_RXFIFO_BASE_OFFSET + 16)
         // fifo data then starts here from FIFO_DATA_OFFSET for ENTRY_SIZE*FIFO_SIZE bytes.
#define MHL_IPC_RXFIFO_SIZE 8 // entries.
#define MHL_IPC_RXFIFO_FIFO_GUARD_OFFSET     (MHL_IPC_RXFIFO_FIFO_DATA_OFFSET + sizeof(MHL_IPC_EVENT_T)*MHL_IPC_RXFIFO_SIZE)

#define MHL_IPC_RX_DEVCAP_OFFSET             (MHL_IPC_RXFIFO_FIFO_GUARD_OFFSET + 4) // mhl device capability block. Each word only contains 1 byte.
#define    MHL_IPC_RX_DEVCAP_LENGTH          (16)  // size of devcap block in words
#define MHL_IPC_RX_DEVCAP_VALID_OFFSET        (MHL_IPC_RX_DEVCAP_OFFSET + 4*MHL_IPC_RX_DEVCAP_LENGTH)
#define MHL_IPC_RX_DEVCAP_GUARD_OFFSET        (MHL_IPC_RX_DEVCAP_VALID_OFFSET + 4)

// Transmit registers. These are only written by ARM and read by VC...
#define MHL_IPC_TX_REGS_BASE_OFFSET       (MHL_IPC_RX_DEVCAP_GUARD_OFFSET + 4)
#define MHL_IPC_TX_CODE_OFFSET               (MHL_IPC_TX_REGS_BASE_OFFSET + 0)     // Value of MSC eventcode, eg: RCP, RCPK, RCPK
#define MHL_IPC_TX_PARAM_OFFSET              (MHL_IPC_TX_REGS_BASE_OFFSET + 4)     // Parameter for this event, eg: keycode.
// Sync registers. These are read and written by both sides...
#define MHL_IPC_TX_COUNT_OFFSET              (MHL_IPC_TX_REGS_BASE_OFFSET + 8)     // ARM inc, VC decr. Result is number of events outstanding or missed by VC.


#if defined( __cplusplus )
}
#endif
#endif // MHL_IPC_REGS_H
