/*
 *  otp_core.c - core otp driver
 *
 *  Copyright (C) 2010 Broadcom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This device provides a shared mechanism for writing to the mailboxes,
 * semaphores, doorbells etc. that are shared between the ARM and the VideoCore
 * processor
 */
#include <linux/module.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <linux/syscalls.h>
#include <linux/proc_fs.h>
#include <linux/semaphore.h>
#include <asm/sizes.h>
#include <asm/io.h>
#include <mach/platform.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <mach/ipc.h>

#include "otp.h"
#include "otp_regs.h"
#include "otp_driver.h"

/* macros */
#define MAKEFOURCC(ch0, ch1, ch2, ch3) ((unsigned int)(unsigned char)(ch0) \
		| ((unsigned int)(unsigned char)(ch1) << 8) \
		| ((unsigned int)(unsigned char)(ch2) << 16) \
		| ((unsigned int)(unsigned char)(ch3) << 24 ))

#define OTP_FOURCC MAKEFOURCC( 'O', 'T', 'P', 'P')

/* structure decalaration */
static struct {
	volatile unsigned char *base_address;
	struct resource *mem_res;
	unsigned int irq;
	struct proc_dir_entry *dump_info;
	struct semaphore work_status; /* 0 = busy processing, 1 = free */
	struct semaphore command_complete;
	int initialized;
	int test_mode;
	uint32_t crcLUT[256];
}otp_state;

static uint32_t vc_otp_crc32 (uint8_t* block, uint32_t length);
static void vc_otp_crc32LUTgen ( void );

/* global function */
// returns non-zero if an error happened
int vc_otp_read(uint32_t addr, uint32_t length, char __user *buf)
{
	int ret = 0;
	uint32_t crc32 = 0;

	otp_print("-enter\n");
	if (!otp_state.initialized) {
		printk(KERN_ERR"otp driver not yet initialized\n");
		ret = -ENXIO;
	    goto done;
	}

	/* mark the otp state as busy */
	down(&otp_state.work_status);

	OTP_REGISTER_RW( OTP_ADDRESS_OFFSET ) = addr;

	OTP_REGISTER_RW( OTP_REQUEST_LENGTH_OFFSET ) = length;
	//fire the 'ready bit'
	OTP_REGISTER_RW( OTP_CONTROL_OFFSET ) = (OTP_CONTROL_GO)
	                                        | (OTP_CONTROL_READ )
	                                        | (otp_state.test_mode & OTP_CONTROL_TEST);


	/* ring the doorbell */
	ipc_notify_vc_event(otp_state.irq);

	/* wait for the done bit */
	otp_print("-waiting\n");
	ret = down_interruptible(&otp_state.command_complete);
	if (ret < 0) {
		printk(KERN_ERR"Failed to get the semaphore error code=%d\n", ret);
		goto release_sem;
	}

	while( !(OTP_REGISTER_RW( OTP_CONTROL_OFFSET) & OTP_CONTROL_DONE) );
	otp_print("-done waiting\n");

	crc32 = OTP_REGISTER_RW( OTP_DATA_CRC );
	if( crc32 != vc_otp_crc32((uint8_t*)(otp_state.base_address + OTP_REQUEST_DATA_OFFSET), (length * 4))) {
		printk(KERN_ERR"otp:crc32 for transfer does not match !\n");
		ret = -ENXIO;
		goto release_sem;
	}

	if (OTP_REGISTER_RW(OTP_STATUS_OFFSET)) {
		ret = -ENXIO;
		goto release_sem;
	}

	ret = copy_to_user(buf, (const void *)(otp_state.base_address + OTP_REQUEST_DATA_OFFSET), (length *4));

	if( ret < 0 ) {
		printk(KERN_ERR"[%s] error in copying data to user buffer\n", __func__);
		goto release_sem;
	}

	/* clear the done bit */
	OTP_REGISTER_RW( OTP_CONTROL_OFFSET) = 0x0;//(~OTP_CONTROL_DONE);

release_sem:
	/* mark otp state as free */
	up(&otp_state.work_status);

done:	
	otp_print("-exiting\n");
	return ret;
}
EXPORT_SYMBOL(vc_otp_read);

// returns non zero if an error happened
int vc_otp_write(uint32_t addr, uint32_t length, const char __user *buf)
{
	int ret = 0;
	uint32_t crc32 = 0;

	otp_print("-enter\n");
	if (!otp_state.initialized) {
		printk(KERN_ERR"otp driver not yet initialized\n");
		ret = -ENXIO;
		goto done;
	}

	/* mark the otp state as busy */
	down(&otp_state.work_status);

	OTP_REGISTER_RW( OTP_ADDRESS_OFFSET ) = addr;

	OTP_REGISTER_RW( OTP_REQUEST_LENGTH_OFFSET ) = length;

	if ((length * 4) > OTP_REQUEST_BUFFER_SIZE) {
		printk(KERN_ERR"invalid buffer size\n");
		goto release_sem;
	}

	ret = copy_from_user((void*)(otp_state.base_address + OTP_REQUEST_DATA_OFFSET), buf, length * 4);

	if( ret < 0 ) {
		printk(KERN_ERR"[%s] error in copying data from user buffer\n", __func__);
		goto release_sem;
	}

	// compute crc32
	crc32 = vc_otp_crc32((uint8_t*)(otp_state.base_address + OTP_REQUEST_DATA_OFFSET), length * 4);
	OTP_REGISTER_RW( OTP_DATA_CRC ) = crc32;

	//fire the 'ready bit'
	OTP_REGISTER_RW( OTP_CONTROL_OFFSET ) = ( OTP_CONTROL_GO )
		| (otp_state.test_mode & OTP_CONTROL_TEST);

	/* ring the doorbell */
	ipc_notify_vc_event(otp_state.irq);

	/* wait for the done bit */
	otp_print("-waiting\n");
	ret = down_interruptible(&otp_state.command_complete);
	if (ret < 0) {
		printk(KERN_ERR"Failed to get the semaphore error code=%d\n", ret);
		goto release_sem;
	}

	while( !(OTP_REGISTER_RW( OTP_CONTROL_OFFSET) & OTP_CONTROL_DONE) );
	otp_print("-done waiting\n");

	/* clear the done bit */
	OTP_REGISTER_RW( OTP_CONTROL_OFFSET) = 0x0;//(~OTP_CONTROL_DONE);

	if( OTP_REGISTER_RW( OTP_STATUS_OFFSET )) {
		printk(KERN_ERR"otp: status failure !\n");
		ret = -ENXIO;
		goto release_sem;
	}

release_sem:
	/* mark otp state as free */
	up(&otp_state.work_status);

done:	
	otp_print("-exiting\n");
	return ret;

}
EXPORT_SYMBOL(vc_otp_write);

// returns non-zero if an error happened
int vc_otp_verify_results(uint32_t addr, uint32_t length, char __user *buf)
{
	int ret = 0;

	otp_print("-enter\n");
	if (!otp_state.initialized) {
		printk(KERN_ERR"otp driver not yet initialized\n");
		ret = -ENXIO;
	    goto done;
	}

	if (OTP_REGISTER_RW( OTP_ADDRESS_OFFSET ) != addr)
	{
		printk(KERN_ERR"otp address does not match the current address offset in memory, thus verify_results will not correlate\n");
		ret = -EINVAL;
		goto done;
	}

	/* mark the otp state as busy */
	down(&otp_state.work_status);

	ret = copy_to_user(buf, (const void *)(otp_state.base_address + OTP_REQUEST_VERIFY_RESULTS_OFFSET), (length *4));

	if( ret < 0 ) {
		printk(KERN_ERR"[%s] error in copying data to user buffer\n", __func__);
	}

	/* mark otp state as free */
	up(&otp_state.work_status);

done:	
	otp_print("-exiting\n");
	return ret;
}
EXPORT_SYMBOL(vc_otp_verify_results);

void vc_otp_test(int test)
{
	otp_state.test_mode = test;
}


// computes crc32
static uint32_t vc_otp_crc32 (uint8_t* block, uint32_t length)
{
	uint32_t crc;
	uint32_t i;

	crc = 0xFFFFFFFF;
	for (i = 0; i < length; i++)
		crc = ((crc >> 8) & 0x00FFFFFF) ^ otp_state.crcLUT[(crc ^ *block++) & 0xFF];

	return (crc ^ 0xFFFFFFFF);
}

// pre-compute crc32 polynom table
static void vc_otp_crc32LUTgen ( void )
{
	uint32_t crc, poly;
	int i, j;
	poly = 0xEDB88320L;

	for (i = 0; i < 256; i++)
	{
	  crc = i;
	  for (j = 8; j > 0; j--)
	  {
		if (crc & 1)
			crc = (crc >> 1) ^ poly;
		else
			crc >>= 1;
	}
	  otp_state.crcLUT[i] = crc;
  }
}


static irqreturn_t bcm2708_otp_isr(int irq, void *dev_id)
{
	otp_print("-enter");
	up( &otp_state.command_complete);
	otp_print("-done");

	return IRQ_HANDLED;
}

static int __devexit bcm2708_otp_remove(struct platform_device *pdev)
{
	if( otp_state.initialized ) {
		otp_driver_exit();
		remove_proc_entry("vc_otp", NULL);
		free_irq(otp_state.irq, NULL);
		release_region(otp_state.mem_res->start, resource_size(otp_state.mem_res));
		otp_state.initialized = 0;
	}

	return 0;
}

static int __devinit bcm2708_otp_probe(struct platform_device *pdev)
{
	struct resource *res;
	int ret = -ENOENT;

	otp_print("probe=%p\n", pdev);

	/* initialize the otp struct */
	memset(&otp_state, 0, sizeof(otp_state));

	/* initialize crc32 LUT table */
	vc_otp_crc32LUTgen();

	/* get the otp memory region */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "failed to get memory region resource\n");
		ret = -ENODEV;
		goto err_platform_res;
	}

	otp_print("otp memory region start=%p end=%p\n", (void *)res->start,
			(void *)res->end);

	otp_state.base_address = (void __iomem *)(res->start);

	/* Request memory region */
	otp_state.mem_res = request_mem_region_exclusive(res->start, resource_size(res), "bcm2708otp_regs");
	if (otp_state.mem_res == NULL) {
		dev_err(&pdev->dev, "Unable to request otp memory region\n");
		ret = -ENOMEM;
		goto err_mem_region;

	}

	/* find our ipc id */
	otp_state.irq = platform_get_irq(pdev, 0);

	if( otp_state.irq < 0 ) {
		dev_err(&pdev->dev, "failed to get irq for otp fourcc=0x%x\n", OTP_FOURCC);
		ret = -ENODEV;
		goto err_irq;
	}

	otp_print("otp irqno =%d\n", otp_state.irq);

	/* initialize the semaphores */
	sema_init( &otp_state.work_status, 1);
	sema_init( &otp_state.command_complete, 0);

	/* register irq for otp service */
	ret = request_irq( otp_state.irq, bcm2708_otp_isr,
			IRQF_DISABLED, "bcm2708 otp interrupt", NULL);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to install otp irq handler(%d)\n",
				otp_state.irq);
		ret = -ENOENT;
		goto err_irq_handler;
	}

	/* register the driver */
	ret = otp_driver_init();
	if( ret < 0 ) {
		dev_err(&pdev->dev, "failed to register the driver\n");
		ret = -ENOENT;
		goto err_driver;
	}

	otp_state.initialized = 1;
	return 0;

err_driver:
	/* free irq */
	free_irq(otp_state.irq, NULL);
err_irq_handler:
err_irq:
	/* release region */
	release_region(otp_state.mem_res->start, resource_size(otp_state.mem_res));
err_mem_region:
err_platform_res:	
	return ret;
}

static struct platform_driver bcm2708_otp_driver = {
	.probe		= bcm2708_otp_probe,
	.remove		= __devexit_p(bcm2708_otp_remove),
	.driver		= {
		.name = "bcm2835_OTPP",
		.owner = THIS_MODULE,
	},
};

static char __initdata banner[] = "VC4 Otp Driver,(C) 2010 Broadcom Corp\n";

static int __init bcm2708_otp_init( void )
{
	printk(KERN_ERR"%s", banner);
	return platform_driver_register(&bcm2708_otp_driver);
}

static void __exit bcm2708_otp_exit( void )
{
	platform_driver_unregister(&bcm2708_otp_driver);
}

module_init(bcm2708_otp_init);
module_exit(bcm2708_otp_exit);

MODULE_AUTHOR("Broadcom");
MODULE_DESCRIPTION("Otp Kernel Driver");

/* ************************************ The End ***************************************** */
