#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/completion.h>
#include <mach/platform.h>
#include <asm/sizes.h>
#include <asm/io.h>
#include <mach/bcm2708_smi.h>
#include <linux/slab.h>

#include "bcm2708_smi_regs.h"
#include "bcm2708_dma.h"


#define SMI_DMA_CHANNEL	4 /* smi dma channel */
// FIFO setup defines
#define FIFO_NON_DMA 0
#define FIFO_DMA     1

#define FIFO_READ    0
#define FIFO_WRITE   1


struct bcm2708_smi_info {
	struct platform_device *pdev;
	struct resource *smi_res;
	void __iomem *smi_base;
	void __iomem *dma_base;
	void *dma_mem;
	volatile struct bcm2708_dma_cb *dma_cb;
	struct completion comp;		/* needed for dma */
};

static struct bcm2708_smi_info *smi_info = NULL;

int32_t smi_wait_transfer_complete(void)
{
	uint32_t smics;

	do {
		smics = readl((smi_info->smi_base + SMICS));
	}while( !(smics & (1 << SMICS_DONE)) ); 

	return 0;
}
EXPORT_SYMBOL(smi_wait_transfer_complete);


int32_t smi_wait_direct_complete(void)
{
	uint32_t smidcs;

	do {
		smidcs = readl((smi_info->smi_base + SMIDCS));
	}while( !(smidcs & (1 << SMIDCS_DONE)) ); 

	return 0;
}
EXPORT_SYMBOL(smi_wait_direct_complete);


static int32_t smi_check_transfer_complete( void )
{
	int32_t success = -1; //fail by default
	int timeout = 0;
	int transfer_completed = 0;

	do
	{
		// check the transfer is done
		if ( readl((smi_info->smi_base + SMICS)) & (1 << SMICS_DONE))
		{
			//check the peripheral is no longer active
			if ( !(readl((smi_info->smi_base + SMICS)) & (1 << SMICS_ACTIVE)))
			{
				transfer_completed = 1;
			}
		}
	}
	while ( !transfer_completed && (timeout++ < 100000) );

	// did we time out?
	if (transfer_completed)
	{
		//ensure that the fifo is empty
		if ( !(readl((smi_info->smi_base + SMIFD)) & (0x1F << SMIFD_FCNT) /*fifo level*/ ))
		{
			//check all fifo flags are cleared
			if ( (readl((smi_info->smi_base + SMICS)) >> 24)
					== (( (1<<SMICS_TXE) | (1<<SMICS_TXD) | (1<<SMICS_TXW)) >> 24) )
			{
				//success!
				success = 0;
			}
		}
	}

	if (success)
		printk(KERN_ERR"[%s]: smi fifo error\n", __func__);

	return success;
}


static void smi_setup_fifo(const uint32_t slot,
		const uint32_t address,
		const uint32_t data_size_in_bytes,
		const uint32_t dma,
		const uint32_t write)
{
	uint32_t smics, smil, smia, smidc;

	/* disable SMI */
	smics = 0;
	writel(smics, (smi_info->smi_base + SMICS));

	// we do 8bit transfers so just length
	smil = data_size_in_bytes;
	writel(smil, (smi_info->smi_base + SMIL));

	smia = (address & 0x3F)        // address
		| (slot << SMIA_DEVICE);  // which settings to use.
	writel(smia, (smi_info->smi_base + SMIA));

	// set dma mode if required
	smidc = readl(smi_info->smi_base + SMIDC);
	if( dma ) {
		smidc |= (1 << SMIDC_DMAEN);
		writel(smidc, (smi_info->smi_base + SMIDC));
	} else {
		smidc &= ~(1 << SMIDC_DMAEN);
		writel(smidc, (smi_info->smi_base + SMIDC));
	}

	// once the other registers are set up we can enable the peripheral
	smics = (1 << SMICS_ENABLE);
	writel(smics, (smi_info->smi_base + SMICS));

	// we use SMI_PIXEL_FORMAT_NAND, see nand_common.c
	smics |= (1 << SMICS_PXLDAT);
	writel(smics, (smi_info->smi_base + SMICS));

	smics |= (write << SMICS_WRITE) // set the write flag if needed
		|  (1 << SMICS_CLEARFIFO); // clear the FIFO
	writel(smics, (smi_info->smi_base + SMICS));
	smics |= (1 << SMICS_START);    // start the transfer
	writel(smics, (smi_info->smi_base + SMICS));
}


/* SMI Reads */
int32_t smi_read_direct (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{
	int ret = -1;
	uint32_t smidcs, smida;
	uint32_t read_data;

	bcm2708_smi_print("slot=%d, address=0x%x, size=%d, blocking=%d, data=%p\n",
			slot, address, data_size_in_bytes, blocking, data);

	smida = (address & 0x3F) | (slot << SMIDA_DEVICE);
	writel(smida, (smi_info->smi_base + SMIDA));

	//start the transfer
	smidcs = (1 << SMIDCS_ENABLE);
	writel(smidcs, (smi_info->smi_base + SMIDCS));

	smidcs |= (1 << SMIDCS_START);
	writel(smidcs, (smi_info->smi_base + SMIDCS));

	/* wait until data is transferred */
	smi_wait_direct_complete();
	//TODO
	read_data = readl((smi_info->smi_base + SMIDD));

	switch(data_size_in_bytes)
	{
		case 1:
			*((uint8_t *)data) = read_data;
			break;
		case 2:
			*((uint16_t *)data) = read_data;
			break;
		case 4:
			*((uint32_t *)data) = read_data;
			break;
		default:
			printk(KERN_ERR"%s: incorrect data size=%d\n", __func__,
					data_size_in_bytes);
			ret = -EINVAL;
			goto exit_err;
	}
	
	smidcs = 1 << SMIDCS_DONE;
	writel(smidcs, (smi_info->smi_base + SMIDCS));
	return 0;
exit_err:
	return ret;

}
EXPORT_SYMBOL(smi_read_direct);


int32_t smi_read_fifo (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{
	int32_t success = -1;
	uint32_t timeout = 0;

	uint32_t dummy;
	uint8_t * ptr = (uint8_t *)data;
	uint32_t data_size = data_size_in_bytes;

	printk(KERN_ERR"%s called\n", __func__);
	smi_setup_fifo(slot, address, data_size_in_bytes, FIFO_NON_DMA, FIFO_READ);

	// do the read
	while(data_size) 
	{
		timeout = 0;

		 //wait until there is data in the read FIFO
		while( !( readl((smi_info->smi_base + SMICS)) & (1 << SMICS_RXR)) &&
				(timeout++ < 10000) );

		if( timeout > 10000) {
			printk(KERN_ERR "[%s]: timeout while waiting for data\n", __func__);
			success = -1;
			goto exit_err;
		}

		if (data_size / 4) {
			*((uint32_t *)ptr) = readl( (smi_info->smi_base + SMID) );
			ptr += 4;
			data_size -= 4;
		}
		else {
			//read data from the FIFO
			dummy =  readl((smi_info->smi_base + SMID));
			// handle alignment issues by using memcpy
			memcpy(ptr, &dummy, data_size);
			ptr += data_size;
			data_size -= data_size;
		}
	}

	// verify that the transfer has completed
	success += smi_check_transfer_complete();

exit_err:
	return success;
}
EXPORT_SYMBOL(smi_read_fifo);


int32_t smi_read_dma (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{
	int ret = -1;
	dma_addr_t dma_data_addr, dma_cb_addr;

	BUG_ON(smi_info->dma_cb == NULL);

	bcm2708_smi_print("slot=%d, address=0x%x, size=%d, blocking=%d, data=%p\n",
			slot, address, data_size_in_bytes, blocking, data);

	init_completion(&smi_info->comp);
	/* setup the smi for DMA */
	smi_setup_fifo( slot, address, data_size_in_bytes, FIFO_DMA, FIFO_READ );

	/*
	 * Should we handle high mem case? 
	 *   - No since we are running on embedded system where phy ram is <1GB
	 *   incase we ever go beyond 1GB we need to revisit this.
	 */
	dma_data_addr = dma_map_single( &smi_info->pdev->dev, (void *)data, data_size_in_bytes,
			DMA_FROM_DEVICE );

	if( dma_mapping_error( &smi_info->pdev->dev, dma_data_addr ) ) {

		dev_err(&smi_info->pdev->dev,
				"Couldn't DMA map a %d byte buffer\n", data_size_in_bytes);
		goto out_copy;
	}


	smi_info->dma_cb->ti = (SMI_PERIPH_DREQ << DMA_TI_PERMAP) 
							| ( 1 << DMA_TI_DEST_INC) 
							| ( 1 << DMA_TI_SRC_DREQ)
							| ( 1 << DMA_TI_DEST_WIDTH)
							| ( 1 << DMA_TI_INTEN );

	smi_info->dma_cb->source_ad = (void *) (VC_SMI_BASE + SMID);
	smi_info->dma_cb->dest_ad = (void *)  dma_data_addr;
	smi_info->dma_cb->txfr_len = data_size_in_bytes;
	smi_info->dma_cb->nextconbk = 0x0;
	smi_info->dma_cb->res1 = 0x0;
	smi_info->dma_cb->res2 = 0x0;

	bcm2708_smi_print("cb=0x%p ti=0x%x source_ad=0x%p dest_ad=0x%p len=0x%x"
			"stride=0x%x nextcon=0x%p\n", smi_info->dma_cb, smi_info->dma_cb->ti,
			 smi_info->dma_cb->source_ad, smi_info->dma_cb->dest_ad, 
			 smi_info->dma_cb->txfr_len, smi_info->dma_cb->stride, 
			 smi_info->dma_cb->nextconbk);

	/*
	 * Map Control block 
	 */
	dma_cb_addr = dma_map_single( &smi_info->pdev->dev, (void *)smi_info->dma_cb, 
			sizeof( struct bcm2708_dma_cb ), DMA_TO_DEVICE );

	if( dma_mapping_error( &smi_info->pdev->dev, dma_cb_addr ) ) {

		dev_err(&smi_info->pdev->dev,
				"Couldn't DMA map a control block buffer\n");
		goto out_copy_cb;
	}

	bcm2708_smi_print("Control block: Virtual=0x%p Physical=0x%x\n",
			 smi_info->dma_cb, dma_cb_addr );
	bcm2708_smi_print("Data: Virtual=0x%p Physical=0x%x\n",
			data, dma_data_addr );

	//reset dma 
	writel( 1 << DMA_CS_RESET, (smi_info->dma_base + DMA0_CS) );
	// write cb block addr
	writel( dma_cb_addr, (smi_info->dma_base + DMA0_CONBLOCK_AD));
	// activate dma
	writel( 1 << DMA_CS_ACTIVE, (smi_info->dma_base + DMA0_CS) );

	//wait for dma to finish
	//while( !( readl(smi_info->dma_base + DMA0_CS) & (1 << DMA_CS_END) ) );
	wait_for_completion( &smi_info->comp );

	dma_unmap_single( &smi_info->pdev->dev, dma_cb_addr, 
			sizeof( struct bcm2708_dma_cb ), DMA_TO_DEVICE );

	dma_unmap_single( &smi_info->pdev->dev, dma_data_addr,
			data_size_in_bytes, DMA_FROM_DEVICE );

	return 0;

out_copy_cb:
	dma_unmap_single( &smi_info->pdev->dev, dma_data_addr,
			data_size_in_bytes, DMA_FROM_DEVICE );
out_copy:
	printk(KERN_ERR"[%s] using fifo mode to transfer data\n", __func__);

	ret = smi_read_fifo( slot, address, data_size_in_bytes, blocking, data);

	return ret;
}
EXPORT_SYMBOL(smi_read_dma);


/* SMI Writes */
int32_t smi_write_direct (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{
	int ret = -1;
	uint32_t smidcs, smida;
	uint32_t write_data = 0;

	bcm2708_smi_print("slot=%d, address=0x%x, size=%d, blocking=%d, data=%p\n",
			slot, address, data_size_in_bytes, blocking, data);

	smida = (address & 0x3F) | (slot << SMIDA_DEVICE);
	writel(smida, (smi_info->smi_base + SMIDA));

	switch(data_size_in_bytes)
	{
		case 1:
			write_data = *((uint8_t *) data);
			break;
		case 2:
			write_data = *((uint16_t *) data);
			break;
		case 4:
			write_data = *((uint32_t *) data);
			break;
		default:
			printk(KERN_ERR"%s: incorrect data size=%d\n", __func__,
					data_size_in_bytes);
			ret = -EINVAL;
			goto exit_err;
	}

	writel(write_data, (smi_info->smi_base + SMIDD));	

	//start the transfer
	smidcs = (1 << SMIDCS_ENABLE | 1 << SMIDCS_WRITE);
	writel(smidcs, (smi_info->smi_base + SMIDCS));

	smidcs |= (1 << SMIDCS_START);
	writel(smidcs, (smi_info->smi_base + SMIDCS));

	//wait ofr data complete
	smi_wait_direct_complete();
	
	smidcs = (1 << SMIDCS_DONE);
	writel(smidcs, (smi_info->smi_base + SMIDCS));

	return 0;
exit_err:
	return ret;
}
EXPORT_SYMBOL(smi_write_direct);


int32_t smi_write_fifo (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{

	int32_t success = -1;
	uint32_t timeout = 0;
	uint32_t dummy = 0;
	uint8_t * ptr = (uint8_t *)data;
	uint32_t data_size = data_size_in_bytes;

	printk(KERN_ERR"%s called\n", __func__);
	smi_setup_fifo(slot, address, data_size_in_bytes, FIFO_NON_DMA, FIFO_WRITE);

	// do the write
	while(data_size)
	{
		timeout = 0;

		 //wait until there is space in the write FIFO.
		while ( !(readl((smi_info->smi_base + SMICS)) & (1 << SMICS_TXW)) &&
				(timeout++ < 5000) );
		if( timeout > 5000) {
			printk(KERN_ERR "[%s]: timeout while waiting for data\n", __func__);
			success = -1;
			goto exit_err;
		}

		// handle alignment issues using memcpy
		if( data_size / 4 ) {
			memcpy(&dummy, ptr, 4);
			data_size -= 4;
			ptr += 4;
		}
		else {
			memcpy(&dummy, ptr, data_size);
			data_size -= data_size;
			ptr += data_size;
		}

		// write data to the FIFO
		writel(dummy, (smi_info->smi_base + SMID));
	}

	// verify that the transfer has completed
	success += smi_check_transfer_complete();

exit_err:
	return(success);
}
EXPORT_SYMBOL(smi_write_fifo);


int32_t smi_write_dma (const uint32_t slot, const uint32_t address,
		const uint32_t data_size_in_bytes, const uint32_t blocking,
		const void *data)
{
	int ret = -1;
	dma_addr_t dma_data_addr, dma_cb_addr;

	BUG_ON(smi_info->dma_cb == NULL);

	bcm2708_smi_print("slot=%d, address=0x%x, size=%d, blocking=%d, data=%p\n",
			slot, address, data_size_in_bytes, blocking, data);

	init_completion(&smi_info->comp);

	/* setup the smi for DMA */
	smi_setup_fifo( slot, address, data_size_in_bytes, FIFO_DMA, FIFO_WRITE );

	/*
	 * Should we handle high mem case? 
	 *   - No since we are running on embedded system where phy ram is <1GB
	 *   incase we ever go beyond 1GB we need to revisit this.
	 */
	dma_data_addr = dma_map_single( &smi_info->pdev->dev, (void *)data, data_size_in_bytes,
			DMA_TO_DEVICE );

	if( dma_mapping_error( &smi_info->pdev->dev, dma_data_addr ) ) {

		dev_err(&smi_info->pdev->dev,
				"Couldn't DMA map a %d byte buffer\n", data_size_in_bytes);
		goto out_copy;
	}


	smi_info->dma_cb->ti = (SMI_PERIPH_DREQ << DMA_TI_PERMAP) 
							| ( 1 << DMA_TI_SRC_INC) 
							| ( 1 << DMA_TI_DEST_DREQ)
							| ( 1 << DMA_TI_SRC_WIDTH)
							| ( 1 << DMA_TI_INTEN);

	smi_info->dma_cb->source_ad = (void *)dma_data_addr;	
	smi_info->dma_cb->dest_ad = (void *) (VC_SMI_BASE + SMID);

	smi_info->dma_cb->txfr_len = data_size_in_bytes;
	smi_info->dma_cb->nextconbk = 0x0;
	smi_info->dma_cb->res1 = 0x0;
	smi_info->dma_cb->res2 = 0x0;

	bcm2708_smi_print("cb=0x%p ti=0x%x source_ad=0x%p dest_ad=0x%p len=0x%x"
			"stride=0x%x nextcon=0x%p\n", smi_info->dma_cb, smi_info->dma_cb->ti,
			 smi_info->dma_cb->source_ad, smi_info->dma_cb->dest_ad, 
			 smi_info->dma_cb->txfr_len, smi_info->dma_cb->stride, 
			 smi_info->dma_cb->nextconbk);

	/*
	 * Map Control block 
	 */
	dma_cb_addr = dma_map_single( &smi_info->pdev->dev, (void *)smi_info->dma_cb, 
			sizeof( struct bcm2708_dma_cb ), DMA_TO_DEVICE );

	if( dma_mapping_error( &smi_info->pdev->dev, dma_cb_addr ) ) {

		dev_err(&smi_info->pdev->dev,
				"Couldn't DMA map a control block buffer\n");
		goto out_copy_cb;
	}

	bcm2708_smi_print("Control block:Virtual=0x%p Physical=0x%x\n",
			 smi_info->dma_cb, dma_cb_addr );
	bcm2708_smi_print("Data:Virtual=0x%p Physical=0x%x\n",
			 data, dma_data_addr );

	//reset dma 
	writel( 1 << DMA_CS_RESET, (smi_info->dma_base + DMA0_CS) );
	// write cb block addr
	writel( dma_cb_addr, (smi_info->dma_base + DMA0_CONBLOCK_AD));
	// activate dma
	writel( 1 << DMA_CS_ACTIVE, (smi_info->dma_base + DMA0_CS) );

	//wait for dma to finish
	//while( !( readl(smi_info->dma_base + DMA0_CS) & (1 << DMA_CS_END) ) );
	wait_for_completion( &smi_info->comp);

	dma_unmap_single( &smi_info->pdev->dev, dma_cb_addr, 
			sizeof( struct bcm2708_dma_cb ), DMA_TO_DEVICE );

	dma_unmap_single( &smi_info->pdev->dev, dma_data_addr,
			data_size_in_bytes, DMA_TO_DEVICE );

	return 0;

out_copy_cb:
	dma_unmap_single( &smi_info->pdev->dev, dma_data_addr,
			data_size_in_bytes, DMA_TO_DEVICE );
out_copy:
	printk(KERN_ERR"[%s] using fifo mode to transfer data\n", __func__);

	ret = smi_write_fifo( slot, address, data_size_in_bytes, blocking, data);

	return ret;
}
EXPORT_SYMBOL(smi_write_dma);


int32_t smi_setup_timing( const uint32_t slot, 
		const struct smi_periph_setup *periph_setup, const uint32_t smi_freq_hz )
{
	int ret = -1;
	uint32_t mode68, format32, pixel_swap;
	uint32_t write_width, read_width, clock_period_in_ns = 1000000000/smi_freq_hz;
	uint32_t read_strobe_time, read_pace_time, read_hold_time, read_setup_time;
	uint32_t write_strobe_time, write_pace_time, write_hold_time, write_setup_time;
	uint32_t val;
	uint32_t smidsr, smidsw, smics;


	bcm2708_smi_print("slot=%d periph_setup=%p smi_freq_hz=%d\n",
			slot, periph_setup, smi_freq_hz);

	if(!smi_info) {
		printk(KERN_ERR "SMI driver not initialized\n");
		ret = -ENODEV;
		goto exit_err_init;
	}
	
	if(slot > 3) {
		printk(KERN_ERR "slot should be less than 3\n");
		ret = -ENOENT;
		goto exit_err_slot;
	}

	/* disable smi */
	val = readl(smi_info->smi_base + SMICS);
	val &= ~1;
	writel((smi_info->smi_base + SMICS), val);
	
	switch(periph_setup->read_timings.transfer_width) 
	{
		case SMI_TRANSFER_WIDTH_8BIT:
			bcm2708_smi_print("Read xfer width=8bit\n");
			read_width = 0;
			break;
		default:
			printk(KERN_ERR"[%s]:Invalid read transfer width\n", __func__);
			ret = -EINVAL;
			goto exit_err_xfer;
			break;
	}
	
	switch(periph_setup->write_timings.transfer_width)
	{
		case SMI_TRANSFER_WIDTH_8BIT:
			bcm2708_smi_print("write xfer width=8bit\n");
			write_width = 0;
			break;
		default:
			printk(KERN_ERR"[%s]:Invalid write transfer width\n", __func__);
			ret = -EINVAL;
			goto exit_err_xfer;
			break;
	}
	
	read_strobe_time = (periph_setup->read_timings.strobe_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	read_pace_time = (periph_setup->read_timings.pace_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	read_hold_time = (periph_setup->read_timings.hold_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	read_setup_time = (periph_setup->read_timings.setup_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	
	read_strobe_time = clamp_t(uint32_t, read_strobe_time, 127, 0);
	read_pace_time = clamp_t(uint32_t, read_pace_time, 127, 0);
	read_hold_time = clamp_t(uint32_t, read_hold_time, 63, 0);
	read_setup_time = clamp_t(uint32_t, read_setup_time, 63, 0);

	mode68 = (periph_setup->mode == SMI_MODE_68);

	bcm2708_smi_print("Read: stobe_time=%d pace_time=%d hold_time=%d"
			"setup_time=%d mode68=%d width=%d\n", read_strobe_time,
			read_pace_time, read_hold_time, read_setup_time, mode68,
			read_width);

	smidsr = (read_strobe_time << SMIDS_STROBE)
			| (read_pace_time << SMIDS_PACE)
			| (read_hold_time << SMIDS_HOLD)
			| (mode68  << SMIDS_MODE68)
			| (read_setup_time << SMIDS_SETUP)
			| (read_width << SMIDS_WIDTH);

	/* smidsw */
	write_strobe_time = (periph_setup->write_timings.strobe_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	write_pace_time = (periph_setup->write_timings.pace_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	write_hold_time = (periph_setup->write_timings.hold_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;
	write_setup_time = (periph_setup->write_timings.setup_time_in_ns + 
			clock_period_in_ns - 1) / clock_period_in_ns;

	write_strobe_time = clamp_t( uint32_t, write_strobe_time, 127, 0);
	write_pace_time = clamp_t(uint32_t, write_pace_time, 127, 0);
	write_hold_time = clamp_t(uint32_t, write_hold_time, 63, 0);
	write_setup_time = clamp_t(uint32_t, write_setup_time, 63, 0);
	
	pixel_swap = (periph_setup->pixel_bits_swapped && 
			(periph_setup->pixel_format != SMI_PIXEL_FORMAT_DONT_CARE)) ? 1 : 0;

	format32 = (periph_setup->pixel_format == SMI_PIXEL_FORMAT_32BIT_RGB888);

	bcm2708_smi_print("write: stobe_time=%d pace_time=%d hold_time=%d"
			"setup_time=%d pixel_swap=%d format32=%d width=%d\n", 
			write_strobe_time,write_pace_time, write_hold_time,
			write_setup_time, pixel_swap, format32, write_width);

	smidsw = (write_strobe_time << SMIDS_STROBE)
		| (write_pace_time << SMIDS_PACE)
		| (write_hold_time << SMIDS_HOLD)
		| (pixel_swap  << SMIDS_SWAP)
		| (format32 << SMIDS_FORMAT)
		| (write_setup_time << SMIDS_SETUP)
		| (write_width << SMIDS_WIDTH);


	/* sync settings */
	writel(smidsr, (smi_info->smi_base + SMIDSR0 + (slot * (SMIDSR1 - SMIDSR0))));
	writel(smidsw, (smi_info->smi_base + SMIDSW0 + (slot * (SMIDSW1 - SMIDSW0))));

	/*smics*/
	smics = readl(smi_info->smi_base + SMICS);
	if (periph_setup->tearing_effect)
		smics |= SMICS_TEEN;
	else
		smics &= ~SMICS_TEEN;

	if (periph_setup->hvs_input)
		smics |= SMICS_PVMODE;
	else
		smics &= ~SMICS_PVMODE;
	writel(smics, (smi_info->smi_base + SMICS));

	return 0;

exit_err_xfer:
exit_err_slot:
exit_err_init:
	return ret;
}
EXPORT_SYMBOL(smi_setup_timing);


static irqreturn_t bcm2708_dma4_irq_handler(int irq, void *dev_id)
{
	struct bcm2708_smi_info *info = (struct bcm2708_smi_info *) dev_id;
	uint32_t dma_cs;

	BUG_ON( info == NULL );

	/* notify the task dma is done */
	complete( &(info->comp) );

	/* clear the interrupt */
	dma_cs = readl(info->dma_base + DMA0_CS );
	dma_cs |= (1 << DMA_CS_INT );
	writel(dma_cs, (info->dma_base + DMA0_CS) );
	
	return IRQ_HANDLED;
}

static int bcm2708_smi_probe(struct platform_device *pdev)
{
	struct bcm2708_smi_info *info;
	int ret;
	struct resource *res;
	
	bcm2708_smi_print("probe=%p\n", pdev);	

	info = kmalloc(sizeof(*info), GFP_KERNEL);
	if (info == NULL) {
		dev_err(&pdev->dev, "no memory for smi info\n");
		ret = -ENOMEM;
		goto exit_error;
	}

	/* get the smi  memory region */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "failed to get memory region resource\n");
		ret = -ENODEV;
		goto exit_rsrc_error;
	}

	printk(KERN_INFO"[%s] SMI start=%p end=%p\n", __func__, (void *)res->start,
			(void *)res->end);

	/* request memory region */
	info->smi_res = request_mem_region_exclusive(res->start,
			resource_size(res), "bcm2708smi_regs");

	if (info->smi_res == NULL) {
		dev_err(&pdev->dev, "Unable to request smi memory region\n");
		ret = -ENOENT;
		goto exit_mem_error;
	}

	info->smi_base = ioremap(res->start, SZ_4K);
	if( info->smi_base == NULL) { 
		dev_err(&pdev->dev, "Unable to request smi memory region\n");
		ret = -ENOMEM;
		goto exit_ioremap_error;
	}

	/* get the dma memory region */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (res == NULL) {
		dev_err(&pdev->dev, "failed to get memory region resource\n");
		ret = -ENODEV;
		goto exit_dma_rsrc_error;
	}

	printk(KERN_INFO "[%s] DMA start=%p end=%p\n", __func__, (void *)res->start,
			(void *)res->end);

	info->dma_base = __io_address(res->start);

	/* use DMA_CHANNEL 4 */
	info->dma_base = (void __iomem *) ((char *)info->dma_base + (SMI_DMA_CHANNEL * 0x100));
		
	/* Allocate memory for dma control block */
	
	info->dma_mem = kmalloc( (sizeof(struct bcm2708_dma_cb) + DMA_BYTE_ALIGN),
			GFP_KERNEL );

	if( NULL == info->dma_mem ) {
		dev_err(&pdev->dev, "Unable to allocate memory for dma control block\n");
		ret = -ENOMEM;
		goto exit_dma_kmalloc_error;
	}

	info->dma_cb = (struct bcm2708_dma_cb *) ALIGN_UP(info->dma_mem,
			DMA_BYTE_ALIGN);

	info->pdev = pdev;
	smi_info = info;

	ret = request_irq( IRQ_DMA4, bcm2708_dma4_irq_handler, 0, "DMA4", (void *)smi_info );
	if( ret != 0 ) {
		printk(KERN_ERR "unable to request IRQ %d for DMA (error %d) \n",
				IRQ_DMA4, ret);
		goto exit_irq_error;

	}
	return 0;

exit_irq_error:
	smi_info = NULL;
	kfree(info->dma_mem);
exit_dma_kmalloc_error:
exit_dma_rsrc_error:
	iounmap(info->smi_base);
exit_ioremap_error:
	release_region(info->smi_res->start, resource_size(info->smi_res));
exit_mem_error:
exit_rsrc_error:
	kfree(info);
exit_error:
	return ret;

}


static int __devexit bcm2708_smi_remove(struct platform_device *pdev)
{
	struct bcm2708_smi_info *info = smi_info;

	free_irq( IRQ_DMA4, smi_info);
	kfree(info->dma_mem);
	iounmap(info->smi_base);
	release_region(info->smi_res->start, resource_size(info->smi_res));
	kfree(info);
	smi_info = NULL;

	return 0;
}

#define bcm2708_smi_suspend    NULL
#define bcm2708_smi_resume     NULL

static struct platform_driver bcm2708_smi_driver = {
	.probe      = bcm2708_smi_probe,
	.remove     = __devexit_p(bcm2708_smi_remove),
	.suspend    = bcm2708_smi_suspend,
	.resume     = bcm2708_smi_resume,
	.driver     = {
		.name   = "bcm2708_smi",
		.owner  = THIS_MODULE,
	},

};

static int __init bcm2708_smi_init(void)
{
	printk(KERN_INFO"BCM2708 SMI Driver,(C) 2010 Broadcom Corp\n");
	return platform_driver_register(&bcm2708_smi_driver);

}

static void __exit bcm2708_smi_exit(void)
{
	platform_driver_unregister(&bcm2708_smi_driver);
}

module_init(bcm2708_smi_init);
module_exit(bcm2708_smi_exit);
