/*
 *  linux/drivers/serial/serial_bcm2708.c
 *
 *  Copyright (C) 2010 Broadcom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This provides a "virtual" UART port that uses an area of shared memory
 * shared with a broadcom communications processor.
 */


#if defined(CONFIG_SERIAL_BCM2708_MINI_UART_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif

#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/device.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_core.h>
#include <linux/serial.h>
#include <linux/mm.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/slab.h>

#include <asm/io.h>
#include <asm/sizes.h>
#include <mach/vcio.h>
#include <mach/platform.h>

#include "serial_bcm2708.h"

#define DRIVER_NAME "bcm2708_mini_uart"

#define SERIAL_BCMUART_MAJOR   207          // stolen - really belongs to IMX
#define MINOR_START	        25           // stolen? - really belongs to IMX
#define DEV_NAME		"ttybcm"

/*
 * We wrap our port structure around the generic uart_port.
 */
struct bcm2708_mini_uart_port {
	struct uart_port	port;
    void __iomem *uart_base;
    int rx_enabled;
};

static int uart_noinput = 0;

/* ----------------------------------------------------------------------
 *      UART Device
 * -------------------------------------------------------------------- */

static void bcm2708_mini_uart_stop_tx(struct uart_port *port)
{
}

static void bcm2708_mini_uart_stop_rx(struct uart_port *port)
{
}

static void bcm2708_mini_uart_enable_ms(struct uart_port *port)
{
}

static void bcm2708_mini_uart_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
}

static void bcm2708_mini_uart_break_ctl(struct uart_port *port, int break_state)
{
}

static void bcm2708_mini_tx_chars(struct bcm2708_mini_uart_port *uap)
{
	struct circ_buf *xmit = &uap->port.state->xmit;
	int count;

	if (uap->port.x_char) {
        if ( !(readw(uap->port.membase + AUX_MU_LSR_REG) & UART_STATUS_TX_READY))
			return;
		writew(uart_noinput ? '6' : uap->port.x_char, uap->port.membase + AUX_MU_IO_REG);
		uap->port.icount.tx++;
		uap->port.x_char = 0;
		return;
	}
	if (uart_circ_empty(xmit) || uart_tx_stopped(&uap->port)) {
        writew (0x00000001, uap->port.membase + AUX_MU_IER_REG);
		return;
	}

	count = uap->port.fifosize >> 1;
	do {
        if ( !(readw(uap->port.membase + AUX_MU_LSR_REG) & UART_STATUS_TX_READY))
			break;
		writew(uart_noinput ? '6' : xmit->buf[xmit->tail], uap->port.membase + AUX_MU_IO_REG);
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		uap->port.icount.tx++;
		if (uart_circ_empty(xmit))
			break;
	} while (--count > 0);

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&uap->port);
}

static void bcm2708_mini_uart_start_tx(struct uart_port *port)
{
    struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
    writew (0x00000003, uap->port.membase + AUX_MU_IER_REG);
}

static irqreturn_t bcm2708_mini_uart_rx_chars(int irq, void *port)
{
    struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
	struct tty_struct *tty = uap->port.state->port.tty;
	unsigned int status, ch, flag, max_count = 256;

	status = readw(uap->port.membase + AUX_MU_LSR_REG);
	while ((status & UART_STATUS_RX_READY) && max_count--) {
		ch = readw(uap->port.membase + AUX_MU_IO_REG);
		flag = TTY_NORMAL;
		uap->port.icount.rx++;

		if (uart_noinput || uart_handle_sysrq_char(&uap->port, ch & 255)){
			goto ignore_char;
        }

		uart_insert_char(&uap->port, ch, 0, ch, flag);

	ignore_char:
		status = readw(uap->port.membase + AUX_MU_LSR_REG); //read data available register
	}

	tty_flip_buffer_push(tty);
    
    bcm2708_mini_tx_chars( uap  );
    return IRQ_HANDLED;
}

static unsigned int bcm2708_tx_empty(struct uart_port *port)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
	unsigned int status = readw(uap->port.membase + AUX_MU_STAT_REG);
	return status & BIT_8 ? TIOCSER_TEMT : 0;
}

static unsigned int bcm2708_get_mctrl(struct uart_port *port)
{
	unsigned int tmp = TIOCM_DSR | TIOCM_CAR;

	tmp |= TIOCM_CTS;
	tmp |= TIOCM_RTS;

	return tmp;
}

#ifdef CONFIG_CONSOLE_POLL
static int bcm2708_get_poll_char(struct uart_port *port)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
	unsigned int status;
	do {
		status = readw(uap->port.membase + AUX_MU_LSR_REG);
	} while (uart_noinput || !(status & UART_STATUS_RX_READY));
	return readw(uap->port.membase + AUX_MU_IO_REG);
}

static void bcm2708_put_poll_char(struct uart_port *port,
			 unsigned char ch)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
	while ( !(readw(uap->port.membase + AUX_MU_LSR_REG) & UART_STATUS_TX_READY))
		barrier();

	writew(uart_noinput ? '6' : ch, uap->port.membase + AUX_MU_IO_REG);
}

#endif /* CONFIG_CONSOLE_POLL */

static int bcm2708_mini_uart_startup(struct uart_port *port)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;
	unsigned int baud_rate_reg = ( (VC_SYSTEM_CLOCK_FREQ / (8 * 115200)) - 1 );
	int retval;

    /*
	 * Allocate the IRQ
	 */
	retval = request_irq(uap->port.irq, bcm2708_mini_uart_rx_chars, IRQF_SHARED, "uart-bcm_mini", uap);
	if (retval){
        printk("FAILED IRQ registration");
		return -1;
    }

	/*
	 ** First, Enable the GPIO alt function
	 ** IO_WRITE ((GPIO_BASE + 0xC), 0x00000480)
	 ** THIS IS DONE BY VC platform_init function
	 */

	/* Enable the mini UART */
	writew( BIT_2 | BIT_1 | BIT_0, uap->port.membase + AUX_MU_EN_REG);
	
	/* Setup the Baud rate */
	writew ( baud_rate_reg, uap->port.membase  + AUX_MU_BAUD);
	
	/* Setup the data to be 8 bit */
	writew ( BIT_1 | BIT_0, uap->port.membase + AUX_MU_LCR_REG);
    
    /* Enable the interrupt */
    writew ( BIT_0, uap->port.membase + AUX_MU_IER_REG);

	return 0;
}

static void bcm2708_mini_uart_shutdown(struct uart_port *port)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;

	/* Disable the mini UART */
	writew(0x00000006, uap->port.membase + AUX_MU_EN_REG);

    /* Disable the interrupt */
    writew (0x00000000, uap->port.membase + AUX_MU_IER_REG);
    
	/* free irq */
	free_irq(uap->port.irq, NULL);
}

static void
bcm2708_mini_uart_set_termios(struct uart_port *port, struct ktermios *termios,
			 struct ktermios *old)
{
	struct bcm2708_mini_uart_port *sport = (struct bcm2708_mini_uart_port *)port;
	unsigned long flags;
	unsigned int old_csize = old ? old->c_cflag & CSIZE : CS8;
    
	/*
	 * If we don't support modem control lines, don't allow
	 * these to be set.
	 */
	if (1) {
		termios->c_cflag &= ~(HUPCL | CRTSCTS | CMSPAR);
		termios->c_cflag |= CLOCAL;
	}

	/*
	 * We only support CS7 and CS8.
	 */
	while ((termios->c_cflag & CSIZE) != CS7 &&
	       (termios->c_cflag & CSIZE) != CS8) {
		termios->c_cflag &= ~CSIZE;
		termios->c_cflag |= old_csize;
		old_csize = CS8;
	}

	spin_lock_irqsave(&sport->port.lock, flags);

    uart_update_timeout(port, termios->c_cflag, 115200);
	sport->port.read_status_mask = 0;

	/*
	 * Characters to ignore
	 */
	sport->port.ignore_status_mask = 0;
        
	spin_unlock_irqrestore(&sport->port.lock, flags);
}

static const char *bcm2708_mini_uart_type(struct uart_port *port)
{
	struct bcm2708_mini_uart_port *sport = (struct bcm2708_mini_uart_port *)port;
	return sport->port.type == PORT_BCM_MINI_UART ? "BCM_MINIUART" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'
 */
static void bcm2708_release_port(struct uart_port *port)
{
	struct platform_device *pdev = to_platform_device(port->dev);
	struct resource *mmres;
	mmres = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(mmres->start, mmres->end - mmres->start + 1);
}

/*
 * Request the memory region(s) being used by 'port'
 */
static int bcm2708_request_port(struct uart_port *port)
{
    /* Resources are allocated in probe :DIRTY*/
	return  0;
}

/*
 * Configure/autoconfigure the port.
 */
static void bcm2708_config_port(struct uart_port *port, int flags)
{
	if (flags & UART_CONFIG_TYPE) {
		port->type = PORT_BCM_MINI_UART;
		bcm2708_request_port(port);
	}
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int bcm2708_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_UNKNOWN && ser->type != PORT_BCM_MINI_UART)
		ret = -EINVAL;
	if (ser->baud_base < 9600)
		ret = -EINVAL;
    
    if (ret == -EINVAL)
        printk(KERN_ERR DRIVER_NAME ": serial port type %d is "
                "invalid\n", ser->type);
	return ret;
}

static struct uart_ops amba_bcm2708_mini_uart_pops = {
	.tx_empty	= bcm2708_tx_empty,
	.set_mctrl	= bcm2708_mini_uart_set_mctrl,
	.get_mctrl	= bcm2708_get_mctrl,
	.stop_tx	= bcm2708_mini_uart_stop_tx,
	.start_tx	= bcm2708_mini_uart_start_tx,
	.stop_rx	= bcm2708_mini_uart_stop_rx,
	.enable_ms	= bcm2708_mini_uart_enable_ms,
	.break_ctl	= bcm2708_mini_uart_break_ctl,
	.startup	= bcm2708_mini_uart_startup,
	.shutdown	= bcm2708_mini_uart_shutdown,
	.set_termios	= bcm2708_mini_uart_set_termios,
	.type		= bcm2708_mini_uart_type,
	.release_port	= bcm2708_release_port,
	.request_port	= bcm2708_request_port,
	.config_port	= bcm2708_config_port,
	.verify_port	= bcm2708_verify_port,
#ifdef CONFIG_CONSOLE_POLL
	.poll_get_char = bcm2708_get_poll_char,
	.poll_put_char = bcm2708_put_poll_char,
#endif //CONFIG_CONSOLE_POLL
};

static struct bcm2708_mini_uart_port *bcm_console_port;

/* ----------------------------------------------------------------------
 *      Console Device
 * -------------------------------------------------------------------- */

#ifdef CONFIG_SERIAL_BCM2708_MINI_UART_CONSOLE
static void bcm2708_mini_uart_console_putchar(struct uart_port *port, int ch)
{
	struct bcm2708_mini_uart_port *uap = (struct bcm2708_mini_uart_port *)port;

	while ( !(readw(uap->port.membase + AUX_MU_LSR_REG) & UART_STATUS_TX_READY))
		barrier();

	writew(uart_noinput ? '6' : ch, uap->port.membase + AUX_MU_IO_REG);
}

static void
bcm2708_mini_uart_console_write(struct console *co, const char *s, unsigned int count)
{
	struct bcm2708_mini_uart_port *uap = bcm_console_port;

	uart_console_write(&uap->port, s, count, bcm2708_mini_uart_console_putchar);

	/*
	 *	Finally, wait for transmitter to become empty
	 */
	while (!bcm2708_tx_empty(&uap->port))
                continue;
    
}

static int __init bcm2708_mini_uart_console_setup(struct console *co, char *options)
{
	struct bcm2708_mini_uart_port *uap;
	int baud = 115200;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';
    int ret;

	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	if (co->index == -1 || co->index >= 1)
		co->index = 0;
	uap = bcm_console_port;
	if (!uap)
		return -ENODEV;

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);

	if (baud == 0) {
		printk("baud == 0: no input allowed\n");
		uart_noinput = 1;
		baud = 115200;
	}
    ret = uart_set_options(&uap->port, co, baud, parity, bits, flow);
    printk(KERN_INFO DRIVER_NAME ": registered console - rc %d\n", ret);
	return ret;
}

static struct uart_driver bcm2708_mini_reg;
static struct console bcm_mini_console = {
	.name		= DEV_NAME,
	.write		= bcm2708_mini_uart_console_write,
	.device		= uart_console_device,
	.setup		= bcm2708_mini_uart_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &bcm2708_mini_reg,
};

#define BCM_MINI_CONSOLE	&bcm_mini_console

#else
#define BCM_MINI_CONSOLE	NULL
#endif

/* ----------------------------------------------------------------------
 *      Platform Device
 * -------------------------------------------------------------------- */

static struct uart_driver bcm2708_mini_reg = {
	.owner          = THIS_MODULE,
	.driver_name    = DRIVER_NAME,
	.dev_name       = DEV_NAME,
	.major          = SERIAL_BCMUART_MAJOR,
	.minor          = MINOR_START,
	.nr             = 1,
	.cons           = BCM_MINI_CONSOLE,
};

static int bcm2708_mini_uart_probe(struct platform_device *dev)
{
	struct bcm2708_mini_uart_port *sport;
	int ret = 0;
    void __iomem *base;
    struct resource *res;
    
	sport = kzalloc(sizeof(*sport), GFP_KERNEL);
	if (!sport) {
        printk(KERN_ERR DRIVER_NAME ": failed to allocate "
                    "port memory\n");
		return -ENOMEM;
    }

	res = platform_get_resource(dev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		ret = -ENODEV;
		goto free;
	}

	base = ioremap(res->start, SZ_4K);
	if( base == NULL) { 
		ret = -ENOMEM;
		goto free;
	}
    sport->uart_base = base;
    sport->rx_enabled = 0;

    writew(BIT_2 | BIT_1 | BIT_0, base + AUX_MU_EN_REG);
	
	/* Setup the Baud rate */
	writew (0x10e, base  + AUX_MU_BAUD);
	
	/* Setup the data to be 8 bit */
	writew (BIT_1 | BIT_0, base + AUX_MU_LCR_REG);

	sport->port.dev = &dev->dev;
	sport->port.mapbase = UART1_BASE;
	sport->port.membase = base;
	sport->port.iotype = UPIO_MEM;
	sport->port.irq = IRQ_SPARE;
	sport->port.fifosize = 64;
	sport->port.ops = &amba_bcm2708_mini_uart_pops;
	sport->port.flags = UPF_BOOT_AUTOCONF;
	sport->port.line = 0;

	bcm_console_port = sport;

	ret = uart_add_one_port(&bcm2708_mini_reg, &sport->port);

	if (ret != 0) {
                printk(KERN_ERR DRIVER_NAME ": failed to add UART port "
                       "on line %d - rc %d\n", sport->port.line, ret);
		ret = -ENODEV;
		goto deinit;
    }

    platform_set_drvdata(dev, &sport->port);

    printk(KERN_INFO DRIVER_NAME ": located device\n");
	return 0;

        
deinit:
	iounmap(sport->port.membase);
free:
	kfree(sport);

	return ret;
}

static int bcm2708_mini_uart_remove(struct platform_device *dev)
{
	struct bcm2708_mini_uart_port *sport = platform_get_drvdata(dev);

	platform_set_drvdata(dev, NULL);

	if (sport) {
		uart_remove_one_port(&bcm2708_mini_reg, &sport->port);
	}

	iounmap(sport->port.membase);
	kfree(sport);
	return 0;
}

#ifdef CONFIG_PM
static int bcm2708_mini_uart_suspend(struct platform_device *dev, pm_message_t state)
{
	struct bcm2708_mini_uart_port *sport = platform_get_drvdata(dev);

	if (sport)
		uart_suspend_port(&bcm2708_mini_reg, &sport->port);

	return 0;
}

static int bcm2708_mini_uart_resume(struct platform_device *dev)
{
	struct bcm2708_mini_uart_port *sport = platform_get_drvdata(dev);

	if (sport)
		uart_resume_port(&bcm2708_mini_reg, &sport->port);

	return 0;
}
#endif

static struct platform_driver bcm2708_mini_uart_driver = {
	.probe		= bcm2708_mini_uart_probe,
	.remove		= bcm2708_mini_uart_remove,
#ifdef CONFIG_PM
	.suspend	= bcm2708_mini_uart_suspend,
	.resume		= bcm2708_mini_uart_resume,
#endif
	.driver		= {
		.name	= "bcm2708_mini_uart",
		.owner	= THIS_MODULE,
	},
};

static int __init bcm2708_mini_uart_init(void)
{
	int ret;
	printk(KERN_INFO "Serial: BROADCOM 2708 mini UART driver\n");

	ret = uart_register_driver(&bcm2708_mini_reg);
	if (ret == 0) {
		ret = platform_driver_register(&bcm2708_mini_uart_driver);
		if (ret != 0) {
			uart_unregister_driver(&bcm2708_mini_reg);
			printk(KERN_ERR DRIVER_NAME ": failed to register "
				   "on platform\n");
		}
		else
			printk(KERN_INFO DRIVER_NAME ": registered UART\n");
	}
	else {
		printk(KERN_ERR DRIVER_NAME ": failed to register UART\n");
	}
	return ret;
}

static void __exit bcm2708_mini_uart_exit(void)
{
	platform_driver_unregister(&bcm2708_mini_uart_driver);
	uart_unregister_driver(&bcm2708_mini_reg);
}

/*
 * While this can be a module, if builtin it's most likely the console
 * So let's leave module_exit but move module_init to an earlier place
 */
arch_initcall(bcm2708_mini_uart_init);
module_exit(bcm2708_mini_uart_exit);

MODULE_AUTHOR("BROADCOM");
MODULE_DESCRIPTION("BCM2708 mini UART driver");
MODULE_LICENSE("GPL");
