#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/time.h>
#include <linux/dma-mapping.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/acct.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/kthread.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/wait.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>

#include <mach/hardware.h>
#include <mach/io.h>
#include <mach/memory.h>
#include <mach/platform.h>
#include <mach/ipc.h>

#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

#include "bcm2708_fb.h"

struct bcm2708_fb {
	char *dev_name;
	unsigned int fbnum;
	atomic_t usage;
	unsigned int is_yuv;
	unsigned int is_suspended;
	/* struct resource *fbmem_res; */
	phys_addr_t phys_fbbase;
	void __iomem *reg_base;
	void __iomem *fbmem;
	uint32_t ipc_id;
	int irq;
	spinlock_t lock;
	wait_queue_head_t wait_for_irq;
	uint8_t update_count;
	int rotation;
	struct fb_info fb;
	u32	cmap[16];
#ifdef CONFIG_ANDROID_POWER
	android_early_suspend_t early_suspend;
#endif
};

struct rect {
	uint16_t x;
	uint16_t y;
	uint16_t w;
	uint16_t h;
};

struct yuv_pixfmt {
	uint16_t fmt;
	uint16_t colour_space;
};

typedef enum {
	FB_PIXEL_FORMAT_RGB565 = 0,
	FB_PIXEL_FORMAT_RGB888,
	FB_PIXEL_FORMAT_RGBA32,
	FB_PIXEL_FORMAT_BGR565,
	FB_PIXEL_FORMAT_BGR888,
	FB_PIXEL_FORMAT_YUV_NV12,
	FB_PIXEL_FORMAT_YUV_NV21
} pixel_format_t;

typedef enum {
    CSC_ITUR_BT_601_5 = 1,
    CSC_ITUR_BT_709_3,
    CSC_JPEG_JIF
} color_space_t;

typedef enum {
    FB0_CACHE_POLICY_UNCACHED = 0,
    FB0_CACHE_POLICY_WRITETHROUGH,
#if 0
    FB0_CACHE_POLICY_WRITEBACK
#endif
} cache_policy_t;

#define DEFAULT_OVL_WIDTH 96
#define DEFAULT_OVL_HEIGHT 96

/*
	fb is /dev/fb0. 
	Otherwise ovl<num> maps to /dev/fb<num>. 

	- 0 is the main grahics layer.
	- 1 and 2 are for the grid screen.
	- 3 is another smaller overlay that can be used while
	- games are running. 
	- 4 through 7 are all for cursors. 
*/

/* FB0 width and height */
static uint32_t fb_width = FRAMEBUFFER_WIDTH;
static uint32_t fb_height = FRAMEBUFFER_HEIGHT;

static uint32_t ovl1_width = FRAMEBUFFER_WIDTH;
static uint32_t ovl1_height = FRAMEBUFFER_HEIGHT;
static uint32_t ovl2_width = FRAMEBUFFER_WIDTH;
static uint32_t ovl2_height = FRAMEBUFFER_HEIGHT;
static uint32_t ovl3_width  = DEFAULT_OVL_WIDTH;
static uint32_t ovl3_height = DEFAULT_OVL_HEIGHT;
static uint32_t ovl4_width  = DEFAULT_OVL_WIDTH;
static uint32_t ovl4_height = DEFAULT_OVL_HEIGHT;
static uint32_t ovl5_width  = DEFAULT_OVL_WIDTH;
static uint32_t ovl5_height = DEFAULT_OVL_HEIGHT;
static uint32_t ovl6_width  = DEFAULT_OVL_WIDTH;
static uint32_t ovl6_height = DEFAULT_OVL_HEIGHT;
static uint32_t ovl7_width  = DEFAULT_OVL_WIDTH;
static uint32_t ovl7_height = DEFAULT_OVL_HEIGHT;

static uint32_t fb0_pixel_format  = FB_PIXEL_FORMAT_RGBA32;
static uint32_t ovl1_pixel_format = FB_PIXEL_FORMAT_YUV_NV12;
static uint32_t ovl2_pixel_format = FB_PIXEL_FORMAT_YUV_NV21;
static uint32_t ovl3_pixel_format = FB_PIXEL_FORMAT_RGBA32;
static uint32_t ovl4_pixel_format = FB_PIXEL_FORMAT_RGBA32;
static uint32_t ovl5_pixel_format = FB_PIXEL_FORMAT_RGBA32;
static uint32_t ovl6_pixel_format = FB_PIXEL_FORMAT_RGBA32;
static uint32_t ovl7_pixel_format = FB_PIXEL_FORMAT_RGBA32;
static uint32_t fb0_cache_policy = FB0_CACHE_POLICY_UNCACHED;

static bool ovl1_double  = false;
static bool ovl2_double  = false;
static bool ovl3_double  = false;
static bool ovl4_double  = false;
static bool ovl5_double  = false;
static bool ovl6_double  = false;
static bool ovl7_double  = false;

/*	
	The z-order. 
*/
/* cursors */
static uint32_t ovl4_layer = 15;
static uint32_t ovl5_layer = 14;
static uint32_t ovl6_layer = 13;
static uint32_t ovl7_layer = 12;
/* general overlay */
static uint32_t ovl3_layer = 11;
/* main graphics */
static uint32_t fb0_layer = 10;
/* grid screen */
static uint32_t ovl1_layer = 9;
static uint32_t ovl2_layer = 8;

module_param_named(fbwidth, fb_width, uint, 0);
MODULE_PARM_DESC(fb_width, "FB0 Horizontal resolution");
module_param_named(fbheight, fb_height, uint, 0);
MODULE_PARM_DESC(fb_height, "FB0 Vertical resolution");

module_param_named(ovl1width, ovl1_width, uint, 0);
MODULE_PARM_DESC(ovl1_width, "OVL1 Horizontal resolution");
module_param_named(ovl1height, ovl1_height, uint, 0);
MODULE_PARM_DESC(ovl1_height, "OVL1 Vertical resolution");

module_param_named(ovl2width, ovl2_width, uint, 0);
MODULE_PARM_DESC(ovl2_width, "OVL2 Horizontal resolution");
module_param_named(ovl2height, ovl2_height, uint, 0);
MODULE_PARM_DESC(ovl2_height, "OVL2 Vertical resolution");

module_param_named(ovl3width, ovl3_width, uint, 0);
MODULE_PARM_DESC(ovl3_width, "OVL3 Horizontal resolution");
module_param_named(ovl3height, ovl3_height, uint, 0);
MODULE_PARM_DESC(ovl3_height, "OVL3 Vertical resolution");

module_param_named(ovl4width, ovl4_width, uint, 0);
MODULE_PARM_DESC(ovl4_width, "OVL4 Horizontal resolution");
module_param_named(ovl4height, ovl4_height, uint, 0);
MODULE_PARM_DESC(ovl4_height, "OVL4 Vertical resolution");

module_param_named(ovl5width, ovl5_width, uint, 0);
MODULE_PARM_DESC(ovl5_width, "OVL5 Horizontal resolution");
module_param_named(ovl5height, ovl5_height, uint, 0);
MODULE_PARM_DESC(ovl5_height, "OVL5 Vertical resolution");

module_param_named(ovl6width, ovl6_width, uint, 0);
MODULE_PARM_DESC(ovl6_width, "OVL6 Horizontal resolution");
module_param_named(ovl6height, ovl6_height, uint, 0);
MODULE_PARM_DESC(ovl6_height, "OVL6 Vertical resolution");

module_param_named(ovl7width, ovl7_width, uint, 0);
MODULE_PARM_DESC(ovl7_width, "OVL7 Horizontal resolution");
module_param_named(ovl7height, ovl7_height, uint, 0);
MODULE_PARM_DESC(ovl7_height, "OVL7 Vertical resolution");

/* FB0 pixel format
 * 0 = RGB565
 * 1 = RGB888
 * 2 = RGBA32 -> default
 * 3 = BGR565 -> not supported yet
 * 4 = BGR888 -> not supported yet
 * 5 = YUV NV12 -> fb1 default
 * 6 = YUV NV21 -> fb2 default 
 */

module_param_named(fb0pixfmt, fb0_pixel_format, uint, 0);
MODULE_PARM_DESC(fb0_pixel_format, "FB0 Pixel Format");
module_param_named(ovl1pixfmt, ovl1_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl1_pixel_format, "OVL1 Pixel Format");
module_param_named(ovl2pixfmt, ovl2_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl2_pixel_format, "OVL2 Pixel Format");
module_param_named(ovl3pixfmt, ovl3_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl3_pixel_format, "OVL3 Pixel Format");
module_param_named(ovl4pixfmt, ovl4_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl4_pixel_format, "OVL4 Pixel Format");
module_param_named(ovl5pixfmt, ovl5_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl5_pixel_format, "OVL5 Pixel Format");
module_param_named(ovl6pixfmt, ovl6_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl6_pixel_format, "OVL6 Pixel Format");
module_param_named(ovl7pixfmt, ovl7_pixel_format, uint, 0);
MODULE_PARM_DESC(ovl7_pixel_format, "OVL7 Pixel Format");

module_param_named(fb0layer, fb0_layer, uint, 0);
MODULE_PARM_DESC(fb0_layer, "FB0 Layer");
module_param_named(ovl1layer, ovl1_layer, uint, 0);
MODULE_PARM_DESC(ovl1_layer, "OVL1 Layer");
module_param_named(ovl2layer, ovl2_layer, uint, 0);
MODULE_PARM_DESC(ovl2_layer, "OVL2 Layer");
module_param_named(ovl3layer, ovl3_layer, uint, 0);
MODULE_PARM_DESC(ovl3_layer, "OVL3 Layer");
module_param_named(ovl4layer, ovl4_layer, uint, 0);
MODULE_PARM_DESC(ovl4_layer, "OVL4 Layer");
module_param_named(ovl5layer, ovl5_layer, uint, 0);
MODULE_PARM_DESC(ovl5_layer, "OVL5 Layer");
module_param_named(ovl6layer, ovl6_layer, uint, 0);
MODULE_PARM_DESC(ovl6_layer, "OVL6 Layer");
module_param_named(ovl7layer, ovl7_layer, uint, 0);
MODULE_PARM_DESC(ovl7_layer, "OVL7 Layer");

module_param_named(ovl1double, ovl1_double, bool, 0);
MODULE_PARM_DESC  (ovl1double,"OVL1 double buffered?");
module_param_named(ovl2double, ovl2_double, bool, 0);
MODULE_PARM_DESC  (ovl2double,"OVL2 double buffered?");
module_param_named(ovl3double, ovl3_double, bool, 0);
MODULE_PARM_DESC  (ovl3double,"OVL3 double buffered?");
module_param_named(ovl4double, ovl4_double, bool, 0);
MODULE_PARM_DESC  (ovl4double,"OVL4 double buffered?");
module_param_named(ovl5double, ovl5_double, bool, 0);
MODULE_PARM_DESC  (ovl5double,"OVL5 double buffered?");
module_param_named(ovl6double, ovl6_double, bool, 0);
MODULE_PARM_DESC  (ovl6double,"OVL6 double buffered?");
module_param_named(ovl7double, ovl7_double, bool, 0);
MODULE_PARM_DESC  (ovl7double,"OVL7 double buffered?");

module_param_named(fb0cache, fb0_cache_policy, uint, 0);
MODULE_PARM_DESC(fb0_cache_policy, "FB0 Cache Policy");

static uint64_t now_cpu_ns(void)
{
	uint64_t ns;
	/*should be safe because there is only one cpu.*/
	/*maybe should switch to ktime_get() if that is stable. */
	ns = cpu_clock(0);
	return ns;
}

static void inline set_default_rectangles(struct bcm2708_fb *fb, struct fb_var_screeninfo *var)
{
	struct rect *s = (struct rect *)&var->reserved[0];
	struct rect *t = (struct rect *)&var->reserved[2];

	s->x = s->y = t->x = t->y = 0;
	s->w = t->w = var->xres;
	s->h = t->h = var->yres;
}

static inline int check_bounds(struct rect *r, struct fb_var_screeninfo *var)
{
	return ((r->x > var->xres) || (r->y > var->yres) ||
		((r->x + r->w) > var->xres) || ((r->y + r->h) > var->yres));
}

static int inline is_rect_set(struct bcm2708_fb *fb, struct rect *s, struct rect *t)
{
	/* Now we only fix the source and destination rectangles if framebuffer
	 * is in RGB mode using its xres and yres
	 */

	if ((s->x + s->y + s->w + s->h) && (t->x + t->y + t->w + t->h))
		goto out;

	if (fb->is_yuv) {
		bcm2708fb_error("FB(%d) YUV Framebuffer must set src/target rectangles\n", fb->fbnum);
		return 0;
	}

	set_default_rectangles(fb, &fb->fb.var);

out:
	return 1;
}

static irqreturn_t bcm2708_fb_interrupt(int irq, void *dev_id)
{
	struct bcm2708_fb *fb = (struct bcm2708_fb *)dev_id;

	/* Just declare that we got an interrupt */
	wake_up_all(&fb->wait_for_irq);

	return IRQ_HANDLED;
}

static int bcm2708fb_set_pixfmt(struct fb_var_screeninfo *var, unsigned int nonstd)
{
	switch (nonstd) {
		case 0 :
		switch (var->bits_per_pixel) {
			case 16 :
				/* RGB 565 */
				var->red.offset = 11;
				var->red.length = 5;
				var->green.offset = 5;
				var->green.length = 6;
				var->blue.offset = 0;
				var->blue.length = 5;
				var->transp.offset = 0;
				var->transp.length = 0;
				break;

			case 24 :
				/* RGB 888 */
				var->red.offset = 16;
				var->red.length = 8;
				var->green.offset = 8;
				var->green.length = 8;
				var->blue.offset = 0;
				var->blue.length = 8;
				var->transp.offset = 0;
				var->transp.length = 0;
				break;

			case 32 :
				/* RGBA 8888 */
				var->red.offset = 16;
				var->red.length = 8;
				var->green.offset = 8;
				var->green.length = 8;
				var->blue.offset = 0;
				var->blue.length = 8;
				var->transp.offset = 24;
				var->transp.length = 8;
				break;

			default :
				bcm2708fb_error("Incorrect standard pixel format\n");
				return -EINVAL;

		}

		break;

		case 1 :
		switch (var->bits_per_pixel) {
			case 16 :
				/* BGR 565 */
				var->blue.offset = 11;
				var->blue.length = 5;
				var->green.offset = 5;
				var->green.length = 6;
				var->red.offset = 0;
				var->red.length = 5;
				var->transp.offset = 0;
				var->transp.length = 0;
				break;

			case 24 :
				/* BGR 888 */
				var->blue.offset = 16;
				var->blue.length = 8;
				var->green.offset = 8;
				var->green.length = 8;
				var->red.offset = 0;
				var->red.length = 8;
				var->transp.offset = 0;
				var->transp.length = 0;
				break;

			default :
				bcm2708fb_error("Incorrect standard pixel format\n");
				return -EINVAL;
		}
        
		break;

		default :
			bcm2708fb_error("nonstd is set incorrectly\n");
			return -EINVAL;
	}

	return 0;
}

/* returns 1 if we are yet to get an ack from the device for the last update
 * else returns 0
 **/
static int is_vsync_pending(struct bcm2708_fb *fb)
{
	uint32_t control, status;
	uint32_t vc_update_count;
	int pending;

	control = readl(fb->reg_base + FB_CONTROL);
	status = readl(fb->reg_base + FB_STATUS);

	if (control & CTRL_UPDT_EN_MASK) {
		/* We have an update pending */
		if (status & STAT_UPDT_EN_MASK) {
			/* verify update count */
			vc_update_count = (status & STAT_UPDATE_COUNT_MASK) >> STAT_UPDATE_COUNT_SHIFT;
			pending = (fb->update_count & 0xff) - vc_update_count;
		} else {
			pending = 1;
		}

	} else {
		/* There are no updates pending at the moment */
		pending = 0;
	}

	return !!pending;
}

/* Waits for last fb update to finish
 * Returns 0 if success, 1 if times out
 */
static int wait_for_vsync(struct bcm2708_fb *fb)
{
	int ret;

	/* Wait with timeout of 1/2 second and check for status update bit
	 * to be set
	 **/
	ret = wait_event_timeout(fb->wait_for_irq,
			(!is_vsync_pending(fb)), HZ / 2);
	if (!ret) {
		bcm2708fb_error("Wait for vsync timed out ..\n");
		/* Something has gone wrong */
		bcm2708fb_error("May be update count is out of sync !!\n");
		bcm2708fb_error("FB(%d) control = 0x%08x, status = 0x%08x update count = 0x%08x!!\n",
					fb->fbnum, readl(fb->reg_base + FB_CONTROL),
					readl(fb->reg_base + FB_STATUS), fb->update_count);
	}

	return !ret;

}

#ifdef CONFIG_ANDROID_POWER
static void bcm2708_fb_early_suspend(android_early_suspend_t *h)
{
	struct bcm2708_fb *fb = container_of(h, struct bcm2708_fb, early_suspend);
	bcm2708fb_info("TODO: BRCM fb early suspend ...\n");
}

static void bcm2708_fb_late_resume(android_early_suspend_t *h)
{
	struct bcm2708_fb *fb = container_of(h, struct bcm2708_fb, early_suspend);
	bcm2708fb_info("TODO: BRCM fb late resume ...\n");
}
#endif

static int var_to_pixel_format(struct fb_var_screeninfo *var, unsigned int nonstd)
{
	int ret = 2; /* RGBA32 by default */

	switch (nonstd) {
		case 0 :
		switch (var->bits_per_pixel) {
			case 16 :
				/* RGB 565 */
				ret = FB_FORMAT_RGB565;
				break;

			case 24 :
				/* RGB 888 */
				ret = FB_FORMAT_RGB888;
				break;

			case 32 :
				/* RGBA 8888 */
				ret = FB_FORMAT_RGBA32;
				break;

			default :
				bcm2708fb_error("Incorrect standard pixel format\n");
		}

		break;

		case 1 :
		switch (var->bits_per_pixel) {
			case 16 :
				/* BGR 565 */
				ret = FB_FORMAT_BGR565;
				break;

			case 24 :
				/* BGR 888 */
				ret = FB_FORMAT_BGR888;
				break;

			default :
				bcm2708fb_error("Incorrect standard pixel format\n");
		}

		break;

		default :
			bcm2708fb_error("nonstd is set incorrectly\n");
	}

	return ret;
}

static int enable_fb_device(struct bcm2708_fb *fb, struct fb_var_screeninfo *var)
{
	int ret = 0;
	uint32_t control, status;
	int pixel_format;
	uint32_t opacity;
	uint32_t ready_mask = STAT_EN_MASK | STAT_DISP_MASK;

	/* 
	 * Setting framebuffer resolution, this will never change
	 */

	writel(var->xres, fb->reg_base + FB_WIDTH);
	writel(var->yres, fb->reg_base + FB_HEIGHT);

	/* Setup pixel format, this will also never change
	 * once driver has started
	 */
	pixel_format = readl(fb->reg_base + FB_SET_FORMAT);
	pixel_format &= ~(PIXEL_FORMAT_MASK);
	if (!fb->is_yuv) {
		if (unlikely(fb0_pixel_format >= FB_PIXEL_FORMAT_BGR565))
			pixel_format |= var_to_pixel_format(var, 1);
		else
			pixel_format |= var_to_pixel_format(var, 0);

		if (!strcmp(fb->dev_name, "bcm2835_OVL3")) {
			writel(ovl3_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL4")) {
			writel(ovl4_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL5")) {
			writel(ovl5_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL6")) {
			writel(ovl6_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL7")) {
			writel(ovl7_layer, fb->reg_base + FB_TARGET_LAYER);
		}
		opacity = var->reserved[4];
	} else {
		if (!strcmp(fb->dev_name, "bcm2835_OVL1")) {
			pixel_format |= FB_FORMAT_YUVNV12;
			writel(ovl1_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL2")) {
			pixel_format |= FB_FORMAT_YUVNV21;
			writel(ovl2_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL3")) {
			writel(ovl3_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL4")) {
			writel(ovl4_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL5")) {
			writel(ovl5_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL6")) {
			writel(ovl6_layer, fb->reg_base + FB_TARGET_LAYER);
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL7")) {
			writel(ovl7_layer, fb->reg_base + FB_TARGET_LAYER);
		}
		/* default ITU-R BT.601-5 */
		pixel_format |= (COLOUR_SPACE_601 << COLOUR_SPACE_SHIFT);

		/* Set to max */
		opacity = 255;
	}

	writel(pixel_format, fb->reg_base + FB_SET_FORMAT);
	writel((opacity & 255), fb->reg_base + FB_SET_OPACITY);

	writel((var->yres_virtual/var->yres), fb->reg_base + FB_SET_NUM_BUFFERS);

	/* Set z-order, src (x, y, w, h)
	 * target(x, y, w, h) and target display number
	 */
	writel(0, fb->reg_base + FB_SRC_X);
	writel(0, fb->reg_base + FB_SRC_Y);
	writel(var->xres, fb->reg_base + FB_SRC_WIDTH);
	writel(var->yres, fb->reg_base + FB_SRC_HEIGHT);

	writel(0, fb->reg_base + FB_TARGET_X);
	writel(0, fb->reg_base + FB_TARGET_Y);
	writel(var->xres, fb->reg_base + FB_TARGET_WIDTH);
	writel(var->yres, fb->reg_base + FB_TARGET_HEIGHT);

	/* z-order = 10  by default*/
	if (fb->fbnum == 0)
		writel(fb0_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 1)
		writel(ovl1_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 2)
		writel(ovl2_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 3)
		writel(ovl3_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 4)
		writel(ovl4_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 5)
		writel(ovl5_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 6)
		writel(ovl6_layer, fb->reg_base + FB_TARGET_LAYER);
	else if(fb->fbnum == 7)
		writel(ovl7_layer, fb->reg_base + FB_TARGET_LAYER);
	else
		bcm2708fb_warning("Incorrect fbnum (%d), wont set framebuffer layer\n", fb->fbnum);

	writel(FRAMEBUFFER_TARGET_DISPLAY, fb->reg_base + FB_SET_DISPLAY);

	/* Write control register bits ...*/
	control = readl(fb->reg_base + FB_CONTROL);
	control &= ~(CTRL_BUFF_IDX_MASK);
	control |= ((1 << CTRL_ENABLE_SHIFT) | (1 << CTRL_DISP_EN_SHIFT)); 

#ifdef DEBUG_UPDATE_COUNT
	control &= ~(CTRL_UPDT_CNT_DBG_MASK);
	control |= (1 << CTRL_UPDT_CNT_DBG_SHIFT);
#endif
	writel(control, fb->reg_base + FB_CONTROL);

	/* ring the doorbell */
	ipc_notify_vc_event(fb->irq);

	/* Wait for Framebuffer to get enabled .. */
	ret = wait_event_timeout(fb->wait_for_irq,
			(((status = readl(fb->reg_base + FB_STATUS)) & ready_mask) == ready_mask),
			HZ / 2);
	if (!ret) {
		bcm2708fb_warning("Wait for fb interrupt timed out ..\n");
		ret = -ETIMEDOUT;
		goto out;
	} else {
		ret = 0;
	}

	fb->update_count = (status & STAT_UPDATE_COUNT_MASK) >> STAT_UPDATE_COUNT_SHIFT;
	fb->phys_fbbase = readl(fb->reg_base + FB_GET_BUFFER_PHYS);
	if (!fb->phys_fbbase) {
		bcm2708fb_error("FB phys memory pointer is NULL\n");
		ret = -ENODEV;
		goto out;
	}

	/* Convert VC phys addr to Kernel phys addr */
	fb->phys_fbbase = __bus_to_phys(fb->phys_fbbase);

out:
	return ret;
}

static int disable_fb_device(struct bcm2708_fb *fb)
{
	uint32_t control;
	uint32_t disable_mask = (STAT_EN_MASK | STAT_DISP_MASK);
	int ret;

	/* Write control register bits ...*/
	control = readl(fb->reg_base + FB_CONTROL);
	control &= ~CTRL_ENABLE_MASK; 
	control &= ~CTRL_UPDT_EN_MASK;
#ifdef DEBUG_UPDATE_COUNT
	control &= ~CTRL_UPDT_CNT_DBG_MASK;
#endif
	writel(control, fb->reg_base + FB_CONTROL);

	ipc_notify_vc_event(fb->irq);
	ret = wait_event_timeout(fb->wait_for_irq,
			(!(readl(fb->reg_base + FB_STATUS) & disable_mask)), HZ / 2);
	if (!ret) {
		bcm2708fb_error("Disable BCM2708 FB device timed out \n");
		ret = -ETIMEDOUT;
		goto out;
	} else {
		ret = 0;
	}

	bcm2708fb_debug("BCM2708 FB device disabled successfully\n");
out:
	return ret;
}

static void init_yuv_fixed_info(struct bcm2708_fb *fb, uint32_t width,
			uint32_t bpp, struct fb_ops *ops) 
{
	struct fb_info *info = &fb->fb;

	fb->is_yuv		= 1;
	if (ops)
		info->fbops	= ops;
	info->pseudo_palette	= NULL;
	info->flags		= FBINFO_FLAG_DEFAULT;
	info->node		= -1;
	info->fix.type		= FB_TYPE_PACKED_PIXELS;
	info->fix.xpanstep	= 0;
	info->fix.ypanstep	= 1;
	info->fix.line_length	= width * bpp/8;
}

static void __init
init_yuv_fb_info(struct bcm2708_fb *fb, uint32_t width, uint32_t height,
		uint32_t pixfmt, struct fb_ops *ops)
{
	struct fb_info *info = &fb->fb;
	struct yuv_pixfmt *fmt = (struct yuv_pixfmt *)&info->var.reserved[4];

	if ((width & 15) || (height & 15)) {
		uint32_t new_width, new_height;
		bcm2708fb_warning("Framebuffer width and height has to be alligned to 16\n");
		new_width = (width + 15) & ~15;
		new_height = (height + 15) & ~15;
		bcm2708fb_warning("Alligning (%dx%d) to be (%dx%d)\n", width, height, new_width, new_height);
		width = new_width;
		height = new_height;
	}

	init_yuv_fixed_info(fb, width, 8, ops);

	/* Clear rectangles and format space */
	memset(&info->var.reserved, 0, sizeof(info->var.reserved));

	info->var.xres		= width;
	info->var.yres		= height;
	info->var.xres_virtual	= width;
	info->var.yres_virtual	= height;
	info->var.bits_per_pixel = 8;
	/* This is a non-std framebuffer, and set default pixel format
	 * and colour space
	 **/
	info->var.nonstd	= 1;

	fmt->fmt		= pixfmt;
	fmt->colour_space	= CSC_ITUR_BT_601_5;

	info->var.activate	= FB_ACTIVATE_NOW;
	info->var.height	= height;
	info->var.width		= width;
	info->var.vmode		= FB_VMODE_NONINTERLACED;

}

static void init_rgb_fixed_info(struct bcm2708_fb *fb, uint32_t width,
				uint32_t bpp, struct fb_ops *ops)
{
	struct fb_info *info = &fb->fb;
	fb->is_yuv		= 0;
	if (ops)
		info->fbops	= ops;
	info->flags		= FBINFO_FLAG_DEFAULT;
	info->pseudo_palette	= fb->cmap;
	info->fix.type		= FB_TYPE_PACKED_PIXELS;
	info->fix.visual	= FB_VISUAL_TRUECOLOR;
	info->fix.line_length	= width * bpp/8;
	info->fix.accel		= FB_ACCEL_NONE;
	info->fix.ypanstep	= 1;

}

static void __init
init_rgb_fb_info(struct bcm2708_fb *fb, uint32_t width, uint32_t height, 
                 bool double_buffered, struct fb_ops *ops)
{
	uint32_t bpp;
	struct fb_info *info = &fb->fb;

	switch (fb0_pixel_format) {
		case 0:
		case 3:
			bpp = 16;
			break;
		case 1:
		case 4:
			bpp = 24;
			break;
		case 2:
			bpp = 32;
			break;
		default:
			bcm2708fb_warning("Setting to default 32bpp\n");
			bpp = 32;
	}

	if ((width & 15) || (height & 15)) {
		uint32_t new_width, new_height;
		bcm2708fb_warning("Framebuffer width and height has to be alligned to 16\n");
		new_width = (width + 15) & ~15;
		new_height = (height + 15) & ~15;
		bcm2708fb_warning("Alligning (%dx%d) to be (%dx%d)\n", width, height, new_width, new_height);
		width = new_width;
		height = new_height;
	}

	init_rgb_fixed_info(fb, width, bpp, ops);

	info->var.xres		= width;
	info->var.yres		= height;
	info->var.xres_virtual	= width;
    if (double_buffered)
        info->var.yres_virtual	= height * 2;
    else        
        info->var.yres_virtual	= height;
	info->var.bits_per_pixel = bpp;
	info->var.activate	= FB_ACTIVATE_NOW;
	info->var.height	= height;
	info->var.width		= width;
	info->var.nonstd = 0; 	/* this is set only for yuv framebuffers */

	/* set src/target rectangles to full screen */
	set_default_rectangles(fb, &info->var);

	/* set max opacity value */
	info->var.reserved[4] = 255;

	if (unlikely(fb0_pixel_format >= FB_PIXEL_FORMAT_BGR565))
		bcm2708fb_set_pixfmt(&info->var, 1);
	else
		bcm2708fb_set_pixfmt(&info->var, 0);

}

static int switch_fb_device_format(struct bcm2708_fb *fb, struct fb_var_screeninfo *var,
		uint32_t *framesize)
{
	struct yuv_pixfmt *fmt = (struct yuv_pixfmt *)&var->reserved[4];
	int ret = 0;

	if (fb->is_yuv) {
		init_rgb_fixed_info(fb, var->xres, var->bits_per_pixel, NULL);

		*framesize = var->xres * var->yres_virtual * var->bits_per_pixel/8;

        /* Set max opacity value */
        var->reserved[4] = 255;
		var->yoffset = var->xoffset = 0;

		bcm2708fb_set_pixfmt(var, 0);

		set_default_rectangles(fb, var);

	} else {

		if (var->yres_virtual != var->yres) {
			bcm2708fb_error("Double buffering not supported on YUV framebuffers\n");
			bcm2708fb_error("RGB->YUV switch will not be allowed ..\n");
			ret = -EINVAL;
			goto out;
		}
		
		init_yuv_fixed_info(fb, var->xres, var->bits_per_pixel, NULL);

		/* Clear rectangles and format space */
		memset(&var->reserved, 0, sizeof(var->reserved));

		fmt->fmt = fb->fbnum == 1 ? FB_PIXEL_FORMAT_YUV_NV12 : FB_PIXEL_FORMAT_YUV_NV21;
		fmt->colour_space = CSC_ITUR_BT_601_5;

		*framesize = ((3 * var->xres * var->yres_virtual * var->bits_per_pixel/8) >> 1);
	}

	/* 
	 * set src/target rectangles to full screen
	 * x/y offset to start of the buffers
	 */

	var->activate = FB_ACTIVATE_NOW;
	var->xoffset = var->yoffset = 0;

out:
	return ret;
}

static int bcm2708_fb_open(struct fb_info *info, int user)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	size_t framesize;
	int ret = 0;

	/* No framebuffer console on overlay device */
	if (user == 0) {
		bcm2708fb_warning("No framebuffer console on overlay framebuffers\n");
		ret = -ENODEV;
		goto out;
	}

	if (atomic_read(&fb->usage) >= 1) {
		/* Its already open, dont do anything and return */
		bcm2708fb_debug("FB(%d) already opened .. doing nothing, usage(%d)\n", 
						fb->fbnum, atomic_read(&fb->usage));
		atomic_inc(&fb->usage);
		goto done;
	}

	atomic_inc(&fb->usage);

	/* Set all rectangles to match full screen again if we are
	 * opening the framebuffer is RGB mod
	 */

	if (!fb->is_yuv) {
		set_default_rectangles(fb, &info->var);
		/* Set maximum opacity value */
		info->var.reserved[4] = 255;
    }

	ret = enable_fb_device(fb, &info->var);
	if (ret) {
		bcm2708fb_error("Failed to enable overlay device\n");
		goto err_enable_fb_device;
	}

	if (fb->is_yuv)
		framesize = ((3 * fb->fb.var.xres * fb->fb.var.yres_virtual * fb->fb.var.bits_per_pixel/8) >> 1);
	else
		framesize = fb->fb.var.xres * fb->fb.var.yres_virtual * fb->fb.var.bits_per_pixel/8;

	info->fix.smem_start = fb->phys_fbbase;
	info->fix.smem_len = framesize;

	bcm2708fb_debug("Requesting and mapping resource (0x%08x)-(0x%08x)\n", fb->phys_fbbase, fb->phys_fbbase + framesize);
	info->screen_base = ioremap(fb->phys_fbbase, framesize); 
	if (info->screen_base == NULL) {
		ret = -ENOMEM;
		bcm2708fb_error("Unable to ioremap fb memory resource\n");
		goto err_ioremap_fbmem;
	}
	bcm2708fb_debug("FB(%d) : IORMAP -> (0x%08x)\n", fb->fbnum, info->screen_base);

	bcm2708fb_debug("Framebuffer starts at phys[0x%08x], and virt[0x%08x]\n",
			fb->phys_fbbase, (uint32_t)fb->fb.screen_base);

done:
	bcm2708fb_debug("bcm2708 FB (%d) device open [ %s ], usage(%d)\n", fb->fbnum, ret ? "FAILED" : "OK", atomic_read(&fb->usage));

	return 0;

err_ioremap_fbmem:
	if (disable_fb_device(fb))
		bcm2708fb_error("Failed to disable FB(%d)\n", fb->fbnum);
err_enable_fb_device:
	atomic_dec(&fb->usage);
out:
	return ret;
}

static int
bcm2708_fb_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long start;
	uint32_t len = info->fix.smem_len;
	unsigned long off = vma->vm_pgoff << PAGE_SHIFT;
	//struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	int ret = -EINVAL;

	if (off > len)
		goto out;

	if ((vma->vm_end - vma->vm_start) > (len - off))
		goto out;

	start = info->fix.smem_start;
	start &= PAGE_MASK;
	start += off;

	/* Change prot bits to reflect cache write-through/write-back */
	switch (fb0_cache_policy) {
	case FB0_CACHE_POLICY_WRITETHROUGH :
		vma->vm_page_prot = pgprot_writethrough(vma->vm_page_prot);
		break;
#if 0
	case FB0_CACHE_POLICY_WRITEBACK :
		vma->vm_page_prot = pgprot_writeback(vma->vm_page_prot);
		break;
#endif
	default:
		bcm2708fb_warning("We should not have come here if FB0 cache policy was uncached\n");
		bcm2708fb_warning("Setting the cache policy to UNCACHED\n");
		vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
	}

	/* VM_IO = I am a memory-mapped device, dont core-dump me and
	 * VM_RESERVED = please dont swap me out .. */
	vma->vm_flags |= VM_IO | VM_RESERVED;

	ret = io_remap_pfn_range(vma, vma->vm_start, start >> PAGE_SHIFT,
			vma->vm_end - vma->vm_start, vma->vm_page_prot);
out:
	return ret;
}

/* We are protected against another ioctl by info->lock taken in do_fb_ioctl()
 * so, we are safe against pan_display() for vsync check
 **/

static int bcm2708_fb_ioctl(struct fb_info *info, unsigned int cmd,
			unsigned long arg)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	/* uint64_t ts1, ts2, latency_us; */
	int ret;

	switch (cmd) {
	case FBIO_WAITFORVSYNC :
		/* ts1 = now_cpu_ns(); */

		if (fb->is_suspended == FB_BLANK_POWERDOWN) {
			bcm2708fb_error("FB (%d) Display is powered down no need to wait for vsync!\n", fb->fbnum);
			ret = -ENODEV;
			goto out;
		}

		if (is_vsync_pending(fb)) {
			ret = wait_for_vsync(fb);
			if (ret) {
				bcm2708fb_error("Wait for vsync timed out\n");
				ret = -ETIMEDOUT;
			}
		} else {
			ret = 0;
		}

		/* ts2 = now_cpu_ns();

		latency_us = (ts2 - ts1)/1000;

		if(latency_us > 20000 && printk_ratelimit())
			bcm2708fb_error("fb=%u: ioctl vsync took %llu usec\n", fb->fbnum,latency_us);
		*/

		break;
	default :
		bcm2708fb_error("Invalid Ioctl command on device\n");
		ret = -ENOIOCTLCMD;
	}
out:
	return ret;
}

static int bcm2708_fb_blank(int blank_mode, struct fb_info *info)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	uint32_t control, status;
	int ret = 0;

	if (fb->is_suspended == blank_mode) {
		bcm2708fb_info("FB(%d) blank mode is already set to (%d), doing nothing\n", fb->fbnum, blank_mode);
		goto out;
	}

	status = readl(fb->reg_base + FB_STATUS);
	control = readl(fb->reg_base + FB_CONTROL);

	switch(blank_mode) {
	/* For blanking disable displaying in control register */
	case FB_BLANK_POWERDOWN:
	case FB_BLANK_NORMAL:

		if (!(status & STAT_DISP_MASK)) {
			bcm2708fb_info("FB(%d) Display is not enabled ..\n", fb->fbnum);
			goto out;
		}
		/* Set opacity to 0 and update the framebuffer again with
		 * all other settings intact
		 **/
		writel(0, fb->reg_base + FB_SET_OPACITY);

		fb->update_count++;

		/* Never round off to zero, bad things will happen
		 * since the code above checks for update_count before it
		 * checks for vsynd pending status
		 *
		 * Update count in status register also behaves the same way
		 * now
		 */

		if (!fb->update_count)
			fb->update_count++;

		control |= (1 << CTRL_UPDT_EN_SHIFT);
#ifdef DEBUG_UPDATE_COUNT
		control &= ~(CTRL_UPDATE_COUNT_MASK);
		control |= (fb->update_count << CTRL_UPDATE_COUNT_SHIFT);
#endif
		writel(control, fb->reg_base + FB_CONTROL);

		ipc_notify_vc_event(fb->irq);

		break;

	case FB_BLANK_UNBLANK:

		if (!(status & STAT_DISP_MASK)) {
			bcm2708fb_info("FB(%d) display is not enabled...\n", fb->fbnum);
			goto out;
		}

		/* Set opacity 255 for yuv framebuffer and var.reserved[4] for
		 * RGB and update the framebuffer again with
		 * all other settings intact
		 **/

		if (fb->is_yuv)
			writel(255, fb->reg_base + FB_SET_OPACITY);
		else
			writel(((uint32_t)info->var.reserved[4] & 255), fb->reg_base + FB_SET_OPACITY);

		fb->update_count++;

		/* Never round off to zero, bad things will happen
		 * since the code above checks for update_count before it
		 * checks for vsynd pending status
		 *
		 * Update count in status register also behaves the same way
		 * now
		 */

		if (!fb->update_count)
			fb->update_count++;

		control |= (1 << CTRL_UPDT_EN_SHIFT);
#ifdef DEBUG_UPDATE_COUNT
		control &= ~(CTRL_UPDATE_COUNT_MASK);
		control |= (fb->update_count << CTRL_UPDATE_COUNT_SHIFT);
#endif
		writel(control, fb->reg_base + FB_CONTROL);

		ipc_notify_vc_event(fb->irq);

		break;

	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	default:
		ret = -EINVAL;
		goto out;
	}

	fb->is_suspended = blank_mode;
out:
	return ret;
}

static int bcm2708_fb_release(struct fb_info *info, int user)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	int ret = 0;

	if(atomic_read(&fb->usage) < 1)
		goto out;

	if (atomic_dec_and_test(&fb->usage)) {
		ret = disable_fb_device(fb);
		if (ret)
			bcm2708fb_error("Failed to disable FB(%d)\n", fb->fbnum);

		bcm2708fb_debug("FB(%d) : IOUMAP -> (0x%08x)\n", fb->fbnum, info->screen_base);
		iounmap(fb->fb.screen_base);
	}

out:
	bcm2708fb_debug("FB(%d) release done with usage(%d)\n", fb->fbnum, atomic_read(&fb->usage));
	return ret;
}

static int bcm2708_fb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
#define CHANGE_RGB_YUV_FMT	(1 << 0)
#define CHANGE_FB_Y_VIRTUAL	(1 << 1)
#define CHANGE_FB_RESOLUTION	(1 << 2)
#define ACTIONS_MASK	(0x00000007)

	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	struct yuv_pixfmt *fmt = (struct yuv_pixfmt *)&var->reserved[4];
	unsigned int actions = 0;
	unsigned int framesize = 0;
	int ret = -EINVAL;

	if (((var->rotate & 1) != (info->var.rotate & 1)) ||
			(var->xres_virtual != info->var.xres_virtual)) {
		bcm2708fb_error("FB(%d) Rotation / changing xres_virtual is not allowed\n", fb->fbnum);
		goto out;
	}

	if (info->var.nonstd != var->nonstd) {
		/* Switching from YUV -> RGB formats */
		bcm2708fb_debug("FB(%d) Trying to switch from (%s) -> (%s) Format\n", fb->fbnum,
				fb->is_yuv ? "yuv" : "rgb", fb->is_yuv ? "rgb" : "yuv");
		actions |= CHANGE_RGB_YUV_FMT;
	}

	if (info->var.yres_virtual != var->yres_virtual)
		actions |= CHANGE_FB_Y_VIRTUAL;

	if ((var->xres != info->var.xres) || (var->yres != info->var.yres))
		actions |= CHANGE_FB_RESOLUTION;

	if ((fb->fbnum == 0) && actions) {
		bcm2708fb_error("FB(%d) Does not support any change to its var_screeninfo\n", fb->fbnum);
		goto out;
	}

	if (actions & CHANGE_FB_Y_VIRTUAL) {
		/* if there's no change of format, Y_VIRTUAL changes should be
		 * (2 * yres) or (yres) and it should be RGB framebuffer
		 */
		if (!(actions & CHANGE_RGB_YUV_FMT)) {
			if ((var->yres_virtual != info->var.yres) && (var->yres_virtual != 2 * info->var.yres)) {
				bcm2708fb_error("Invalid yres_virtual change (%d)->(%d) with action(0x%08x)\n",
								info->var.yres_virtual, var->yres_virtual, actions);
				goto out;
			}

			if (fb->is_yuv) {
				bcm2708fb_warning("FB(%d) Just changin y_virtual for YUV framebuffer is not allowed\n", fb->fbnum);
				goto out;
			}

		}

		bcm2708fb_debug("Changing yres_virtual from (%d) to (%d)\n", info->var.yres_virtual, var->yres_virtual);
	}

	if (actions & CHANGE_FB_RESOLUTION) {
		if (!(actions & CHANGE_RGB_YUV_FMT)) { /* Resolution cannot be changed without changing the framebuffer format as of now */
			bcm2708fb_error("fb_check_var_failed : No resolution change is allowed w/o changing the pixel format\n");
			goto out;
		}

		if ((var->xres > 4096) || (var->yres > 4096)) {
			bcm2708fb_error("FB(%d) Maximum XRESxYRES while changing format is 4096x4096\n", fb->fbnum);
			bcm2708fb_error("FB(%d) Cant set resolution to (%d)x(%d)\n", fb->fbnum, var->xres, var->yres);
			goto out;
		}
	}

	if (fb->is_yuv && !(actions & CHANGE_RGB_YUV_FMT)) {
		if (var->yoffset) {
			bcm2708fb_error("FB(%d) Only single buffer is supported!\n", fb->fbnum);
			goto out;
		}

		if ((fmt->fmt < FB_PIXEL_FORMAT_YUV_NV12) || (fmt->fmt > FB_PIXEL_FORMAT_YUV_NV21) ||
				(fmt->colour_space < CSC_ITUR_BT_601_5) || (fmt->colour_space > CSC_JPEG_JIF)) {
			bcm2708fb_error("Pixel Format (fmt = %i, colour = %i) is invalid\n",
					fmt->fmt, fmt->colour_space);
			goto out;
		}
	}

	ret = 0;

	if (actions & ACTIONS_MASK) {

		/* 1. Disable Framebuffer
		 * 2. If change format is set, then change framebuffer format
		 *    from YUV->RGB and vice versa
		 * 3. Enable framebuffer (with number of buffers set according
		 *    to new yres_virtual)
		 */
		bcm2708fb_debug("FB(%d) : IOUMAP -> (0x%08x)\n", fb->fbnum, info->screen_base);

		iounmap(info->screen_base);

		ret = disable_fb_device(fb);
		if (ret) {
			bcm2708fb_error("Failed to disable overlay device\n");
			bcm2708fb_error("FB Format change failed !!\n");
			goto out;
		}

		if (actions & CHANGE_RGB_YUV_FMT) {
			ret = switch_fb_device_format(fb, var, &framesize);
			if (ret) {
				bcm2708fb_error("Switching FB format failed, restoring old format\n");
				memcpy(var, &info->var, sizeof(*var));
				/*
				 * calculate framesize for the re-ioremap below
				 */
				if (fb->is_yuv)
					framesize = ((3 * var->xres * var->yres_virtual * var->bits_per_pixel/8) >> 1);
				else
					framesize = var->xres * var->yres_virtual * var->bits_per_pixel/8;
			}
		} else if (actions & CHANGE_FB_Y_VIRTUAL) {
			/* just calculate new framesize after y_virtual change
			 * for RGB framebuffer, as YUV should not even come
			 * here
			 */
			framesize = var->xres * var->yres_virtual * var->bits_per_pixel/8;
		}
	
		ret = enable_fb_device(fb, var);
		if (ret) {
			bcm2708fb_error("Failed to enable fb (%d)\n", fb->fbnum);
			bcm2708fb_error("Try closing - reopening the device again\n");
			goto out;
		}

		info->fix.smem_start = fb->phys_fbbase;
		info->fix.smem_len = framesize;
		bcm2708fb_debug("Requesting and mapping resource (0x%08x)-(0x%08x)\n", fb->phys_fbbase, fb->phys_fbbase + framesize);

		info->screen_base = ioremap(fb->phys_fbbase, framesize); 
		if (info->screen_base == NULL) {
			ret = -ENOMEM;
			bcm2708fb_error("Unable to ioremap fb memory resource\n");
			disable_fb_device(fb);
			goto out;
		}
		bcm2708fb_debug("FB(%d) : IORMAP -> (0x%08x)\n", fb->fbnum, info->screen_base);

		bcm2708fb_debug("Framebuffer starts at phys[0x%08x], and virt[0x%08x]\n",
				fb->phys_fbbase, (uint32_t)fb->fb.screen_base);

		/* Cheat with the user here a bit, copy our var structure into
		 * the one passed by the user, so fb subsystem copies it back
		 * into info->var and we are good with yoffset = 0 and
		 * rectangles initialised to their default values
		 */
	} 

out:
	return ret;
}

static int bcm2708_fb_set_par(struct fb_info *info)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	if(fb->rotation != fb->fb.var.rotate) {
		bcm2708fb_warning("Rotation is not supported yet !\n");
		return -EINVAL;
	}

	return 0;
}

static inline u32 convert_bitfield(int val, struct fb_bitfield *bf)
{
	unsigned int mask = (1 << bf->length) - 1;

	return (val >> (16 - bf->length) & mask) << bf->offset;
}

static int
bcm2708_fb_setcolreg(unsigned int regno, unsigned int red, unsigned int green,
		 unsigned int blue, unsigned int transp, struct fb_info *info)
{
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);

	if (regno < 16) {
		fb->cmap[regno] = convert_bitfield(transp, &fb->fb.var.transp) |
				  convert_bitfield(blue, &fb->fb.var.blue) |
				  convert_bitfield(green, &fb->fb.var.green) |
				  convert_bitfield(red, &fb->fb.var.red);
		return 0;
	}
	else {
		return 1;
	}
}

/* Here we are using var->reserved[5] to read src/target(x,y,w,h)
 * They should be written by the application in following format
 *
 *               31---------------15----------------0
 *                |                |                |
 * reserved[0] -> |     src_y      |      src_x     |
 * reserved[1] -> |     src_h      |      src_w     |
 * reserved[2] -> |    target_y    |     target_x   |
 * reserved[3] -> |    target_h    |     target_w   |
 *                -----------------------------------
**/


static int bcm2708_fb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	unsigned long irq_flags;
	int ret;
	struct bcm2708_fb *fb = container_of(info, struct bcm2708_fb, fb);
	uint32_t buff_idx, control, pixel_format;
	struct rect *src_rect = (struct rect *)&var->reserved[0];
	struct rect *target_rect = (struct rect *)&var->reserved[2];
	struct yuv_pixfmt *fmt;

	
	if (fb->is_suspended == FB_BLANK_POWERDOWN) {
		bcm2708fb_error("FB (%d) Display is powered down, the update wont go through!\n", fb->fbnum);
		ret = -ENODEV;
		goto out;

	}

	if (!(readl(fb->reg_base + FB_STATUS) & (STAT_EN_MASK | STAT_DISP_MASK))) {
		bcm2708fb_error("FB (%d) Display update arrived without framebuffer enable/display set\n", fb->fbnum);
		ret = -ENODEV;
		goto out;
	}

	/* We are here only if yoffset is '0' or 'yres',
	 * so if yoffset = 0, update first buffer or update second
	 *
	 * We should also shout out loud if we are updating a yuv framebuffer
	 * from second buffer, as its not allowed at the moment
	 */

	BUG_ON(fb->is_yuv && var->yoffset);

	buff_idx = var->yoffset ? 1 : 0;

	/* if source and destination rectangles are set then check bounds
	 * if not the function will set default rectangles to match
	 * fullscreen update
	 **/

	if (is_rect_set(fb, src_rect, target_rect)) {
		if (check_bounds(src_rect, var)) {
			bcm2708fb_error("src/target rectangles in var screeninfo are out of bounds\n");
			bcm2708fb_error("src rect (%i, %i, %i, %i)\n", src_rect->x, src_rect->y,
					src_rect->w, src_rect->h);
			bcm2708fb_error("target rect (%i, %i, %i, %i)\n", target_rect->x, target_rect->y,
					target_rect->w, target_rect->h);
			bcm2708fb_error("This will not be updated on the screen\n");
			ret = -EINVAL;
			goto out;
		}
	} else {
		bcm2708fb_error("FB(%d) must set src/target rectangles for updates\n", fb->fbnum);
		ret = -EINVAL;
		goto out;
	}

	/* check for vsync only if we have already sent an update */
	if (fb->update_count && is_vsync_pending(fb)) {
		bcm2708fb_debug("FB update is still pending, now waiting for it to finish\n");
		bcm2708fb_debug("control = 0x%08x, status = 0x%08x update count (0x%02x) !!\n",
				readl(fb->reg_base + FB_CONTROL), readl(fb->reg_base + FB_STATUS), fb->update_count);

		ret = wait_for_vsync(fb);
		if (ret) {
			bcm2708fb_error("Wait for vsync timed out\n");
			ret = -ETIMEDOUT;
			goto out;
		}
	}

#if 0
	/* IF RGB double buffered framebuffer, and we are doing write-back
	 * caches, then try and find the vma of mmaped region and flush
	 * the cache range associated with this update
	 */
	if (!fb->is_yuv && (fb0_cache_policy == FB0_CACHE_POLICY_WRITEBACK)) {
		/* FB memory does not have a struct page allocated with it,
		 * so we need to ue dmac_* cache flush API to flush the cache
		 * range associated with this update using process virtual
		 * addresses. I think dmac_cache_flush() fits the bill here
		 */
	}
#endif
	spin_lock_irqsave(&fb->lock, irq_flags);

	/* Clear update bit in control register */
	control = readl(fb->reg_base + FB_CONTROL);
	control &= ~(CTRL_UPDT_EN_MASK);
	writel(control, fb->reg_base + FB_CONTROL);

	/* Set up source and target rectangles */
	writel(src_rect->x, fb->reg_base + FB_SRC_X);
	writel(src_rect->y, fb->reg_base + FB_SRC_Y);
	writel(src_rect->w, fb->reg_base + FB_SRC_WIDTH);
	writel(src_rect->h, fb->reg_base + FB_SRC_HEIGHT);

	writel(target_rect->x, fb->reg_base + FB_TARGET_X);
	writel(target_rect->y, fb->reg_base + FB_TARGET_Y);
	writel(target_rect->w, fb->reg_base + FB_TARGET_WIDTH);
	writel(target_rect->h, fb->reg_base + FB_TARGET_HEIGHT);

	bcm2708fb_debug("src_rect(%i, %i, %i, %i)\n", src_rect->x, src_rect->y, src_rect->w, src_rect->h);
	bcm2708fb_debug("target_rect(%i, %i, %i, %i)\n", target_rect->x, target_rect->y, target_rect->w, target_rect->h);

	if (fb->is_yuv) {
		fmt = (struct yuv_pixfmt *)&var->reserved[4];
		pixel_format = (fmt->colour_space << COLOUR_SPACE_SHIFT) | (fmt->fmt & PIXEL_FORMAT_MASK);
		writel(pixel_format, fb->reg_base + FB_SET_FORMAT);
		bcm2708fb_debug("YUV format fmt=(%i), col=(%i)\n", fmt->fmt, fmt->colour_space);
	}

	/* Dont setup the opacity if we are blanked out already */
	if (!fb->is_suspended) {
		if (fb->is_yuv)
			writel(255, fb->reg_base + FB_SET_OPACITY); /* Maximum opacity value for yuv framebuffers */
		else
			writel((var->reserved[4] & 255), fb->reg_base + FB_SET_OPACITY);
	}

	/* Tell FB to update display .. */
	control &= ~(CTRL_BUFF_IDX_MASK);
	/* Write active buffer and update bit to control register */
	control |= (buff_idx << CTRL_BUFF_IDX_SHIFT);
	control |= (1 << CTRL_UPDT_EN_SHIFT);

	fb->update_count++;

	/* Never round off to zero, bad things will happen
	 * since the code above checks for update_count before it
	 * checks for vsynd pending status
	 *
	 * Update count in status register also behaves the same way
	 * now
	 */

	if (!fb->update_count)
		fb->update_count++;

#ifdef DEBUG_UPDATE_COUNT
	control &= ~(CTRL_UPDATE_COUNT_MASK);
	control |= (fb->update_count << CTRL_UPDATE_COUNT_SHIFT);
#endif
	writel(control, fb->reg_base + FB_CONTROL);

	/* ring the doorbell */
	ipc_notify_vc_event(fb->irq);

	spin_unlock_irqrestore(&fb->lock, irq_flags);

out:
	return ret;
}

static struct fb_ops bcm2708_fb0_ops = {
	.owner          = THIS_MODULE,
	.fb_ioctl	= bcm2708_fb_ioctl,
	.fb_blank	= bcm2708_fb_blank,
	.fb_check_var   = bcm2708_fb_check_var,
	.fb_set_par     = bcm2708_fb_set_par,
	.fb_setcolreg   = bcm2708_fb_setcolreg,
	.fb_pan_display = bcm2708_fb_pan_display,
};

static struct fb_ops bcm2708_fb_ops = {
	.owner          = THIS_MODULE,
	.fb_open	= bcm2708_fb_open,
	.fb_ioctl	= bcm2708_fb_ioctl,
	.fb_blank	= bcm2708_fb_blank,
	.fb_release	= bcm2708_fb_release,
	.fb_check_var   = bcm2708_fb_check_var,
	.fb_set_par     = bcm2708_fb_set_par,
	.fb_pan_display = bcm2708_fb_pan_display,
};

static int __init init_fb_info(struct bcm2708_fb *fb)
{
	uint32_t width, height;
	uint16_t ovl_pixel_format;
	int ret = 0;
    bool double_buffered=true;

	/* Is this overlay framebuffer */
	if (strcmp(fb->dev_name, "bcm2835_FBUF")) {

		if (!strcmp(fb->dev_name, "bcm2835_OVL1")) {
			fb->fbnum = 1;
			width = ovl1_width;
			height = ovl1_height;
			ovl_pixel_format = ovl1_pixel_format;
            double_buffered = ovl1_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL2")) {
			fb->fbnum = 2;
			width = ovl2_width;
			height = ovl2_height;
			ovl_pixel_format = ovl2_pixel_format;
            double_buffered = ovl2_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL3")) {
			fb->fbnum = 3;
			width = ovl3_width;
			height = ovl3_height;
			ovl_pixel_format = ovl3_pixel_format;
            double_buffered = ovl3_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL4")) {
			fb->fbnum = 4;
			width = ovl4_width;
			height = ovl4_height;
			ovl_pixel_format = ovl4_pixel_format;
            double_buffered = ovl4_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL5")) {
			fb->fbnum = 5;
			width = ovl5_width;
			height = ovl5_height;
			ovl_pixel_format = ovl5_pixel_format;
            double_buffered = ovl5_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL6")) {
			fb->fbnum = 6;
			width = ovl6_width;
			height = ovl6_height;
			ovl_pixel_format = ovl6_pixel_format;
            double_buffered = ovl6_double;
		} else if (!strcmp(fb->dev_name, "bcm2835_OVL7")) {
			fb->fbnum = 7;
			width = ovl7_width;
			height = ovl7_height;
			ovl_pixel_format = ovl7_pixel_format;
            double_buffered = ovl7_double;
		}

		/* Its not really mandatory to have this alignment on framebuffer, but YUV Images
		 * for HVS are always allocated with this alignment. if the resolution is not aligned,
		 * framebuffer will be allocated with padding at the end of each Y-line (for width) and
		 * at the end of Y-plane (for height).
		 *
		 * Also, in that case, application will have to write the data by taking this padding into account.
		 * which never gets communicated.
		 *
		 * If we force the alignment, no special care has to be taken by
		 * applications writing into framebuffer.
		 */

		if ((width & 31) || (height & 15)) {
			bcm2708fb_error("For overlay framebuffers width has to be aligned to 32 and height to 16\n");
			bcm2708fb_error("Cant set resolution to (%ix%i)\n", width, height);
			return -EINVAL;
		}

		if (ovl_pixel_format == FB_PIXEL_FORMAT_YUV_NV12 ||
			ovl_pixel_format ==	FB_PIXEL_FORMAT_YUV_NV21)
		{
			init_yuv_fb_info(fb, width, height, ovl_pixel_format, &bcm2708_fb_ops);
		}
		else{
			init_rgb_fb_info(fb, width, height, double_buffered, &bcm2708_fb_ops);
		}
		atomic_set(&fb->usage, 0);

	} else {

		if (fb0_cache_policy && (fb0_cache_policy <= FB0_CACHE_POLICY_WRITETHROUGH))
		/* if (fb0_cache_policy && (fb0_cache_policy < FB0_CACHE_POLICY_WRITEBACK)) */
			bcm2708_fb0_ops.fb_mmap = bcm2708_fb_mmap;

		fb->fbnum = 0;
		init_rgb_fb_info(fb, fb_width, fb_height, true, &bcm2708_fb0_ops);
	}

	return ret;

}

static int bcm2708_fb_probe(struct platform_device *pdev)
{
	int ret = -ENXIO;
	struct resource *r;
	struct bcm2708_fb *fb;
	size_t framesize;

	fb = kzalloc(sizeof(struct bcm2708_fb), GFP_KERNEL);
	if (fb == NULL) {
		bcm2708fb_error("Unable to allocate framebuffer structure\n");
		ret = -ENOMEM;
		goto err_fb_alloc_failed;
	}

#ifdef DEBUG_FRAMEBUFFER
	global_hack = fb;
#endif

	spin_lock_init(&fb->lock);
	init_waitqueue_head(&fb->wait_for_irq);
	fb->dev_name = kzalloc(FB_DEV_NAME_SIZE_MAX, GFP_KERNEL);
	if (fb->dev_name == NULL) {
		bcm2708fb_error("Unable to allocate memory for FB device name\n");
		ret = -ENOMEM;
		goto err_no_dev_name;
	}

	strcpy(fb->dev_name, pdev->name);
	platform_set_drvdata(pdev, fb);

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(r == NULL) {
		bcm2708fb_error("Unable to get framebuffer memory resource\n");
		ret = -ENODEV;
		goto err_no_io_base;
	}

	bcm2708fb_info("FB registers start-end (0x%08x)-(0x%08x)\n", r->start, r->end);

	/* 
	 * IPC driver puts virtual address here so we dont need to
	 * use the __io_address()
	 */
	fb->reg_base = (void __iomem *)r->start;

	fb->irq = platform_get_irq(pdev, 0);
	if(fb->irq < 0) {
		bcm2708fb_error("Unable to get framebuffer irq resource\n");
		ret = -ENODEV;
		goto err_no_irq;
	}
	
	ret = request_irq(fb->irq, bcm2708_fb_interrupt, IRQF_DISABLED,
				"bcm2708 fb interrupt", (void *)fb);
	if (ret) {
		bcm2708fb_error("Unable to register Interrupt for bcm2708 FB\n");
		goto err_no_irq;
	}

	ret = init_fb_info(fb);
	if (ret) {
		bcm2708fb_error("Failed to initialise fb info with given params\n");
		goto err_init_fb_info;
	}

	if (fb->fbnum == 0) {
		ret = enable_fb_device(fb, &fb->fb.var);
		if (ret) {
			bcm2708fb_error("Failed to enable FB device\n");
			goto err_enable_fb_failed;
		}

		framesize = fb->fb.var.xres * fb->fb.var.yres_virtual * fb->fb.var.bits_per_pixel/8;


		fb->fb.fix.smem_start = fb->phys_fbbase;
		fb->fb.fix.smem_len = framesize;

		bcm2708fb_info("Requesting and mapping resource (0x%08x)-(0x%08x)\n", fb->phys_fbbase, fb->phys_fbbase + framesize);

		fb->fb.screen_base = ioremap(fb->phys_fbbase, framesize); 
		if (fb->fb.screen_base == NULL) {
			ret = -ENOMEM;
			bcm2708fb_error("Unable to ioremap fb memory resource\n");
			goto err_ioremap_fbmem_failed;
		}
		bcm2708fb_debug("IORMAP -> (0x%08x)\n", fb->fb.screen_base);

		bcm2708fb_info("Framebuffer starts at phys[0x%08x], and virt[0x%08x]\n",
				fb->phys_fbbase, (uint32_t)fb->fb.screen_base);
	}

	ret = register_framebuffer(&fb->fb);
	if (ret) {
		bcm2708fb_error("Framebuffer registration failed\n");
		goto err_fb_register_failed;
	}

	bcm2708fb_info("BCM2708 Framebuffer probe successfull\n");

#ifdef CONFIG_ANDROID_POWER
	fb->early_suspend.suspend = bcm2708_fb_early_suspend;
	fb->early_suspend.resume = bcm2708_fb_late_resume;
	android_register_early_suspend(&fb->early_suspend);
#endif

	return 0;

err_fb_register_failed:
	bcm2708fb_debug("FB(%d) : IOUMAP -> (0x%08x)\n", fb->fbnum, fb->fb.screen_base);
	iounmap(fb->fb.screen_base);
err_ioremap_fbmem_failed:
	disable_fb_device(fb);
err_enable_fb_failed:
err_init_fb_info :
	free_irq(fb->irq, (void *)fb);
err_no_irq:
err_no_io_base:
	kfree(fb->dev_name);
err_no_dev_name:
	kfree(fb);
err_fb_alloc_failed:
	bcm2708fb_alert("BCM2708 Framebuffer probe FAILED !!\n");
	return ret;
}

static int __devexit bcm2708_fb_remove(struct platform_device *pdev)
{
	struct bcm2708_fb *fb = platform_get_drvdata(pdev);

#ifdef CONFIG_ANDROID_POWER
        android_unregister_early_suspend(&fb->early_suspend);
#endif
	unregister_framebuffer(&fb->fb);
	if (fb->fbnum == 0) {
		bcm2708fb_debug("FB(%d) : IOUMAP -> (0x%08x)\n", fb->fbnum, fb->fb.screen_base);
		iounmap(fb->fb.screen_base);
		/* release_region(fb->fbmem_res->start, resource_size(fb->fbmem_res)); */
		disable_fb_device(fb);
	}
	free_irq(fb->irq, (void *)fb);
	kfree(fb);
	bcm2708fb_info("BCM2708 FB removed !!\n");
	return 0;
}

static struct platform_driver bcm2708_fb0_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_FBUF"
	}
};

static struct platform_driver bcm2708_fb1_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL1"
	}
};

static struct platform_driver bcm2708_fb2_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL2"
	}
};

static struct platform_driver bcm2708_fb3_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL3"
	}
};

static struct platform_driver bcm2708_fb4_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL4"
	}
};

static struct platform_driver bcm2708_fb5_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL5"
	}
};

static struct platform_driver bcm2708_fb6_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL6"
	}
};

static struct platform_driver bcm2708_fb7_driver = {
	.probe		= bcm2708_fb_probe,
	.remove		= __devexit_p(bcm2708_fb_remove),
	.driver = {
		.name = "bcm2835_OVL7"
	}
};

static bool check_and_align_param(const char* name, int *value, int min, int max)
{
	if (*value <=min || *value > max) {
		printk(KERN_ERR"(%s) should be > than (%i) and <= (%i)\n", 
			name,min,max);
		return false;
	}

	if (*value & 15) {
		*value = (*value + 15) & ~15;
		bcm2708fb_warning("Alligned %s to (%d)\n", name, *value);
	}

	return true;
}

static int __init check_module_params(void)
{
	if (!check_and_align_param("fbwidth"   ,&fb_width    , 0 , FRAMEBUFFER_WIDTH)) return -EINVAL;
	if (!check_and_align_param("fbheight"  ,&fb_height   , 0 , FRAMEBUFFER_HEIGHT)) return -EINVAL;
	if (!check_and_align_param("ovl1width" ,&ovl1_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl1height",&ovl1_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl2width" ,&ovl2_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl2height",&ovl2_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl3width" ,&ovl3_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl3height",&ovl3_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl4width" ,&ovl4_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl4height",&ovl4_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl5width" ,&ovl5_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl5height",&ovl5_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl6width" ,&ovl6_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl6height",&ovl6_height , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl7width" ,&ovl7_width  , 0 , 4096)) return -EINVAL;
	if (!check_and_align_param("ovl7height",&ovl7_height , 0 , 4096)) return -EINVAL;

	if (fb0_pixel_format > FB_PIXEL_FORMAT_BGR888) {
		printk(KERN_ERR"Invalid (fb0pixfmt), Setting to default ARGB8888\n");
		fb0_pixel_format = FB_PIXEL_FORMAT_RGBA32;
	}

	if ((ovl1_pixel_format < FB_PIXEL_FORMAT_YUV_NV12) &&
			(ovl1_pixel_format > FB_PIXEL_FORMAT_YUV_NV21)) {
		printk(KERN_ERR"Invalid (ovl1pixfmt), You should always start with (%d) or (%d)\n",
					FB_PIXEL_FORMAT_YUV_NV12, FB_PIXEL_FORMAT_YUV_NV21);
		printk(KERN_WARNING"Setting (ovl1pixfmt) to default (%d)\n", FB_PIXEL_FORMAT_YUV_NV12);
		ovl1_pixel_format = FB_PIXEL_FORMAT_YUV_NV12;
	}

	if ((ovl2_pixel_format < FB_PIXEL_FORMAT_YUV_NV12) &&
			(ovl2_pixel_format > FB_PIXEL_FORMAT_YUV_NV21)) {
		printk(KERN_ERR"Invalid (ovl2pixfmt), You should always start with (%d) or (%d)\n",
					FB_PIXEL_FORMAT_YUV_NV12, FB_PIXEL_FORMAT_YUV_NV21);
		printk(KERN_WARNING"Setting (ovl2pixfmt) to default (%d)\n", FB_PIXEL_FORMAT_YUV_NV21);
		ovl1_pixel_format = FB_PIXEL_FORMAT_YUV_NV21;
	}

	if (fb0_cache_policy > FB0_CACHE_POLICY_WRITETHROUGH) {
		printk(KERN_ERR"Invalid (fb0cache) value, accepted range is (%d - %d)\n",
					FB0_CACHE_POLICY_UNCACHED, FB0_CACHE_POLICY_WRITETHROUGH);
		printk(KERN_WARNING"Setting (fb0cache) to default (Uncached) cache policy\n");
		fb0_cache_policy = FB0_CACHE_POLICY_UNCACHED;
	}

	return 0;
}

static int __init bcm2708_fb_init(void)
{
	int ret;

	/* Check if module params are valid */
	ret = check_module_params();
	if (ret) {
		printk(KERN_ERR"%s : Invalid module parameters\n", __func__);
		goto out;
	}

	ret = platform_driver_register(&bcm2708_fb0_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB0 driver\n", __func__);
		goto out;
	}

	ret = platform_driver_register(&bcm2708_fb1_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB1 driver\n", __func__);
		goto out_fb1;
	}

	ret = platform_driver_register(&bcm2708_fb2_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB2 driver\n", __func__);
		goto out_fb2;
	}

	ret = platform_driver_register(&bcm2708_fb3_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB3 driver\n", __func__);
		goto out_fb3;
	}

	ret = platform_driver_register(&bcm2708_fb4_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB4 driver\n", __func__);
		goto out_fb4;
	}

	ret = platform_driver_register(&bcm2708_fb5_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB5 driver\n", __func__);
		goto out_fb5;
	}

	ret = platform_driver_register(&bcm2708_fb6_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB6 driver\n", __func__);
		goto out_fb6;
	}

	ret = platform_driver_register(&bcm2708_fb7_driver);
	if (ret) {
		printk(KERN_ERR"%s : Unable to register BRCM FB7 driver\n", __func__);
		goto out_fb7;
	}


	printk(KERN_INFO"BRCM Framebuffer Init %s !\n", ret ? "FAILED" : "OK");

	return ret;

out_fb7:
	platform_driver_unregister(&bcm2708_fb6_driver);
out_fb6:
	platform_driver_unregister(&bcm2708_fb5_driver);
out_fb5:
	platform_driver_unregister(&bcm2708_fb4_driver);
out_fb4:
	platform_driver_unregister(&bcm2708_fb3_driver);
out_fb3:
	platform_driver_unregister(&bcm2708_fb2_driver);
out_fb2:
	platform_driver_unregister(&bcm2708_fb1_driver);
out_fb1:
	platform_driver_unregister(&bcm2708_fb0_driver);
out:
	return ret;
}

static void __exit bcm2708_fb_exit(void)
{
	/* Clean up .. */
	platform_driver_unregister(&bcm2708_fb0_driver);
	platform_driver_unregister(&bcm2708_fb1_driver);
	platform_driver_unregister(&bcm2708_fb2_driver);
	platform_driver_unregister(&bcm2708_fb3_driver);
	platform_driver_unregister(&bcm2708_fb4_driver);
	platform_driver_unregister(&bcm2708_fb5_driver);
	platform_driver_unregister(&bcm2708_fb6_driver);
	platform_driver_unregister(&bcm2708_fb7_driver);

	printk(KERN_INFO"BRCM Framebuffer exit OK\n");
}

module_init(bcm2708_fb_init);
module_exit(bcm2708_fb_exit);

MODULE_AUTHOR("Broadcom");
MODULE_DESCRIPTION("BCM2708 FB Driver");
MODULE_LICENSE("GPL");
