#include <linux/device.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/syscalls.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include "audio_regs.h"
#include "bcm2835.h"

#ifdef SHORT_FUNCTION_NAMES
#define bcm2835_control_set_clear b_c_s_c
#endif

#ifdef AUDIO_VERBOSE_DEBUG_ENABLE
/* Must be called with ARM<->VC fifo lock held */
static inline void dump_fifo(bcm2835_alsa_stream_t *alsa_stream)
{

//	audio_debug("Out fifo->\n");
	audio_debug("Out fifo-> write - 0x%08x read - 0x%08x base - 0%08x size - %d entry_size - %d",
      *(uint32_t *)alsa_stream->out_fifo.write, *(uint32_t *)alsa_stream->out_fifo.read, (uint32_t)alsa_stream->out_fifo.base, alsa_stream->out_fifo.size,
			alsa_stream->out_fifo.entry_size);

//	audio_debug("In fifo->\n");
	audio_debug("In fifo-> write - 0x%08x read - 0x%08x base - 0%08x size - %d entry_size - %d",
      *(uint32_t *)alsa_stream->in_fifo.write, *(uint32_t *)alsa_stream->in_fifo.read, (uint32_t)alsa_stream->in_fifo.base, alsa_stream->in_fifo.size,
			alsa_stream->in_fifo.entry_size);


}
#else
#define dump_fifo(x) do {} while(0)
#endif

/* routine to atomically manipulate the volatile control word */
/* this guards against the interrupt routine manipulating it while we are manipulating it */
/* It also makes it convenient to check for a particular crash condition */
void bcm2835_control_set_clear(bcm2835_alsa_stream_t *alsa_stream, uint32_t setmask, uint32_t clearmask)
{
//	unsigned long flags;
    uint32_t ctl = readl(alsa_stream->reg_base + AUDIO_CONTROL_OFFSET);
    uint32_t lockup = ctl & setmask & (AUDIO_CONTROL_PLAY|AUDIO_CONTROL_ENABLE);
    audio_debug(" .. IN: setm=%x clrm=%x ctl=%x lckup=%x\n",setmask,clearmask,ctl,lockup);
    if (lockup)
    {
        audio_alert(" .. lockup : warning setting bit, already set ctl=%x setmask=%x\n",ctl,setmask);
        BUG_ON(lockup);
    }
	BUG_ON(!(ctl & AUDIO_CONTROL_ENABLE));

//	spin_lock_irqsave(&alsa_stream->lock, flags);
	ctl = readl(alsa_stream->reg_base + AUDIO_CONTROL_OFFSET);
	ctl |= setmask;
	ctl &= ~clearmask;
	alsa_stream->control = ctl;
	writel(ctl, alsa_stream->reg_base + AUDIO_CONTROL_OFFSET);
//	spin_unlock_irqrestore(&alsa_stream->lock, flags);
	/* ring the doorbell */
	ipc_notify_vc_event(alsa_stream->irq);
}

static irqreturn_t bcm2835_audio_irq(int irq, void *dev_id)
{
	bcm2835_alsa_stream_t *alsa_stream = (bcm2835_alsa_stream_t *)dev_id;
  uint32_t status, changed;


  // Process any status changes
  status  = readl(alsa_stream->reg_base + AUDIO_STATUS_OFFSET);
  changed = status ^ alsa_stream->status;
  audio_debug(" .. IN status=%x changed=%x\n",status,changed);

  alsa_stream->status = status;
  // AUDIO_CONTROL_ENABLE is edge triggered
  if ((changed & AUDIO_CONTROL_ENABLE) != 0)
  {
    audio_error("enable changed to %08x", status);
    up( &alsa_stream->enable_sem );
  }
  // AUDIO_CONTROL_SETUP is level triggered
  else if ((status & AUDIO_CONTROL_SETUP) != 0)
  {
    audio_debug("setup complete %08x", status);
    if (status & AUDIO_CONTROL_ERROR)
    {
      audio_alert(": setup failure %08x\n", status);
      alsa_stream->setup_err = true;
      // video core will reset AUDIO_CONTROL_ERROR and AUDIO_CONTROL_SETUP when arm
      // clears the AUDIO_CONTROL_SETUP bit in the control resigter
    }
    up(&alsa_stream->setup_sem);
  }
  // AUDIO_CONTROL_FLUSH is level triggered
  else if ((status & AUDIO_CONTROL_FLUSH) != 0)
  {
    audio_debug("flush complete %08x", status);
    up(&alsa_stream->flush_sem);
  }

  if (alsa_stream->in_fifo.write != NULL)
  {
   // ignore audioclock(pos) return value, videocore can't keep track of it properly
   // alsa_stream->pos = readl(alsa_stream->reg_base + AUDIO_AUDIOCLOCK_OFFSET);
   // We track the incoming buffers from videocore

    unsigned int current_returned_write = *(uint32_t *)alsa_stream->in_fifo.write;
    unsigned int delta = current_returned_write - alsa_stream->last_in_fifo_write;
    if (delta)
    {
      alsa_stream->pos += delta * AUDIO_IPC_BLOCK_BUFFER_SIZE;
      alsa_stream->last_in_fifo_write = current_returned_write;
    }
    audio_debug("delta=%d pos %08x", delta, alsa_stream->pos);
    audio_debug("ARM ->  VC r %08x w %08x", *alsa_stream->out_fifo.read, *alsa_stream->out_fifo.write);
    audio_debug("VC  -> ARM r %08x w %08x", *alsa_stream->in_fifo.read,  *alsa_stream->in_fifo.write);
  }

  if (alsa_stream->enable_fifo_irq && alsa_stream->fifo_irq_handler)
    (void) alsa_stream->fifo_irq_handler(irq, dev_id);

	audio_debug(" .. OUT\n");

	return IRQ_HANDLED;
}

int bcm2835_audio_open(bcm2835_alsa_stream_t *alsa_stream)
{
	int err;
  uint32_t control;
#ifdef DUMP_RAW_DATA
	mm_segment_t old_fs;
	int open_flags;
#endif


	err = request_irq(alsa_stream->irq, bcm2835_audio_irq, IRQF_DISABLED, "bcm2835 audio irq", (void *)alsa_stream);
  if (err < 0)
  {
		printk(KERN_ERR"Failed to request IRQ for bcm2835 playback\n");
    return err;
	}

	/* setup the destination and enable device*/
	writel(alsa_stream->dest, alsa_stream->reg_base + AUDIO_DESTINATION_OFFSET);
  control = readl(alsa_stream->reg_base + AUDIO_CONTROL_OFFSET);
  if ((control & AUDIO_CONTROL_ENABLE) != 0)
  {
		audio_alert("Alsa device was already enabled !\n");
    return -EPERM;
  }

  control &= ~AUDIO_CONTROL_MUTE;
  control |= AUDIO_CONTROL_ENABLE;
		writel(alsa_stream->volume, alsa_stream->reg_base + AUDIO_VOLUME_OFFSET);
  writel(control,             alsa_stream->reg_base + AUDIO_CONTROL_OFFSET);

		/* ring the doorbell */
		ipc_notify_vc_event(alsa_stream->irq);

	/* wait for it .. */
  audio_debug("waiting for audio to be enabled ..\n");
  down(&alsa_stream->enable_sem);

  ipc_fifo_setup_no_reset(&alsa_stream->out_fifo, (alsa_stream->reg_base + AUDIO_ARM_TO_VIDEOCORE_WRITE_PTR_OFFSET),
          (alsa_stream->reg_base + AUDIO_ARM_TO_VIDEOCORE_READ_PTR_OFFSET),
          ipc_bus_to_virt(readl(alsa_stream->reg_base + AUDIO_ARM_TO_VIDEOCORE_FIFO_START_OFFSET)),
          readl(alsa_stream->reg_base + AUDIO_ARM_TO_VIDEOCORE_FIFO_SIZE_OFFSET),
          readl(alsa_stream->reg_base + AUDIO_ARM_TO_VIDEOCORE_FIFO_ENTRY_OFFSET));

  ipc_fifo_setup_no_reset(&alsa_stream->in_fifo, (alsa_stream->reg_base + AUDIO_VIDEOCORE_TO_ARM_WRITE_PTR_OFFSET),
          (alsa_stream->reg_base + AUDIO_VIDEOCORE_TO_ARM_READ_PTR_OFFSET),
          ipc_bus_to_virt(readl(alsa_stream->reg_base + AUDIO_VIDEOCORE_TO_ARM_FIFO_START_OFFSET)),
          readl(alsa_stream->reg_base + AUDIO_VIDEOCORE_TO_ARM_FIFO_SIZE_OFFSET),
          readl(alsa_stream->reg_base + AUDIO_VIDEOCORE_TO_ARM_FIFO_ENTRY_OFFSET));
  alsa_stream->out_remaining_bytes = 0;
  alsa_stream->out_buffer.buffer_size = 0;

	dump_fifo(alsa_stream);

#ifdef DUMP_RAW_DATA
	/* save the sys stack */
	old_fs = get_fs();
	set_fs(get_ds());

	// open the file
	open_flags = O_WRONLY | O_CREAT;

	alsa_stream->file = sys_open("/system/alsa.raw", open_flags, 0777);
	if (alsa_stream->file < 0) {
		printk(KERN_EMERG"%s : Failed to open raw file\n",__func__);
	}
	set_fs(old_fs);
#endif /* DUMP_RAW_DATA */

	audio_debug(" .. OUT\n");

	return 0;
}

void bcm2835_audio_close(bcm2835_alsa_stream_t *alsa_stream)
{
	int waits;
#ifdef DUMP_RAW_DATA
	mm_segment_t old_fs;
#endif
	audio_debug(" .. IN\n");

	/* Stop period elapsed irqs right now */
	alsa_stream->enable_fifo_irq = 0;

	/* Free up all buffers in our queue */
	bcm2835_audio_flush_buffers(alsa_stream);

	/* Wait till we get playback stopped ack .. */
  audio_debug("Waiting for playback to stop ..\n");

  for (waits = 0; waits < 50; waits++) {
    if ((alsa_stream->status & AUDIO_CONTROL_PLAY) == 0) break;
    msleep(20);
  }

  bcm2835_control_set_clear(alsa_stream, 0, AUDIO_CONTROL_ENABLE);

	/* wait for it .. (with timeout ??) */
  audio_debug("waiting for audio to be disabled ..\n");
  down(&alsa_stream->enable_sem);

  audio_debug("Freeing alsa irq (%d)\n", alsa_stream->irq);
	free_irq(alsa_stream->irq, (void *)alsa_stream);

#ifdef DUMP_RAW_DATA
	old_fs = get_fs();
	set_fs(get_ds());
	sys_close(alsa_stream->file);
	set_fs(old_fs);
#endif
	audio_debug(" .. OUT\n");
	return;
}

int bcm2835_audio_set_params(bcm2835_alsa_stream_t *alsa_stream, uint32_t channels, uint32_t samplerate, uint32_t bps)
{
	audio_debug(" .. IN\n");

	if (channels < 1 || channels > 2) {
		audio_error(" channels (%d) not supported\n", channels);
		return -EINVAL;
	}

	switch (samplerate) {
        case 44100:
        case 8000:
        case 11025:
        case 16000:
        case 22050:
        case 32000:
        case 48000:
        case 64000:
        case 88200:
        case 96000:
        case 176400:
        case 192000:
                        break;
        default:
		    audio_error(" samplerate (%d) not supported\n", samplerate);
		    return -EINVAL;
	}

	if (bps != 8 && bps != 16) {
		audio_error(" Bits per sample (%d) not supported\n", bps);
		return -EINVAL;
	}

  audio_debug(" Setting ALSA channels(%d), samplerate(%d), bits-per-sample(%d)\n", channels, samplerate, bps);

	writel(channels, alsa_stream->reg_base + AUDIO_CHANNELS_OFFSET);
	writel(samplerate, alsa_stream->reg_base + AUDIO_SAMPLE_RATE_OFFSET);
	writel(bps, alsa_stream->reg_base + AUDIO_BIT_RATE_OFFSET);

	audio_debug(" .. OUT\n");
	return 0;
}

int bcm2835_audio_setup(bcm2835_alsa_stream_t *alsa_stream)
{
  int err = 0;
  audio_debug(" .. IN(%p)\n",alsa_stream);

  alsa_stream->setup_err = false;
  bcm2835_control_set_clear(alsa_stream, AUDIO_CONTROL_SETUP, 0);

	alsa_stream->enable_fifo_irq = 1;

    // Wait for ack
  audio_debug("waiting for audio to be setup ..\n");
  down(&alsa_stream->setup_sem);

  // AUDIO_CONTROL_SETUP is level triggered,
  // reset to 0 so next command won't trigger a new setup
  bcm2835_control_set_clear(alsa_stream, 0, AUDIO_CONTROL_SETUP);

  // return error if needed
  if (alsa_stream->setup_err) err = -1;
  audio_debug(" .. OUT err=%d\n", err);

  return err;
}

int bcm2835_audio_start(bcm2835_alsa_stream_t *alsa_stream)
{
  audio_debug(" .. IN(%p)\n", alsa_stream);

  bcm2835_control_set_clear(alsa_stream, AUDIO_CONTROL_PLAY, 0);

  audio_debug(" .. OUT\n");

  return 0;
}

int bcm2835_audio_stop(bcm2835_alsa_stream_t *alsa_stream)
{
  audio_debug(" .. IN(%p) draining=%d\n", alsa_stream, alsa_stream->draining);

  if (alsa_stream->draining)
  {
    audio_debug(".. DRAINING\n");
    writel(0, alsa_stream->reg_base + AUDIO_STOPFLUSH_OFFSET);
  }
  else
  {
    audio_debug(".. DROPPING\n");
    writel(1, alsa_stream->reg_base + AUDIO_STOPFLUSH_OFFSET);
  }
  bcm2835_control_set_clear(alsa_stream, 0, AUDIO_CONTROL_PLAY);
  audio_debug(" .. OUT\n");

  return 0;
}

int bcm2835_audio_set_ctls(bcm2835_chip_t *chip)
{
	int i;

	audio_debug(" .. IN\n");

	/* change ctls for all substreams */
	for (i = 0; i < MAX_SUBSTREAMS; i++) {
		if (chip->avail_substreams & (1 << i)) {
			BUG_ON(!(chip->reg_base[i] && chip->irq[i]));
			writel(chip->volume[i], chip->reg_base[i] + AUDIO_VOLUME_OFFSET);
			writel(chip->dest[i], chip->reg_base[i] + AUDIO_DESTINATION_OFFSET);
			ipc_notify_vc_event(chip->irq[i]);
		}
	}

	audio_debug(" .. OUT\n");

	return 0;
}

static void send_buffer(bcm2835_alsa_stream_t *alsa_stream)
{
  // reset buffer
  alsa_stream->out_mapped[alsa_stream->out_map_id] -= alsa_stream->out_buffer.buffer_size;
  alsa_stream->out_map_id += 1;
  if (alsa_stream->out_map_id==AUDIO_IPC_BLOCK_NUM_BUFFERS)
{
     alsa_stream->out_map_id = 0;
  }
  BUG_ON(ipc_fifo_full(&alsa_stream->out_fifo));
  ipc_fifo_write(&alsa_stream->out_fifo, &alsa_stream->out_buffer);
			ipc_notify_vc_event(alsa_stream->irq);
		}

int bcm2835_audio_write(bcm2835_alsa_stream_t *alsa_stream, uint32_t count, void *src)
{
  audio_debug(" .. IN(%p) count=%d src=%p",alsa_stream, count, src);
//	dump_fifo(alsa_stream);

  while (count != 0 && (readl(alsa_stream->reg_base + AUDIO_CONTROL_OFFSET) & AUDIO_CONTROL_ENABLE) != 0)
  {
    size_t copy_size;

    audio_debug(" Control %08x remaining_bytes=%d fifoempty=%d\n", readl(alsa_stream->reg_base + AUDIO_CONTROL_OFFSET),
                             alsa_stream->out_remaining_bytes, ipc_fifo_empty(&alsa_stream->in_fifo));

    // Do we need another buffer?
    if (alsa_stream->out_remaining_bytes == 0)
    {
      // Wait for a buffer to be available
      while (ipc_fifo_empty(&alsa_stream->in_fifo))
        down(&alsa_stream->buffers_update_sem);

      // Read the FIFO entry
      ipc_fifo_read(&alsa_stream->in_fifo, &alsa_stream->out_buffer);
      BUG_ON(alsa_stream->out_buffer.buffer_ptr == 0);
      alsa_stream->out_remaining_bytes = alsa_stream->out_buffer.buffer_size;

      if (alsa_stream->out_mapped[alsa_stream->out_map_id] == NULL)
      {
         alsa_stream->out_mapped[alsa_stream->out_map_id] = (uint8_t *)ioremap(__bus_to_phys(alsa_stream->out_buffer.buffer_ptr), alsa_stream->out_buffer.buffer_size);
         BUG_ON(alsa_stream->out_mapped[alsa_stream->out_map_id] == NULL);

         audio_debug(" New buffer mapped at %p +%4x with map index %d for substream %p\n", alsa_stream->out_mapped[alsa_stream->out_map_id], alsa_stream->out_buffer.buffer_size,alsa_stream->out_map_id,alsa_stream->substream);
      }   
      else
      {
	 audio_debug(" Reuse buffer mapped at %p +%4x with map index %d for substream %p\n",alsa_stream->out_mapped[alsa_stream->out_map_id], alsa_stream->out_buffer.buffer_size, alsa_stream->out_map_id,alsa_stream->substream);
			}
		}

    copy_size = count > alsa_stream->out_remaining_bytes ? alsa_stream->out_remaining_bytes : count;
    audio_debug(" Copying into buffer (%d) +%4x\n", alsa_stream->out_buffer.buffer_id, copy_size);
    if (copy_from_user( alsa_stream->out_mapped[alsa_stream->out_map_id], (uint8_t *)src, copy_size))
        {
      audio_error(" Copying into buffer (%d) +%4x FAILED\n", alsa_stream->out_buffer.buffer_id, copy_size);
			return -EFAULT;
        }

		count -= copy_size;
    src += copy_size;
    alsa_stream->out_mapped[alsa_stream->out_map_id] += copy_size;
    alsa_stream->out_remaining_bytes -= copy_size;

    // Is out_buffer full?
    if (alsa_stream->out_remaining_bytes == 0)
      send_buffer(alsa_stream);
		}
  if (count != 0)   audio_error(" returning with unwritten samples count=%d\n", count);

	audio_debug(" .. OUT count=%d\n", count);
	return 0;
}

void bcm2835_audio_flush_buffers(bcm2835_alsa_stream_t *alsa_stream)
{
  audio_debug(" .. IN remaining bytes=%d\n", alsa_stream->out_remaining_bytes);

  if (alsa_stream->out_remaining_bytes != 0)
  {
    alsa_stream->out_remaining_bytes = 0;
    send_buffer(alsa_stream);
  }

  audio_debug(" .. OUT\n");
}

/**
  * Forces VC to flush(drop) its filled playback buffers and 
  * return them the us. (VC->ARM)
  */
/* This function is not used at this time, at one time it was used in snd_bcm2835_pcm_prepare */

void bcm2835_audio_flush_playback_buffers(bcm2835_alsa_stream_t *alsa_stream)
{
	audio_debug(" .. IN\n");

  bcm2835_control_set_clear(alsa_stream, AUDIO_CONTROL_FLUSH, 0);

  audio_debug("waiting for audio to flush ..\n");
  down(&alsa_stream->flush_sem);

  // AUDIO_CONTROL_FLUSH is level triggered,
  // reset to 0 so next command won't trigger a new flush
  bcm2835_control_set_clear(alsa_stream, 0, AUDIO_CONTROL_FLUSH);

  audio_debug(" .. OUT\n");
}
