#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include "audio_regs.h"
#include "bcm2835.h"

/* hardware definition */
static struct snd_pcm_hardware snd_bcm2835_playback_hw = {
	.info = ( SNDRV_PCM_INFO_INTERLEAVED |
		SNDRV_PCM_INFO_BLOCK_TRANSFER ),
	.formats =          SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S8,
	.rates =            SNDRV_PCM_RATE_8000_192000,
	.rate_min =         8000,
	.rate_max =         192000,
	.channels_min =     1,
	.channels_max =     2,
	.buffer_bytes_max = AUDIO_IPC_BLOCK_NUM_BUFFERS * AUDIO_IPC_BLOCK_BUFFER_SIZE,
	.period_bytes_min = AUDIO_IPC_BLOCK_BUFFER_SIZE,
	.period_bytes_max = AUDIO_IPC_BLOCK_BUFFER_SIZE,
	.periods_min =      1,
	.periods_max =      AUDIO_IPC_BLOCK_NUM_BUFFERS,
};

#if 0
/* hardware definition */
static struct snd_pcm_hardware snd_bcm2835_capture_hw = {
	.info = (SNDRV_PCM_INFO_MMAP |
			SNDRV_PCM_INFO_INTERLEAVED |
			SNDRV_PCM_INFO_BLOCK_TRANSFER |
			SNDRV_PCM_INFO_MMAP_VALID),
	.formats =          SNDRV_PCM_FMTBIT_S16_LE,
	.rates =            SNDRV_PCM_RATE_8000_48000,
	.rate_min =         8000,
	.rate_max =         48000,
	.channels_min =     2,
	.channels_max =     2,
	.buffer_bytes_max = 32768,
	.period_bytes_min = 4096,
	.period_bytes_max = 32768,
	.periods_min =      1,
	.periods_max =      1024,
};
#endif

static void snd_bcm2835_playback_free(struct snd_pcm_runtime *runtime)
{
	audio_debug("Freeing up alsa stream here ..\n");
	kfree(runtime->private_data);
}

static irqreturn_t bcm2835_playback_fifo_irq(int irq, void *dev_id)
{
	bcm2835_alsa_stream_t *alsa_stream = (bcm2835_alsa_stream_t *)dev_id;
	IPC_FIFO_T *in_fifo = &alsa_stream->in_fifo;

	audio_debug(" .. IN\n");

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

        if (!ipc_fifo_empty(in_fifo))
        {
		BUG_ON(!alsa_stream->open);
		up(&alsa_stream->buffers_update_sem);
    }

    if (alsa_stream->reported_pos != alsa_stream->pos) // avoid unnecessary divides
    {
	    if (alsa_stream->period_size && alsa_stream->reported_pos / alsa_stream->period_size != alsa_stream->pos / alsa_stream->period_size)
	    {
		    audio_debug("Calling snd_pcm_period_elapsed, offset %8x -> %8x\n", alsa_stream->reported_pos, alsa_stream->pos);
		    alsa_stream->reported_pos = alsa_stream->pos;
	snd_pcm_period_elapsed(alsa_stream->substream);
	    }
    }

	audio_debug(" .. OUT\n");

	return IRQ_HANDLED;
}

/* open callback */
static int snd_bcm2835_playback_open(struct snd_pcm_substream *substream)
{
	bcm2835_chip_t *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream;
	int idx;
	int err;

	audio_debug(" .. IN (%d)\n", substream->number);

	idx = substream->number;

	if (idx > MAX_SUBSTREAMS) {
		audio_error("substream(%d) device doesn't exist max(%d) substreams allowed\n", idx, MAX_SUBSTREAMS);
		err = -ENODEV;
		goto out;
	}

	/* Check if we are ready */
	if (!(chip->avail_substreams & (1 << idx))) {
		/* We are not ready yet */
		audio_error("substream(%d) device is not ready yet\n", idx);
		err = -EAGAIN;
		goto out;
	}

	BUG_ON(!(chip->reg_base[idx] && chip->irq[idx]));

	alsa_stream = kzalloc(sizeof(bcm2835_alsa_stream_t), GFP_KERNEL);
	if (alsa_stream == NULL) {
		return -ENOMEM;
	}

	/* Initialise alsa_stream */
	alsa_stream->chip = chip;
	alsa_stream->substream = substream;
	alsa_stream->reg_base = chip->reg_base[idx];
	alsa_stream->irq = chip->irq[idx];

	sema_init(&alsa_stream->buffers_update_sem, 0);
	sema_init(&alsa_stream->enable_sem, 0);
	sema_init(&alsa_stream->setup_sem, 0);
	sema_init(&alsa_stream->flush_sem, 0);

	/* Enabled in start trigger, called on each "fifo irq" after that */
	alsa_stream->enable_fifo_irq = 0;
	alsa_stream->fifo_irq_handler = bcm2835_playback_fifo_irq;

	runtime->private_data = alsa_stream;
	runtime->private_free = snd_bcm2835_playback_free;
	runtime->hw = snd_bcm2835_playback_hw;
	alsa_stream->volume = chip->volume[idx];
    alsa_stream->mute = chip->mute[idx];
    alsa_stream->dest = chip->dest[idx];
	alsa_stream->draining = 1;
        alsa_stream->out_map_id = 0;

	err = bcm2835_audio_open(alsa_stream);
	if (err != 0) {
		kfree(alsa_stream);
		return err;
	}

	alsa_stream->open = 1;
	
out:
	audio_debug(" .. OUT\n");

	return err;
}

/* close callback */
static int snd_bcm2835_playback_close(struct snd_pcm_substream *substream)
{
	/* the hardware-specific codes will be here */

	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = runtime->private_data;

	audio_debug(" .. IN\n");
	audio_debug("Alsa close\n");

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

	/*
	 * Call stop if it's still running. This happens when app
	 * is force killed and we don't get a stop trigger.
	 */
	if (alsa_stream->running) {
		int err;
		err = bcm2835_audio_stop(alsa_stream);
		alsa_stream->running = 0;
		if (err != 0)
			audio_error(" Failed to STOP alsa device\n");
	}

	alsa_stream->period_size = 0;
	alsa_stream->buffer_size = 0;

	if (alsa_stream->open) {
		alsa_stream->open = 0;
		bcm2835_audio_close(alsa_stream);
	}

	/*
	 * Do not free up alsa_stream here, it will be freed up by
	 * runtime->private_free callback we registered in *_open above
	 */

	alsa_stream->reg_base = NULL;
	alsa_stream->irq = 0;

	audio_debug(" .. OUT\n");

	return 0;
}

#if 0
/* open callback */
static int snd_bcm2835_capture_open(struct snd_pcm_substream *substream)
{
	bcm2835_chip_t *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	runtime->hw = snd_bcm2835_capture_hw;
	/* more hardware-initialization will be done here */
	....
		return 0;
}

/* close callback */
static int snd_bcm2835_capture_close(struct snd_pcm_substream *substream)
{
	bcm2835_chip_t *chip = snd_pcm_substream_chip(substream);
	/* the hardware-specific codes will be here */
	....
		return 0;

}
#endif

/* hw_params callback */
static int snd_bcm2835_pcm_hw_params(struct snd_pcm_substream *substream,
		struct snd_pcm_hw_params *params)
{
	int err;
    int waits;
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = (bcm2835_alsa_stream_t *)runtime->private_data;

	audio_debug(" .. IN(%p)\n", alsa_stream);

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

	err = snd_pcm_lib_malloc_pages(substream,
			params_buffer_bytes(params));
	if (err < 0) {
		audio_error(" pcm_lib_malloc failed to allocated pages for buffers\n");
		return err;
	}

    if  (alsa_stream->status & AUDIO_CONTROL_PLAY)
    {
	    audio_info("Waiting for playback to stop ..\n");
	    for (waits = 0; waits < 50; waits++) {
		    if (!(alsa_stream->status & AUDIO_CONTROL_PLAY)) break;
		    msleep(20);
        }
	}

	err = bcm2835_audio_set_params(alsa_stream, params_channels(params),
			params_rate(params), snd_pcm_format_width(params_format(params)));
	if (err < 0) {
		audio_error(" error setting hw params\n");
	}
    else {
        err = bcm2835_audio_setup(alsa_stream);
    }

	audio_debug(" .. OUT err=%d\n", err);

	return err;
}

/* hw_free callback */
static int snd_bcm2835_pcm_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

/* prepare callback */
static int snd_bcm2835_pcm_prepare(struct snd_pcm_substream *substream)
{
    int waits;
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = runtime->private_data;

	audio_debug(" .. IN(%p)\n",alsa_stream);

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

	/* Wait till we get playback stopped ack .. */
    if (alsa_stream->status & AUDIO_CONTROL_PLAY)
    {
    	audio_info("Waiting for playback to stop ..\n");

	    for (waits = 0; waits < 50; waits++) {
		    if (!(alsa_stream->status & AUDIO_CONTROL_PLAY)) break;
		    msleep(20);
        }
	}

	alsa_stream->buffer_size = snd_pcm_lib_buffer_bytes(substream);
	alsa_stream->period_size = snd_pcm_lib_period_bytes(substream);

	alsa_stream->pos = alsa_stream->reported_pos = 0;

    /* store current write fifo pointer of video core */
    alsa_stream->last_in_fifo_write = *alsa_stream->in_fifo.write;

    /* recover any partially filled buffer that hadn't been sent */
    if (alsa_stream->out_remaining_bytes > 0)
    {
        int bytes_in_buffer = alsa_stream->out_buffer.buffer_size - alsa_stream->out_remaining_bytes;
        if (bytes_in_buffer > 0)
        {
            audio_error("Recover partially filled buffer bytecount=%d", bytes_in_buffer);
            alsa_stream->out_remaining_bytes = alsa_stream->out_buffer.buffer_size;
            alsa_stream->out_mapped[alsa_stream->out_map_id] -= bytes_in_buffer;
        }
    }

	audio_debug(" .. OUT\n");
	return 0;
}

/* trigger callback */
static int snd_bcm2835_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = runtime->private_data;
	int err = 0;

	audio_debug(" .. IN(%p)\n", alsa_stream);

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

	switch (cmd) {
		case SNDRV_PCM_TRIGGER_START:
			audio_debug("bcm2835_AUDIO_TRIGGER_START\n");
			if (!alsa_stream->running) {
				err = bcm2835_audio_start(alsa_stream);
				if (err == 0)
				{
					alsa_stream->running = 1;
					alsa_stream->draining = 1;
				}
			}
			break;
		case SNDRV_PCM_TRIGGER_STOP:
			audio_debug("bcm2835_AUDIO_TRIGGER_STOP\n");
			if (runtime->status->state==SNDRV_PCM_STATE_DRAINING)
			{
				audio_debug("DRAINING\n");
				alsa_stream->draining = 1;
			}
                        else
                        {
                                audio_debug("DROPPING\n");
                                alsa_stream->draining=0;
                        }
			if (alsa_stream->running) {
				err = bcm2835_audio_stop(alsa_stream);
				if (err != 0)
					audio_error(" Failed to STOP alsa device\n");
				alsa_stream->running = 0;
			}
			break;
		default:
			err = -EINVAL;
	}

	audio_debug(" .. OUT\n");
	return err;
}

/* pointer callback */
static snd_pcm_uframes_t
snd_bcm2835_pcm_pointer(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = runtime->private_data;
	snd_pcm_sframes_t frames;
	unsigned int      bytes = alsa_stream->pos;

	audio_debug(" .. IN\n");

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

        frames = bytes_to_frames(runtime, bytes % alsa_stream->buffer_size);

	audio_debug("(%p) pcm_pointer %08x bytes=%08x frames %08x (%x))\n", alsa_stream, alsa_stream->pos, bytes,
                         (unsigned int) frames, alsa_stream->buffer_size);
	audio_debug(" .. OUT\n");

	return (snd_pcm_uframes_t) frames;
}

static int snd_bcm2835_pcm_copy(struct snd_pcm_substream *substream, int channel,
		snd_pcm_uframes_t pos, void *src, snd_pcm_uframes_t count)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	bcm2835_alsa_stream_t *alsa_stream = runtime->private_data;

	BUG_ON(!(alsa_stream->reg_base && alsa_stream->irq));

	audio_debug("copy... %x\n", frames_to_bytes(runtime, count));

	return bcm2835_audio_write(alsa_stream, frames_to_bytes(runtime, count), src);
}

#if 0
static int snd_bcm2835_pcm_silence(struct snd_pcm_substream *substream, int channel,
		snd_pcm_uframes_t post, snd_pcm_uframes_t count)
{
	//    memset(buf,0);
	//    bcm2835_audio_write(frames_to_bytes(runetime, count), buf);
}
#endif

/* operators */
static struct snd_pcm_ops snd_bcm2835_playback_ops = {
	.open =        snd_bcm2835_playback_open,
	.close =       snd_bcm2835_playback_close,
	.ioctl =       snd_pcm_lib_ioctl,
	.hw_params =   snd_bcm2835_pcm_hw_params,
	.hw_free =     snd_bcm2835_pcm_hw_free,
	.prepare =     snd_bcm2835_pcm_prepare,
	.trigger =     snd_bcm2835_pcm_trigger,
	.pointer =     snd_bcm2835_pcm_pointer,
	.copy    =     snd_bcm2835_pcm_copy,
	/* .silence =     snd_bcm2835_pcm_silence, */
};

#if 0
/* operators */
static struct snd_pcm_ops snd_bcm2835_capture_ops = {
	.open =        snd_bcm2835_capture_open,
	.close =       snd_bcm2835_capture_close,
	.ioctl =       snd_pcm_lib_ioctl,
	.hw_params =   snd_bcm2835_pcm_hw_params,
	.hw_free =     snd_bcm2835_pcm_hw_free,
	.prepare =     snd_bcm2835_pcm_prepare,
	.trigger =     snd_bcm2835_pcm_trigger,
	.pointer =     snd_bcm2835_pcm_pointer,
};
#endif
/*
 *  definitions of capture are omitted here...
 */

/* create a pcm device */
int __devinit snd_bcm2835_new_pcm(bcm2835_chip_t *chip)
{
	struct snd_pcm *pcm;
	int err;
        int index=0;

	err = snd_pcm_new(chip->card, "bcm2835 ALSA", 0, MAX_SUBSTREAMS, 0, &pcm);
	if (err < 0) 
		return err;
	pcm->private_data = chip;
	strcpy(pcm->name, "bcm2835 ALSA");
	chip->pcm = pcm;
	for (index=0;index<MAX_SUBSTREAMS;index++)
	{
		chip->dest[index] = AUDIO_DESTINATION_ALL;
		chip->volume[index] = 0;        /* 0.0 db */
	}
	/* set operators */
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
			&snd_bcm2835_playback_ops);
#if 0
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&snd_bcm2835_capture_ops);
#endif
	/* pre-allocation of buffers */
	/* NOTE: this may fail */
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_CONTINUOUS,
			snd_dma_continuous_data(GFP_KERNEL), 64*1024, 64*1024);

	audio_debug(" .. OUT\n");

	return 0;
}
