#ifndef __SOUND_ARM_BCM2835_H
#define __SOUND_ARM_BCM2835_H 

#define SUBSTREAM_NUM 1

#include <linux/device.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include "audio_regs.h"

#include <mach/ipc.h>
#include <mach/fifo.h>

/* #define DUMP_RAW_DATA */
//#define AUDIO_DEBUG_ENABLE
//#define AUDIO_VERBOSE_DEBUG_ENABLE

/* Debug output */
static void audio_output(const char *function, int line, const char *level, const char *format, ...) __attribute__((unused));
static void audio_output(const char *function, int line, const char *level, const char *format, ...)
{
  struct timespec time = current_kernel_time();
  va_list         args;
  va_start( args, format );
  printk("%s %8u.%03us %s:%d ", level, (unsigned int) time.tv_sec, (unsigned int) time.tv_nsec / (unsigned int) NSEC_PER_MSEC, function, line);
  vprintk(format, args);
  va_end( args );
}

//#define SHORT_FUNCTION_NAMES
#ifdef SHORT_FUNCTION_NAMES
/* used when turning on AUDIO_DEBUG_ENABLE to reduce verboseness of printk */
#define  bcm2835_audio_setup b_a_setup
#define bcm2835_audio_start b_a_start
#define bcm2835_audio_stop b_a_stop
#define bcm2835_audio_write b_a_wr
#define bcm2835_audio_retrieve_buffers b_a_rtrv
#define bcm2835_audio_set_params b_a_s_p
#endif

/* Debug macros */
#ifdef AUDIO_DEBUG_ENABLE
  #define audio_output_debug audio_output
#else
static void audio_output_debug(const char *function, int line, const char *level, const char *format, ...) __attribute__((unused));
static void audio_output_debug(const char *function, int line, const char *level, const char *format, ...)
{
  (void) function;
  (void) line;
  (void) level;
  (void) format;
}
#endif

#ifdef AUDIO_VERBOSE_DEBUG_ENABLE
  #define audio_debug(format, arg...) audio_output_debug(__func__, __LINE__, KERN_INFO,    format, ##arg)
#else
  #define audio_debug(format, arg...) audio_output_debug(__func__, __LINE__, KERN_DEBUG,   format, ##arg)
#endif

#define audio_info(   format, arg...) audio_output(__func__, __LINE__, KERN_INFO,    format, ##arg)
#define audio_error(  format, arg...) audio_output(__func__, __LINE__, KERN_ERR,     format, ##arg)
#define audio_warning(format, arg...) audio_output(__func__, __LINE__, KERN_WARNING, format, ##arg)
#define audio_alert(  format, arg...) audio_output(__func__, __LINE__, KERN_ALERT,   format, ##arg)

#define MAX_SUBSTREAMS			(8)
#define AVAIL_SUBSTREAMS_MASK		(0xff)

typedef enum {
	PCM_PLAYBACK_VOLUME,
	PCM_PLAYBACK_MUTE,
	PCM_PLAYBACK_DEVICE,
} SND_BCM2835_CTRL_T;

#define VOLUME_CONTROL_MIN_ID (1)
#define VOLUME_CONTROL_MAX_ID (MAX_SUBSTREAMS)
#define MUTE_CONTROL_MIN_ID (VOLUME_CONTROL_MAX_ID+1)
#define MUTE_CONTROL_MAX_ID (VOLUME_CONTROL_MAX_ID+MAX_SUBSTREAMS)
#define ROUTE_CONTROL_MIN_ID (MUTE_CONTROL_MAX_ID+1)
#define ROUTE_CONTROL_MAX_ID (MUTE_CONTROL_MAX_ID+MAX_SUBSTREAMS)

/* definition of the chip-specific record */
typedef struct bcm2835_chip {
	struct snd_card *card;
	struct snd_pcm *pcm;
	/* Bitmat for valid reg_base and irq numbers */
	uint32_t avail_substreams;
	void __iomem *reg_base[MAX_SUBSTREAMS];
	uint32_t irq[MAX_SUBSTREAMS];
	struct platform_device *pdev[MAX_SUBSTREAMS];

	int volume[MAX_SUBSTREAMS];
	int dest[MAX_SUBSTREAMS];
	int mute[MAX_SUBSTREAMS];
} bcm2835_chip_t;

typedef struct bcm2835_audio_buffer {
	uint32_t buffer_id;
	phys_addr_t	bus_addr;
	uint8_t __iomem	*start;
	uint32_t size;
	uint32_t data_left;
	struct list_head link;

} bcm2835_audio_buffer_t; 

typedef struct bcm2835_alsa_stream {
	bcm2835_chip_t *chip;
	struct snd_pcm_substream *substream;
	void __iomem *reg_base;
	uint32_t irq;

        uint32_t status;

	struct semaphore buffers_update_sem;
	struct semaphore enable_sem;
	struct semaphore setup_sem;
	struct semaphore flush_sem;

	int open;
	int running;
	int draining;

	int volume;
	int dest;
	int mute;
    uint32_t control;

#ifdef DUMP_RAW_DATA
	/* for debug */
	int file;
#endif
	unsigned int pos;
	unsigned int reported_pos;
	unsigned int buffer_size;
	unsigned int period_size;

    unsigned int last_in_fifo_write;

	/* out/in fifos, managed by the generic ipc code */
        AUDIO_FIFO_ENTRY_T out_buffer;
        int out_remaining_bytes;
        int out_map_id;
        void *out_mapped[AUDIO_IPC_BLOCK_NUM_BUFFERS];
	IPC_FIFO_T in_fifo;
	IPC_FIFO_T out_fifo;

	uint32_t enable_fifo_irq;
	irq_handler_t fifo_irq_handler;
    bool    fifo_irq_skipped;       // set to true if irq handler couldn't get the fifo_lock
    bool    setup_err;              // set to true by interrupt handler if setup failed on VC
} bcm2835_alsa_stream_t;

int snd_bcm2835_new_ctl(bcm2835_chip_t *chip);
int snd_bcm2835_new_pcm(bcm2835_chip_t *chip);

int bcm2835_audio_open(bcm2835_alsa_stream_t *alsa_stream);
void bcm2835_audio_close(bcm2835_alsa_stream_t *alsa_stream);
int bcm2835_audio_set_params(bcm2835_alsa_stream_t *alsa_stream, uint32_t channels, uint32_t samplerate, uint32_t bps);
int bcm2835_audio_setup(bcm2835_alsa_stream_t *alsa_stream);
int bcm2835_audio_start(bcm2835_alsa_stream_t *alsa_stream);
int bcm2835_audio_stop(bcm2835_alsa_stream_t *alsa_stream);
int bcm2835_audio_set_ctls(bcm2835_chip_t *chip);
int bcm2835_audio_write(bcm2835_alsa_stream_t *alsa_stream, uint32_t count, void *src);
//uint32_t bcm2835_audio_buffers_consumed_bytes(bcm2835_alsa_stream_t *alsa_stream);
uint32_t bcm2835_audio_retrieve_buffers(bcm2835_alsa_stream_t *alsa_stream);
void bcm2835_audio_flush_buffers(bcm2835_alsa_stream_t *alsa_stream);
void bcm2835_audio_flush_playback_buffers(bcm2835_alsa_stream_t *alsa_stream);

#endif /* __SOUND_ARM_BCM2835_H */
